/*
 * cal.c
 *
 * CPU Abstraction Layer implimentation.
 */

/* $Id: cal.c,v 1.12 1999/06/06 17:50:22 nyef Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include "flags.h"
#include "cal.h"

#ifdef CPU_MARAT6502
#include "m6502.h"
#endif

#ifdef CPU_EMU6502
#include "emu6502.h"
#endif

#ifdef CPU_EMU6502A
#include "emu6502a.h"
#endif

#ifdef CPU_EMU65C02
#include "emu65c02.h"
#endif

#ifdef CPU_EMU6280
#include "emu6280.h"
#endif

#ifdef CPU_MARATZ80
#include "mz80.h"
#endif

#ifdef CPU_EMUZ80
#include "emuz80.h"
#endif

#include "emu68k.h"

#ifdef CPU_MARAT6502
void cal_marat6502_init(cal_cpu *cpu);
void cal_marat6502_reset(cal_cpu cpu);
void cal_marat6502_run(cal_cpu cpu);
void cal_marat6502_runfor(cal_cpu cpu, int cycles);
void cal_marat6502_irq(cal_cpu cpu);
void cal_marat6502_nmi(cal_cpu cpu);
void cal_marat6502_setzpage(cal_cpu cpu, void *page0);
int cal_marat6502_timeleft(cal_cpu cpu);
void cal_marat6502_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl);
void cal_marat6502_setiou(cal_cpu cpu, memread8_t ioread, memwrite8_t iowrite);
#endif

#ifdef CPU_EMU6502
void cal_emu6502_init(cal_cpu *cpu);
void cal_emu6502_reset(cal_cpu cpu);
void cal_emu6502_run(cal_cpu cpu);
void cal_emu6502_runfor(cal_cpu cpu, int cycles);
void cal_emu6502_irq(cal_cpu cpu);
void cal_emu6502_nmi(cal_cpu cpu);
void cal_emu6502_setzpage(cal_cpu cpu, void *page0);
int cal_emu6502_timeleft(cal_cpu cpu);
void cal_emu6502_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl);
void cal_emu6502_setiou(cal_cpu cpu, memread8_t ioread, memwrite8_t iowrite);
#endif

#ifdef CPU_EMU6502A
void cal_emu6502a_init(cal_cpu *cpu);
void cal_emu6502a_reset(cal_cpu cpu);
#endif

#ifdef CPU_EMU65C02
void cal_emu65c02_init(cal_cpu *cpu);
void cal_emu65c02_reset(cal_cpu cpu);
#endif

#ifdef CPU_EMU6280
void cal_emu6280_init(cal_cpu *cpu);
void cal_emu6280_reset(cal_cpu cpu);
void cal_emu6280_irq(cal_cpu cpu);
#endif

#ifdef CPU_MARATZ80
void cal_maratz80_init(cal_cpu *cpu);
void cal_maratz80_reset(cal_cpu cpu);
void cal_maratz80_run(cal_cpu cpu);
void cal_maratz80_runfor(cal_cpu cpu, int cycles);
void cal_maratz80_irq(cal_cpu cpu);
void cal_maratz80_nmi(cal_cpu cpu);
void cal_maratz80_setzpage(cal_cpu cpu, void *page0);
int cal_maratz80_timeleft(cal_cpu cpu);
void cal_maratz80_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl);
void cal_maratz80_setiou(cal_cpu cpu, memread8_t ioread, memwrite8_t iowrite);
#endif

#ifdef CPU_EMUZ80
void cal_emuz80_init(cal_cpu *cpu);
void cal_emuz80_reset(cal_cpu cpu);
void cal_emuz80_run(cal_cpu cpu);
void cal_emuz80_runfor(cal_cpu cpu, int cycles);
void cal_emuz80_irq(cal_cpu cpu);
void cal_emuz80_nmi(cal_cpu cpu);
void cal_emuz80_setzpage(cal_cpu cpu, void *page0);
int cal_emuz80_timeleft(cal_cpu cpu);
void cal_emuz80_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl);
void cal_emuz80_setiou(cal_cpu cpu, memread8_t ioread, memwrite8_t iowrite);
#endif

void cal_junk68k_init(cal_cpu *cpu);
void cal_junk68k_reset(cal_cpu cpu);
void cal_junk68k_run(cal_cpu cpu);
void cal_junk68k_runfor(cal_cpu cpu, int cycles);
void cal_junk68k_irq(cal_cpu cpu);
void cal_junk68k_nmi(cal_cpu cpu);
void cal_junk68k_setzpage(cal_cpu cpu, void *page0);
int cal_junk68k_timeleft(cal_cpu cpu);
void cal_junk68k_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl);
void cal_junk68k_setmmu16(cal_cpu cpu, int shift, int mask, memread16_t *rtbl, memwrite16_t *wtbl);

cal_cpu cal_create(int type)
{
    cal_cpu retval;

    retval = calloc(1, sizeof(struct cal_cpu));

    if (retval) {
	retval->type = type;
	switch (type) {
#ifdef CPU_MARAT6502
	case CPUT_MARAT6502:
	    cal_marat6502_init(&retval);
	    break;
#endif
#ifdef CPU_EMU6502
	case CPUT_EMU6502:
	    cal_emu6502_init(&retval);
	    break;
#endif
#ifdef CPU_EMU6502A
	case CPUT_EMU6502A:
	    cal_emu6502a_init(&retval);
	    break;
#endif
#ifdef CPU_EMU65C02
	case CPUT_EMU65C02:
	    cal_emu65c02_init(&retval);
	    break;
#endif
#ifdef CPU_EMU6280
	case CPUT_EMU6280:
	    cal_emu6280_init(&retval);
	    break;
#endif
#ifdef CPU_MARATZ80
	case CPUT_MARATZ80:
	    cal_maratz80_init(&retval);
	    break;
#endif
#ifdef CPU_EMUZ80
	case CPUT_EMUZ80:
	    cal_emuz80_init(&retval);
	    break;
#endif
	case CPUT_JUNK68K:
	    cal_junk68k_init(&retval);
	    break;
	    
	default:
	    printf("Unknown CPU type.\n");
	    free(retval);
	    retval = NULL;
	}
    } else {
	printf("Insufficient memory to create CPU.\n");
    }

    return retval;
}

#ifdef CPU_MARAT6502

void cal_marat6502_init(cal_cpu *cpu)
{
    (*cpu)->data.d_marat6502 = calloc(1, sizeof(M6502));
    if (!(*cpu)->data.d_marat6502) {
	printf("Insufficient memory to create CPU.\n");
	free(*cpu);
	*cpu = NULL;
	return;
    }
    (*cpu)->reset = cal_marat6502_reset;
    (*cpu)->run = cal_marat6502_run;
    (*cpu)->runfor = cal_marat6502_runfor;
    (*cpu)->irq = cal_marat6502_irq;
    (*cpu)->nmi = cal_marat6502_nmi;
    (*cpu)->setzpage = cal_marat6502_setzpage;
    (*cpu)->timeleft = cal_marat6502_timeleft;
    (*cpu)->setmmu8 = cal_marat6502_setmmu8;
    (*cpu)->setiou = cal_marat6502_setiou;
    (*cpu)->data.d_marat6502->User = *cpu;
}

void cal_marat6502_reset(cal_cpu cpu)
{
    Reset6502(cpu->data.d_marat6502);
}

void cal_marat6502_run(cal_cpu cpu)
{
    Run6502(cpu->data.d_marat6502);
}

void cal_marat6502_runfor(cal_cpu cpu, int cycles)
{
    cpu->data.d_marat6502->IPeriod = cycles;
}

void cal_marat6502_irq(cal_cpu cpu)
{
    cpu->data.d_marat6502->IRequest = INT_IRQ;
}

void cal_marat6502_nmi(cal_cpu cpu)
{
    cpu->data.d_marat6502->IRequest = INT_NMI;
}

void cal_marat6502_setzpage(cal_cpu cpu, void *page0)
{
}

int cal_marat6502_timeleft(cal_cpu cpu)
{
    return cpu->data.d_marat6502->ICount;
}

void cal_marat6502_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl)
{
    cpu->data.d_marat6502->memshift = shift;
    cpu->data.d_marat6502->memmask = mask;
    cpu->data.d_marat6502->readtable = rtbl;
    cpu->data.d_marat6502->writetable = wtbl;
}

void cal_marat6502_setiou(cal_cpu cpu, memread8_t ioread, memwrite8_t iowrite)
{
}

#endif


#ifdef CPU_EMU6502

void cal_emu6502_init(cal_cpu *cpu)
{
    (*cpu)->data.d_emu6502 = calloc(1, sizeof(struct emu6502_context));
    if (!(*cpu)->data.d_emu6502) {
	printf("Insufficient memory to create CPU.\n");
	free(*cpu);
	*cpu = NULL;
	return;
    }
    (*cpu)->reset = cal_emu6502_reset;
    (*cpu)->run = cal_emu6502_run;
    (*cpu)->runfor = cal_emu6502_runfor;
    (*cpu)->irq = cal_emu6502_irq;
    (*cpu)->nmi = cal_emu6502_nmi;
    (*cpu)->setzpage = cal_emu6502_setzpage;
    (*cpu)->timeleft = cal_emu6502_timeleft;
    (*cpu)->setmmu8 = cal_emu6502_setmmu8;
    (*cpu)->setiou = cal_emu6502_setiou;
    (*cpu)->data.d_emu6502->user = *cpu;
    init_flag_tables();
    emu6502_init((*cpu)->data.d_emu6502);
}

void cal_emu6502_reset(cal_cpu cpu)
{
    emu6502_reset(cpu->data.d_emu6502);
}

void cal_emu6502_run(cal_cpu cpu)
{
    emu6502_run(cpu->data.d_emu6502);
}

void cal_emu6502_runfor(cal_cpu cpu, int cycles)
{
    cpu->data.d_emu6502->cycles_left += cycles;
}

void cal_emu6502_irq(cal_cpu cpu)
{
    emu6502_IRQ(cpu->data.d_emu6502);
}

void cal_emu6502_nmi(cal_cpu cpu)
{
    emu6502_NMI(cpu->data.d_emu6502);
}

void cal_emu6502_setzpage(cal_cpu cpu, void *page0)
{
    cpu->data.d_emu6502->page_0 = page0;
}

int cal_emu6502_timeleft(cal_cpu cpu)
{
    return cpu->data.d_emu6502->cycles_left;
}

void cal_emu6502_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl)
{
    cpu->data.d_emu6502->mmu_shift = shift;
    cpu->data.d_emu6502->mmu_mask = mask;
    cpu->data.d_emu6502->readtable = rtbl;
    cpu->data.d_emu6502->writetable = wtbl;
}

void cal_emu6502_setiou(cal_cpu cpu, memread8_t ioread, memwrite8_t iowrite)
{
}

#endif


#ifdef CPU_EMU6502A

void cal_emu6502a_init(cal_cpu *cpu)
{
    (*cpu)->data.d_emu6502a = calloc(1, sizeof(struct emu6502_context));
    if (!(*cpu)->data.d_emu6502a) {
	printf("Insufficient memory to create CPU.\n");
	free(*cpu);
	*cpu = NULL;
	return;
    }
    (*cpu)->reset = cal_emu6502a_reset;
    (*cpu)->run = cal_emu6502_run;
    (*cpu)->runfor = cal_emu6502_runfor;
    (*cpu)->irq = cal_emu6502_irq;
    (*cpu)->nmi = cal_emu6502_nmi;
    (*cpu)->setzpage = cal_emu6502_setzpage;
    (*cpu)->timeleft = cal_emu6502_timeleft;
    (*cpu)->setmmu8 = cal_emu6502_setmmu8;
    (*cpu)->setiou = cal_emu6502_setiou;
    (*cpu)->data.d_emu6502a->user = *cpu;
    init_flag_tables();
    emu6502a_init((*cpu)->data.d_emu6502a);
}

void cal_emu6502a_reset(cal_cpu cpu)
{
    emu6502a_reset(cpu->data.d_emu6502a);
}

#endif


#ifdef CPU_EMU65C02

void cal_emu65c02_init(cal_cpu *cpu)
{
    (*cpu)->data.d_emu65c02 = calloc(1, sizeof(struct emu6502_context));
    if (!(*cpu)->data.d_emu65c02) {
	printf("Insufficient memory to create CPU.\n");
	free(*cpu);
	*cpu = NULL;
	return;
    }
    (*cpu)->reset = cal_emu65c02_reset;
    (*cpu)->run = cal_emu6502_run;
    (*cpu)->runfor = cal_emu6502_runfor;
    (*cpu)->irq = cal_emu6502_irq;
    (*cpu)->nmi = cal_emu6502_nmi;
    (*cpu)->setzpage = cal_emu6502_setzpage;
    (*cpu)->timeleft = cal_emu6502_timeleft;
    (*cpu)->setmmu8 = cal_emu6502_setmmu8;
    (*cpu)->setiou = cal_emu6502_setiou;
    (*cpu)->data.d_emu65c02->user = *cpu;
    init_flag_tables();
    emu65c02_init((*cpu)->data.d_emu65c02);
}

void cal_emu65c02_reset(cal_cpu cpu)
{
    emu65c02_reset(cpu->data.d_emu65c02);
}

#endif


#ifdef CPU_EMU6280

void cal_emu6280_init(cal_cpu *cpu)
{
    (*cpu)->data.d_emu6280 = calloc(1, sizeof(struct emu6280_context));
    if (!(*cpu)->data.d_emu6280) {
	printf("Insufficient memory to create CPU.\n");
	free(*cpu);
	*cpu = NULL;
	return;
    }
    (*cpu)->reset = cal_emu6280_reset;
    (*cpu)->run = cal_emu6502_run;
    (*cpu)->runfor = cal_emu6502_runfor;
    (*cpu)->irq = cal_emu6280_irq;
    (*cpu)->nmi = cal_emu6502_nmi;
    (*cpu)->setzpage = cal_emu6502_setzpage;
    (*cpu)->timeleft = cal_emu6502_timeleft;
    (*cpu)->setmmu8 = cal_emu6502_setmmu8;
    (*cpu)->setiou = cal_emu6502_setiou;
    (*cpu)->data.d_emu6280->user = *cpu;
    init_flag_tables();
    emu6280_init((*cpu)->data.d_emu6280);
}

void cal_emu6280_reset(cal_cpu cpu)
{
    emu6280_reset(cpu->data.d_emu6280);
}

void cal_emu6280_irq(cal_cpu cpu)
{
    emu6280_IRQ(cpu->data.d_emu6280);
}

#endif


#ifdef CPU_MARATZ80

void cal_maratz80_init(cal_cpu *cpu)
{
    (*cpu)->data.d_maratz80 = calloc(1, sizeof(Z80));
    if (!(*cpu)->data.d_maratz80) {
	printf("Insufficient memory to create CPU.\n");
	free(*cpu);
	*cpu = NULL;
	return;
    }
    (*cpu)->reset = cal_maratz80_reset;
    (*cpu)->run = cal_maratz80_run;
    (*cpu)->runfor = cal_maratz80_runfor;
    (*cpu)->irq = cal_maratz80_irq;
    (*cpu)->nmi = cal_maratz80_nmi;
    (*cpu)->setzpage = cal_maratz80_setzpage;
    (*cpu)->timeleft = cal_maratz80_timeleft;
    (*cpu)->setmmu8 = cal_maratz80_setmmu8;
    (*cpu)->setiou = cal_maratz80_setiou;
    (*cpu)->data.d_maratz80->User = *cpu;
}

void cal_maratz80_reset(cal_cpu cpu)
{
    ResetZ80(cpu->data.d_maratz80);
}

void cal_maratz80_run(cal_cpu cpu)
{
    RunZ80(cpu->data.d_maratz80);
}

void cal_maratz80_runfor(cal_cpu cpu, int cycles)
{
    cpu->data.d_maratz80->IPeriod = cycles;
}

void cal_maratz80_irq(cal_cpu cpu)
{
    cpu->data.d_maratz80->IRequest = INT_IRQ;
}

void cal_maratz80_nmi(cal_cpu cpu)
{
    cpu->data.d_maratz80->IRequest = INT_NMI;
}

void cal_maratz80_setzpage(cal_cpu cpu, void *page0)
{
}

int cal_maratz80_timeleft(cal_cpu cpu)
{
    return cpu->data.d_maratz80->ICount;
}

void cal_maratz80_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl)
{
/*     cpu->data.d_maratz80->memshift = shift; */
/*     cpu->data.d_maratz80->memmask = mask; */
/*     cpu->data.d_maratz80->readtable = rtbl; */
/*     cpu->data.d_maratz80->writetable = wtbl; */
    cpu->data.d_maratz80->readfunc = *rtbl;
    cpu->data.d_maratz80->writefunc = *wtbl;
}

void cal_maratz80_setiou(cal_cpu cpu, memread8_t ioread, memwrite8_t iowrite)
{
    cpu->data.d_maratz80->io_read = ioread;
    cpu->data.d_maratz80->io_write = iowrite;
}

#endif


#ifdef CPU_EMUZ80

void cal_emuz80_init(cal_cpu *cpu)
{
    (*cpu)->data.d_emuz80 = calloc(1, sizeof(struct emuz80_context));
    if (!(*cpu)->data.d_emuz80) {
	printf("Insufficient memory to create CPU.\n");
	free(*cpu);
	*cpu = NULL;
	return;
    }
    (*cpu)->reset = cal_emuz80_reset;
    (*cpu)->run = cal_emuz80_run;
    (*cpu)->runfor = cal_emuz80_runfor;
    (*cpu)->irq = cal_emuz80_irq;
    (*cpu)->nmi = cal_emuz80_nmi;
    (*cpu)->setzpage = cal_emuz80_setzpage;
    (*cpu)->timeleft = cal_emuz80_timeleft;
    (*cpu)->setmmu8 = cal_emuz80_setmmu8;
    (*cpu)->setiou = cal_emuz80_setiou;
    (*cpu)->data.d_emuz80->user = *cpu;
    emuz80_init((*cpu)->data.d_emuz80);
}

void cal_emuz80_reset(cal_cpu cpu)
{
    emuz80_reset(cpu->data.d_emuz80);
}

void cal_emuz80_run(cal_cpu cpu)
{
    emuz80_run(cpu->data.d_emuz80);
}

void cal_emuz80_runfor(cal_cpu cpu, int cycles)
{
    cpu->data.d_emuz80->cycles_left += cycles;
}

void cal_emuz80_irq(cal_cpu cpu)
{
    emuz80_IRQ(cpu->data.d_emuz80);
}

void cal_emuz80_nmi(cal_cpu cpu)
{
    emuz80_NMI(cpu->data.d_emuz80);
}

void cal_emuz80_setzpage(cal_cpu cpu, void *page0)
{
}

int cal_emuz80_timeleft(cal_cpu cpu)
{
    return cpu->data.d_emuz80->cycles_left;
}

void cal_emuz80_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl)
{
    cpu->data.d_emuz80->mmu_shift = shift;
    cpu->data.d_emuz80->mmu_mask = mask;
    cpu->data.d_emuz80->readtable = rtbl;
    cpu->data.d_emuz80->writetable = wtbl;
}

void cal_emuz80_setiou(cal_cpu cpu, memread8_t ioread, memwrite8_t iowrite)
{
    cpu->data.d_emuz80->io_read = ioread;
    cpu->data.d_emuz80->io_write = iowrite;
}

#endif

void cal_junk68k_init(cal_cpu *cpu)
{
    (*cpu)->data.d_junk68k = calloc(1, sizeof(struct emu68k_context));
    if (!(*cpu)->data.d_junk68k) {
	printf("Insufficient memory to create CPU.\n");
	free(*cpu);
	*cpu = NULL;
	return;
    }
    (*cpu)->reset = cal_junk68k_reset;
    (*cpu)->run = cal_junk68k_run;
    (*cpu)->runfor = cal_junk68k_runfor;
    (*cpu)->irq = cal_junk68k_irq;
    (*cpu)->nmi = cal_junk68k_nmi;
    (*cpu)->setzpage = cal_junk68k_setzpage;
    (*cpu)->timeleft = cal_junk68k_timeleft;
    (*cpu)->setmmu8 = cal_junk68k_setmmu8;
    (*cpu)->setmmu16 = cal_junk68k_setmmu16;
    (*cpu)->data.d_junk68k->cpu = *cpu;
/*     emuz80_init((*cpu)->data.d_junk68k); */
}

void cal_junk68k_reset(cal_cpu cpu)
{
    emu68k_reset(cpu->data.d_junk68k);
}

void cal_junk68k_run(cal_cpu cpu)
{
    emu68k_step(cpu->data.d_junk68k);
}

void cal_junk68k_runfor(cal_cpu cpu, int cycles)
{
/*     cpu->data.d_emuz80->cycles_left += cycles; */
}

void cal_junk68k_irq(cal_cpu cpu)
{
/*     emuz80_IRQ(cpu->data.d_emuz80); */
}

void cal_junk68k_nmi(cal_cpu cpu)
{
/*     emuz80_NMI(cpu->data.d_emuz80); */
}

void cal_junk68k_setzpage(cal_cpu cpu, void *page0)
{
}

int cal_junk68k_timeleft(cal_cpu cpu)
{
/*     return cpu->data.d_emuz80->cycles_left; */
    return 0;
}

void cal_junk68k_setmmu8(cal_cpu cpu, int shift, int mask, memread8_t *rtbl, memwrite8_t *wtbl)
{
    cpu->data.d_junk68k->memshift = shift;
    cpu->data.d_junk68k->memmask = mask;
    cpu->data.d_junk68k->read8table = rtbl;
    cpu->data.d_junk68k->write8table = wtbl;
}

void cal_junk68k_setmmu16(cal_cpu cpu, int shift, int mask, memread16_t *rtbl, memwrite16_t *wtbl)
{
    cpu->data.d_junk68k->read16table = rtbl;
    cpu->data.d_junk68k->write16table = wtbl;
}

/*
 * $Log: cal.c,v $
 * Revision 1.12  1999/06/06 17:50:22  nyef
 * changed maratz80 memory access interface to be faster
 *
 * Revision 1.11  1999/04/17 20:06:10  nyef
 * added 68k support
 *
 * Revision 1.10  1999/01/24 15:58:33  nyef
 * changed all malloc(x) to calloc(1, x).
 *
 * Revision 1.9  1999/01/17 02:10:27  nyef
 * fixed support for irqs in the emu6280 core
 *
 * Revision 1.8  1999/01/13 03:52:23  nyef
 * continued preliminary work on 6280 support.
 *
 * Revision 1.7  1999/01/03 02:25:32  nyef
 * added maratz80 support.
 *
 * Revision 1.6  1999/01/02 06:37:43  nyef
 * fixed stupid mistake with z80 setiou.
 *
 * Revision 1.5  1999/01/02 06:35:40  nyef
 * added io interface for z80 support.
 *
 * Revision 1.4  1999/01/01 07:00:59  nyef
 * added emuz80 support
 *
 * Revision 1.3  1998/12/12 21:34:37  nyef
 * brought remaining cpu interfaces up to spec
 *
 * Revision 1.2  1998/12/12 20:34:22  nyef
 * brought marat6502 interface up to spec.
 *
 * Revision 1.1  1998/11/29 18:28:44  nyef
 * Initial revision
 *
 */
