/*
 * dt6502.c
 *
 * debug tables for 6502 emulation
 */

/* $Id: dt6502.c,v 1.2 1998/07/14 01:45:35 nyef Exp $ */

#include <stdio.h>
#include "debug.h"
#include "dt6502.h"

static char instrs[256][4] = {
    "BRK", "ORA", "???", "???",   "???", "ORA", "ASL", "???",   /* 00-07 */
    "PHP", "ORA", "ASL", "???",   "???", "ORA", "ASL", "???",   /* 08-0F */
    "BPL", "ORA", "???", "???",   "???", "ORA", "ASL", "???",   /* 10-17 */
    "CLC", "ORA", "???", "???",   "???", "ORA", "ASL", "???",   /* 18-1F */
    "JSR", "AND", "???", "???",   "BIT", "AND", "ROL", "???",   /* 20-27 */
    "PLP", "AND", "ROL", "???",   "BIT", "AND", "ROL", "???",   /* 28-2F */
    "BMI", "AND", "???", "???",   "???", "AND", "ROL", "???",   /* 30-37 */
    "SEC", "AND", "???", "???",   "???", "AND", "ROL", "???",   /* 38-3F */
    "RTI", "EOR", "???", "???",   "???", "EOR", "LSR", "???",   /* 40-47 */
    "PHA", "EOR", "LSR", "???",   "JMP", "EOR", "LSR", "???",   /* 48-4F */
    "BVC", "EOR", "???", "???",   "???", "EOR", "LSR", "???",   /* 50-57 */
    "CLI", "EOR", "???", "???",   "???", "EOR", "LSR", "???",   /* 58-5F */
    "RTS", "ADC", "???", "???",   "???", "ADC", "ROR", "???",   /* 60-67 */
    "PLA", "ADC", "ROR", "???",   "JMP", "ADC", "ROR", "???",   /* 68-6F */
    "BVS", "ADC", "???", "???",   "???", "ADC", "ROR", "???",   /* 70-77 */
    "SEI", "ADC", "???", "???",   "???", "ADC", "ROR", "???",   /* 78-7F */
    "???", "STA", "???", "???",   "STY", "STA", "STX", "???",   /* 80-87 */
    "DEY", "???", "TXA", "???",   "STY", "STA", "STX", "???",   /* 88-8F */
    "BCC", "STA", "???", "???",   "STY", "STA", "STX", "???",   /* 90-97 */
    "TYA", "STA", "TXS", "???",   "???", "STA", "???", "???",   /* 98-9F */
    "LDY", "LDA", "LDX", "???",   "LDY", "LDA", "LDX", "???",   /* A0-A7 */
    "TAY", "LDA", "TAX", "???",   "LDY", "LDA", "LDX", "???",   /* A8-AF */
    "BCS", "LDA", "???", "???",   "LDY", "LDA", "LDX", "???",   /* B0-B7 */
    "CLV", "LDA", "TSX", "???",   "LDY", "LDA", "LDX", "???",   /* B8-BF */
    "CPY", "CMP", "???", "???",   "CPY", "CMP", "DEC", "???",   /* C0-C7 */
    "INY", "CMP", "DEX", "???",   "CPY", "CMP", "DEC", "???",   /* C8-CF */
    "BNE", "CMP", "???", "???",   "???", "CMP", "DEC", "???",   /* D0-D7 */
    "CLD", "CMP", "???", "???",   "???", "CMP", "DEC", "???",   /* D8-DF */
    "CPX", "SBC", "???", "???",   "CPX", "SBC", "INC", "???",   /* E0-E7 */
    "INX", "SBC", "NOP", "???",   "CPX", "SBC", "INC", "???",   /* E8-EF */
    "BEQ", "SBC", "???", "???",   "???", "SBC", "INC", "???",   /* F0-F7 */
    "SED", "SBC", "???", "???",   "???", "SBC", "INC", "???",   /* F8-FF */
};

static int addmodes[256] = {
    10,  7, 10, 10, 10,  1,  1, 10,   /* 00-07 */
    10,  0,  2, 10, 10,  4,  4, 10,   /* 08-0F */
    11,  8, 10, 10, 10,  3,  3, 10,   /* 10-17 */
    10,  6, 10, 10, 10,  5,  5, 10,   /* 18-1F */
     4,  7, 10, 10,  1,  1,  1, 10,   /* 20-27 */
    10,  0,  2, 10,  4,  4,  4, 10,   /* 28-2F */
    11,  8, 10, 10, 10,  3,  3, 10,   /* 30-37 */
    10,  6, 10, 10, 10,  5,  5, 10,   /* 38-3F */
    10,  7, 10, 10, 10,  1,  1, 10,   /* 40-47 */
    10,  0,  2, 10,  4,  4,  4, 10,   /* 48-4F */
    11,  8, 10, 10, 10,  3,  3, 10,   /* 50-57 */
    10,  6, 10, 10, 10,  5,  5, 10,   /* 58-5F */
    10,  7, 10, 10, 10,  1,  1, 10,   /* 60-67 */
    10,  0,  2, 10,  9,  4,  4, 10,   /* 68-6F */
    11,  8, 10, 10, 10,  3,  3, 10,   /* 70-77 */
    10,  6, 10, 10, 10,  5,  5, 10,   /* 78-7F */
    10,  7, 10, 10,  1,  1,  1, 10,   /* 80-87 */
    10, 10, 10, 10,  4,  4,  4, 10,   /* 88-8F */
    11,  8, 10, 10,  3,  3, 12, 10,   /* 90-97 */
    10,  6, 10, 10, 10,  5, 10, 10,   /* 98-9F */
     0,  7,  0, 10,  1,  1,  1, 10,   /* A0-A7 */
    10,  0, 10, 10,  4,  4,  4, 10,   /* A8-AF */
    11,  8, 10, 10,  3,  3, 12, 10,   /* B0-B7 */
    10,  6, 10, 10,  5,  5,  6, 10,   /* B8-BF */
     0,  7, 10, 10,  1,  1,  1, 10,   /* C0-C7 */
    10,  1, 10, 10,  4,  4,  4, 10,   /* C8-CF */
    11,  8, 10, 10, 10,  3,  3, 10,   /* D0-D7 */
    10,  6, 10, 10, 10,  5,  5, 10,   /* D8-DF */
     0,  7, 10, 10,  1,  1,  1, 10,   /* E0-E7 */
    10,  0, 10, 10,  4,  4,  4, 10,   /* E8-EF */
    11,  8, 10, 10, 10,  3,  3, 10,   /* F0-F7 */
    10,  6, 10, 10, 10,  5,  5, 10,   /* F8-FF */
};

static char *addmodestrs[13] = {
    "%s #$%02x",          /* 0: Immediate */
    "%s $%02x",           /* 1: Zero Page */
    "%s a",               /* 2: Accumulator */
    "%s $%02x, X",        /* 3: Zero page, X */
    "%s $%04x",           /* 4: Absolute */
    "%s $%04x, X",        /* 5: Absolute, X */
    "%s $%04x, Y",        /* 6: Absolute, Y */
    "%s ($%02x, X)",      /* 7: (Indirect, X) */
    "%s ($%02x), Y",      /* 8: (Indirect), Y */
    "%s ($%04x)",         /* 9: Indirect */
    "%s",                 /* 10: Implied */
    "%s $%04x",           /* 11: Relative */
    "%s $%02x, Y",        /* 12: Zero Page, Y */
};

static int instr_lengths[13] = {2, 2, 1, 2, 3, 3, 3, 2, 2, 3, 1, 2, 2};

static int disassemble_instr(char *buf, int address, UCHAR d1, UCHAR d2, UCHAR d3)
{
    switch(addmodes[d1]) {
    case 2: case 10:
        sprintf(buf, addmodestrs[addmodes[d1]], instrs[d1]);
        break;

    case 0: case 1: case 3: case 7: case 8: case 12:
        sprintf(buf, addmodestrs[addmodes[d1]], instrs[d1], d2);
        break;

    case 4: case 5: case 6: case 9:
        sprintf(buf, addmodestrs[addmodes[d1]], instrs[d1], ((d3<<8)+d2));
        break;

    case 11:
        sprintf(buf, addmodestrs[addmodes[d1]], instrs[d1], address+2+((signed char)d2));
        break;

    default:
        sprintf(buf, "<program error>");
        printf("tool: fault in disassemble_instr()\n");
    }
    return (instr_lengths[addmodes[d1]]);
}

int dt6502_disassemble(int start_addr, int end_addr) {
    int cur_addr;
    int count;
    int len;
    unsigned char d1, d2, d3;
    char buf[128];
  
    cur_addr = start_addr;
    count = 0;
  
    while (cur_addr <= end_addr) {
	d1 = mem_read_byte(cur_addr);
	d2 = mem_read_byte(cur_addr+1);
	d3 = mem_read_byte(cur_addr+2);
    
	len = disassemble_instr(buf, cur_addr, d1, d2, d3);
	if (len == 1) {
	    printf("%04hx: %02x          %s\n", cur_addr, d1, buf);
	} else if (len == 2) {
	    printf("%04hx: %02x %02x       %s\n", cur_addr, d1, d2, buf);
	} else if (len == 3) {
	    printf("%04hx: %02x %02x %02x    %s\n", cur_addr, d1, d2, d3, buf);
	} else {
	    printf("tool: fault in disassemble()\n");
	}
	cur_addr += len;
	count += len;
    }
  
    return count;
}

/*
 * $Log: dt6502.c,v $
 * Revision 1.2  1998/07/14 01:45:35  nyef
 * made all internal data static
 * made disassemble_instr() static
 * moved disassemble() from debug.c as dt6502_disassemble()
 *
 * Revision 1.1  1998/07/11 22:16:35  nyef
 * Initial revision
 *
 */
