/*
 * emu6280.S
 *
 * 6280 emulator core
 */

/* $Id: emu6280.S,v 1.9 1999/01/31 21:11:44 nyef Exp $ */

#include "opc6280.h"
	
/* emulation register map */
/* %edi = effective address */
/* %esi = memaccess scratch */
/* %edx = PC */
/* %ch = P */
/* %ax = opcode scratch */


/* debugger interface routines */
			
/* emulate one instruction */
	.globl SYMNAME(emu6280_step)
SYMLABEL(emu6280_step)
	pushl %ebp
	movl 0x08(%esp), %ebp

	pushal

        movb reg_S, %al
        movl $0x100, reg_S
        movb %al, reg_S

	movzwl reg_PC, %edx
	movb reg_P, %ch

	movl $0f, dispatch
	jmp run_emu6502
		
0:	movw %dx, reg_PC
	movb %ch, reg_P

	popal
	
	popl %ebp
	ret

/* run emulator continuously */
	.globl SYMNAME(emu6280_run)
SYMLABEL(emu6280_run)
	pushl %ebp
	movl 0x08(%esp), %ebp
	
	pushal		

        movb reg_S, %al
        movl $0x100, reg_S
        movb %al, reg_S
	
	movzwl reg_PC, %edx
	movb reg_P, %ch

	movl $run_emu6502, dispatch
	jmp run_emu6502

/* reset CPU */
	.globl SYMNAME(emu6280_reset)
SYMLABEL(emu6280_reset)
	pushl %ebp
	movl 0x08(%esp), %ebp

	movl $opc6280_map, decodemap

	movl $0xfffe, %edi
	MemReadW(Norm)
	movw %ax, reg_PC

	popl %ebp
	ret

/* simulate an NMI */
/* WARNING: Do not call while in emu6502_step() or emu6502_run() */
	.globl SYMNAME(emu6280_NMI)
SYMLABEL(emu6280_NMI)
	pushl %ebp
	movl 0x08(%esp), %ebp

	pushal

        movb reg_S, %al
        movl $0x100, reg_S
        movb %al, reg_S
	
	mov reg_PC, %dx
/* 	decw %dx */
        push(%dh)
	push(%dl)

	xorl %eax, %eax
        movb reg_P, %al
	movl $SYMNAME(flags_to_6502), %esi
	movb (%eax, %esi), %al
	push(%al)
	orb flag_I, reg_P

	movl $0xfffa, %edi
	MemReadW(Norm)
	movw %ax, reg_PC

	popal

	popl %ebp
	ret

/* simulate an IRQ */
/* WARNING: Do not call while in emu6502_step() or emu6502_run() */
	.globl SYMNAME(emu6280_IRQ)
SYMLABEL(emu6280_IRQ)
	pushl %ebp
	movl 0x08(%esp), %ebp

	pushal

        movb reg_S, %al
        movl $0x100, reg_S
        movb %al, reg_S
	
	mov reg_PC, %dx
/* 	decw %dx */
        push(%dh)
	push(%dl)

	xorl %eax, %eax
        movb reg_P, %al
	movl $SYMNAME(flags_to_6502), %esi
	movb (%eax, %esi), %al
	push(%al)
	orb flag_I, reg_P
	
	movl $0xfff8, %edi
	MemReadW(Norm)
	movw %ax, reg_PC

	popal
	
	popl %ebp
	ret

/* initialize all emulator tables */
	.globl SYMNAME(emu6280_init)
SYMLABEL(emu6280_init)
	ret

/*
 * $Log: emu6280.S,v $
 * Revision 1.9  1999/01/31 21:11:44  nyef
 * fixed NMI and IRQ handlers
 *
 * Revision 1.8  1999/01/17 02:09:57  nyef
 * fixed IRQ handler
 *
 * Revision 1.7  1998/11/01 23:29:51  nyef
 * fixed all ZPage addressing modes to use direct ZPage access.
 *
 * Revision 1.6  1998/07/31 01:42:18  nyef
 * removed a bogus SYMNAME around a reference to opc6280_map.
 *
 * Revision 1.5  1998/07/31 01:34:35  nyef
 * added a number of SYMNAME/SYMLABELs for DJGPP compatability.
 * also changed emu6280_NMI to use MemReadW instead of bank_E.
 *
 * Revision 1.4  1998/07/14 02:12:00  nyef
 * changed emu6280_reset to set decodemap
 *
 * Revision 1.3  1998/07/13 01:49:47  nyef
 * fixed emu6280_reset to use correct reset vector
 *
 * Revision 1.2  1998/07/12 20:50:04  nyef
 * changed to use emu6502 main loop.
 *
 * fixed all routines to use context properly and fix stack.
 *
 * Revision 1.1  1998/07/11 22:16:57  nyef
 * Initial revision
 *
 */
