/*
 * emu6502.S
 *
 * 6502 emulator core
 */

/* $Id: emu6502.S,v 1.7 1999/06/14 16:50:32 nyef Exp $ */

#include "opc6502.h"
	

/* cycle_temp is a workspace for calclating page boundary penalties */
	.data
	.globl cycle_temp
cycle_temp:
	.long 0
	.text

/* emulation register map */
/* %edi = effective address */
/* %esi = memaccess scratch */
/* %ebp = context pointer */
/* %edx = PC */
/* %ch = P */
/* %ax = opcode scratch */


/* debugger interface routines */
			
/* emulate one instruction */
	.globl SYMNAME(emu6502_step)
SYMLABEL(emu6502_step)
	pushl %ebp

	movl 0x08(%esp), %ebp
	
	pushal

	movb reg_S, %al
	movl $0x100, reg_S
	movb %al, reg_S
	
	movzwl reg_PC, %edx
	movb reg_P, %ch

	movl $0f, dispatch
	jmp run_emu6502
		
0:	movw %dx, reg_PC
	movb %ch, reg_P

	popal
	
	popl %ebp
	ret

/* run emulator continuously */
	.globl SYMNAME(emu6502_run)
SYMLABEL(emu6502_run)
	pushl %ebp
	
	movl 0x08(%esp), %ebp
	
	pushal		
	
	movb reg_S, %al
	movl $0x100, reg_S
	movb %al, reg_S
	
	movzwl reg_PC, %edx
	movb reg_P, %ch

	movl $run_emu6502, dispatch
	jmp run_emu6502

/* reset CPU */
	.globl SYMNAME(emu6502_reset)
SYMLABEL(emu6502_reset)
	pushl %ebp
	movl 0x08(%esp), %ebp

	movl $SYMNAME(opc6502_map), decodemap	
	movl $0xfffc, %edi
	MemReadW(Norm)
	movw %ax, reg_PC
	
	popl %ebp
	ret

/* simulate an NMI */
/* WARNING: Do not call while in emu6502_step() or emu6502_run() */
	.globl SYMNAME(emu6502_NMI)
SYMLABEL(emu6502_NMI)
	pushl %ebp
	movl 0x08(%esp), %ebp
	pushal
	
	movb reg_S, %al
	movl $0x100, reg_S
	movb %al, reg_S
	
	mov reg_PC, %dx
        push(%dh)
	push(%dl)

	xorl %eax, %eax
        movb reg_P, %al
	movl $SYMNAME(flags_to_6502), %esi
	movb (%eax, %esi, 1), %al
	push(%al)
	orb flag_I, reg_P

	movl $0xfffa, %edi
	MemReadW(Norm)
	movw %ax, reg_PC

	popal
	popl %ebp
	ret

/* simulate an IRQ */
/* WARNING: Do not call while in emu6502_step() or emu6502_run() */
	.globl SYMNAME(emu6502_IRQ)
SYMLABEL(emu6502_IRQ)
	pushl %ebp
	movl 0x08(%esp), %ebp

	pushal
	
	movb reg_S, %al
	movl $0x100, reg_S
	movb %al, reg_S
	
	mov reg_PC, %dx
        push(%dh)
	push(%dl)

	xorl %eax, %eax
        movb reg_P, %al
	movl $SYMNAME(flags_to_6502), %esi
	movb (%eax, %esi, 1), %al
	push(%al)
	orb flag_I, reg_P
	
	movl $0xfffe, %edi
	MemReadW(Norm)
	movw %ax, reg_PC

	popal
	popl %ebp
	ret

/* initialize all emulator tables */
	.globl SYMNAME(emu6502_init)
SYMLABEL(emu6502_init)
	ret

/* debug stuff */
#ifdef DEBUG_CPUTRACE
	.data
	.globl debstuff
debstuff:
	.word 0
	.globl debstuff2
debstuff2:
	.word 0
	.globl debstuff3
debstuff3:
	.word 0
	.text
#endif

/* main loop */

	.globl run_emu6502
run_emu6502:
	decl cycles_left
	js 0f
#ifdef DEBUG_CPUTRACE
	movw debstuff2, %ax
	movw %ax, debstuff3
	movw debstuff, %ax
	movw %ax, debstuff2
	movw %dx, debstuff
#endif
	GetFromPC_B
	MemReadNorm
	movzbl %al, %eax
	movl decodemap, %esi
	jmp *(%esi, %eax, 4)
0:	jmp abort_instr

/*
 * $Log: emu6502.S,v $
 * Revision 1.7  1999/06/14 16:50:32  nyef
 * fixed to compile clean with GNU as 2.9.1
 *
 * Revision 1.6  1998/12/31 03:39:52  nyef
 * added some ifdefed out code to record the last two PC values.
 *
 * Revision 1.5  1998/12/25 16:09:23  nyef
 * cleaned up some abstraction violations involving the interrupt vectors.
 *
 * Revision 1.4  1998/12/19 21:04:20  nyef
 * fixed interrupt routines
 *
 * Revision 1.3  1998/11/01 23:30:18  nyef
 * fixed all ZPage addressing modes to use direct ZPage access.
 *
 * Revision 1.2  1998/07/31 01:39:28  nyef
 * added some SYMNAME/SYMLABELs for DJGPP compatability.
 *
 * Revision 1.1  1998/07/11 22:17:12  nyef
 * Initial revision
 *
 */
