/*
 * emu6502.h
 *
 * 6502 emulation defines
 */

/* $Id: emu6502.h,v 1.5 1999/04/17 20:07:17 nyef Exp $ */

#ifndef __EMU6502_H__
#define __EMU6502_H__

#include "cal.h"

#define BYTE unsigned char
#define WORD unsigned short

#define PACKED __attribute__ ((packed))

/* registers */

/* reserved memory spaces are for future expansion (e.g. 65816 emulation) */
struct emu6502_context {
    BYTE reg_A;
    BYTE reserved1;
    BYTE reg_Y;
    BYTE reserved2;
    BYTE reg_X;
    BYTE reserved3;
    BYTE reg_P;
    BYTE reg_S;
    BYTE reserved4;
    WORD reserved5;
    WORD reg_PC;
    memread8_t *readtable;
    memwrite8_t *writetable;
    void **decodetable;
    void *dispatch;
    long cycles_left;
    void *page_0; /* this should be followed contiguously by page 1 */
    long mmu_shift;
    long mmu_mask;
    void *user;
} PACKED;


/* functions */

void emu6502_step(struct emu6502_context *context);
void emu6502_run(struct emu6502_context *context);

void emu6502_reset(struct emu6502_context *context);
void emu6502_NMI(struct emu6502_context *context);
void emu6502_IRQ(struct emu6502_context *context);

void emu6502_init(struct emu6502_context *context);

#endif /* __EMU6502_H__ */

/*
 * $Log: emu6502.h,v $
 * Revision 1.5  1999/04/17 20:07:17  nyef
 * fixed for new version of the CAL.
 *
 * Revision 1.4  1998/12/21 02:00:11  nyef
 * changed readtable and writetable to be of type memread_t * and
 * memwrite_t *.
 *
 * Revision 1.3  1998/12/12 21:34:53  nyef
 * finished converting to the new memory access spec
 *
 * Revision 1.2  1998/11/29 18:26:15  nyef
 * converted to use the new memory access spec.
 *
 * Revision 1.1  1998/07/11 22:17:10  nyef
 * Initial revision
 *
 */
