/*
 * emuz80.S
 *
 * z80 emulator core
 */

/* $Id: emuz80.S,v 1.2 1999/06/14 16:50:28 nyef Exp $ */

#include "opcz80.h"


/* emulation register map */
/* %edi = effective address */
/* %esi = memaccess scratch */
/* %ebp = context pointer */
/* %edx = PC */
/* %ch = F */
/* %ax = opcode scratch */


/* debugger interface routines */
			
/* emulate one instruction */
	.globl SYMNAME(emuz80_step)
SYMLABEL(emuz80_step)
	pushl %ebp

	movl 0x08(%esp), %ebp
	
	pushal

	movzwl reg_PC, %edx
	movb reg_F, %ch

	movl $0f, dispatch
	jmp run_emuz80
		
0:	movw %dx, reg_PC
	movb %ch, reg_F

	popal
	
	popl %ebp
	ret

/* run emulator continuously */
	.globl SYMNAME(emuz80_run)
SYMLABEL(emuz80_run)
	pushl %ebp
	
	movl 0x08(%esp), %ebp
	
	pushal		
	
	movzwl reg_PC, %edx
	movb reg_F, %ch

	movl $run_emuz80, dispatch
	jmp run_emuz80

/* reset CPU */
	.globl SYMNAME(emuz80_reset)
SYMLABEL(emuz80_reset)
	pushl %ebp
	movl 0x08(%esp), %ebp

	movl $SYMNAME(opcz80_map), decodemap	
	xorl %eax, %eax
	movw %ax, reg_PC
	
	popl %ebp
	ret

/* simulate an NMI */
/* WARNING: Do not call while in emuz80_step() or emuz80_run() */
	.globl SYMNAME(emuz80_NMI)
SYMLABEL(emuz80_NMI)
	pushl %ebp
	movl 0x08(%esp), %ebp
	pushal
	
	mov reg_PC, %dx

	movl $0xfffa, %edi
	MemReadW
	movw %ax, reg_PC

	popal
	popl %ebp
	ret

/* simulate an IRQ */
/* WARNING: Do not call while in emuz80_step() or emuz80_run() */
	.globl SYMNAME(emuz80_IRQ)
SYMLABEL(emuz80_IRQ)
	pushl %ebp
	movl 0x08(%esp), %ebp

	pushal
	
	mov reg_PC, %dx

	movl $0xfffe, %edi
	MemReadW
	movw %ax, reg_PC

	popal
	popl %ebp
	ret

/* initialize all emulator tables */
	.globl SYMNAME(emuz80_init)
SYMLABEL(emuz80_init)
	ret

/* debug stuff */
#ifdef DEBUG_CPUTRACE
	.data
	.globl z80_debstuff
z80_debstuff:
	.word 0
	.globl z80_debstuff2
z80_debstuff2:
	.word 0
	.globl z80_debstuff3
z80_debstuff3:
	.word 0
	.text
#endif

/* main loop */

	.globl run_emuz80
run_emuz80:
	decl cycles_left
	js 0f
#ifdef DEBUG_CPUTRACE
	movw z80_debstuff2, %ax
	movw %ax, z80_debstuff3
	movw z80_debstuff, %ax
	movw %ax, z80_debstuff2
	movw %dx, z80_debstuff
#endif
	GetFromPC_B
	MemRead
	movzbl %al, %eax
	movl decodemap, %esi
	jmp *(%esi, %eax, 4)
0:	jmp z80_abort_instr

/*
 * $Log: emuz80.S,v $
 * Revision 1.2  1999/06/14 16:50:28  nyef
 * fixed to compile clean with GNU as 2.9.1
 *
 * Revision 1.1  1999/01/01 20:43:55  nyef
 * Initial revision
 *
 */
