/*
 * genesis.c
 *
 * Sega Genesis emulation.
 */

/* $Id: genesis.c,v 1.1 1999/04/17 22:24:47 nyef Exp $ */

#include <stdlib.h>
#include <netinet/in.h>

#include "ui.h"
#include "system.h"
#include "cal.h"
#include "tool.h"
#include "gen_vdp.h"

#define GENESIS_RAMSIZE 0x10000 /* FIXME: Get real ram size data */

unsigned char *genesis_ram;
cal_cpu genesis_68k;
cal_cpu genesis_z80;
gen_vdp genesis_vdp;

/*
 * memory map:
 *
 * 000000-3fffff ROM
 * 400000-9fffff Reserved
 * a00000-afffff Z80/Sound
 * b00000-bfffff Reserved
 * c00000-dfffff VDP
 * e00000-feffff RAM Mirrors
 * ff0000-ffffff RAM
 */

unsigned char gen_read8_rom(cal_cpu cpu, unsigned long address)
{
    return romimage[address & 0x3fffff];
}

unsigned short gen_read16_rom(cal_cpu cpu, unsigned long address)
{
    if (address & 1) {
	deb_printf("gen_read16_rom: CPU BUG: non-aligned 16-bit access.\n");
    }
    return htons(*((unsigned short *)(romimage + address)));
}

unsigned char gen_read8_reserved(cal_cpu cpu, unsigned long address)
{
    deb_printf("gen_read8_reserved: address 0x%06x.\n", address);
    return 0;
}

unsigned short gen_read16_reserved(cal_cpu cpu, unsigned long address)
{
    if (address & 1) {
	deb_printf("gen_read16_reserved: CPU BUG: non-aligned 16-bit access.\n");
    }
    deb_printf("gen_read16_reserved: address 0x%06x.\n", address);
    return 0;
}

unsigned char gen_read8_misc(cal_cpu cpu, unsigned long address)
{
    deb_printf("gen_read8_misc: 0x%06x.\n", address);
    return 0;
}

unsigned short gen_read16_misc(cal_cpu cpu, unsigned long address)
{
    if (address & 1) {
	deb_printf("gen_read16_misc: CPU BUG: non-aligned 16-bit access.\n");
    }
    deb_printf("gen_read16_misc: 0x%06x.\n", address);
    return 0;
}

unsigned char gen_read8_vdp(cal_cpu cpu, unsigned long address)
{
    deb_printf("gen_read8_vdp: 0x%06x.\n", address);
    return 0;
}

unsigned short gen_read16_vdp(cal_cpu cpu, unsigned long address)
{
    if (address & 1) {
	deb_printf("gen_read16_vdp: CPU BUG: non-aligned 16-bit access.\n");
    }
    deb_printf("gen_read16_vdp: 0x%06x.\n", address);
    return 0;
}

unsigned char gen_read8_ram(cal_cpu cpu, unsigned long address)
{
    return genesis_ram[address & 0xffff];
}

unsigned short gen_read16_ram(cal_cpu cpu, unsigned long address)
{
    if (address & 1) {
	deb_printf("gen_read16_ram: CPU BUG: non-aligned 16-bit access.\n");
    }
    return htons(*((unsigned short *)(genesis_ram + (address & 0xffff))));
}

void gen_write8_reserved(cal_cpu cpu, unsigned long address, unsigned char data)
{
    deb_printf("gen_write8_reserved: address 0x%06x, data 0x%02x.\n", address, data);
}

void gen_write16_reserved(cal_cpu cpu, unsigned long address, unsigned short data)
{
    if (address & 1) {
	deb_printf("gen_write16_reserved: CPU BUG: non-aligned 16-bit access.\n");
    }
    deb_printf("gen_write16_reserved: address 0x%06x, data 0x%04hx.\n", address, data);
}

void gen_write8_misc(cal_cpu cpu, unsigned long address, unsigned char data)
{
/*     deb_printf("gen_write8_misc: address 0x%06x, data 0x%02x.\n", address, data); */
}

void gen_write16_misc(cal_cpu cpu, unsigned long address, unsigned short data)
{
    if (address & 1) {
	deb_printf("gen_write16_misc: CPU BUG: non-aligned 16-bit access.\n");
    }
/*     deb_printf("gen_write16_misc: address 0x%06x, data 0x%04hx.\n", address, data); */
}

void gen_write8_vdp(cal_cpu cpu, unsigned long address, unsigned char data)
{
    deb_printf("gen_write8_vdp: address 0x%06x, data 0x%02x.\n", address, data);
}

void gen_write16_vdp(cal_cpu cpu, unsigned long address, unsigned short data)
{
    if (address & 1) {
	deb_printf("gen_write16_vdp: CPU BUG: non-aligned 16-bit access.\n");
    }
/*     deb_printf("gen_write16_vdp: address 0x%06x, data 0x%04hx.\n", address, data); */
    if ((address & 0xfffffc) == 0xc00000) {
	genesis_vdp->writeport0(genesis_vdp, data);
    } else if ((address & 0xfffffc) == 0xc00004) {
	genesis_vdp->writeport1(genesis_vdp, data);
/* 	deb_printf("gen_write16_vdp: CNTRL_LO, data 0x%04hx (%c%c).\n", data, (data & 0x8000)? '1': '0', (data & 0x4000)? '1': '0'); */
/*     } else if (address == 0xc00006) { */
/* 	deb_printf("gen_write16_vdp: CNTRL_HI, data 0x%04hx.\n", data); */
    } else {
	deb_printf("gen_write16_vdp: address 0x%06x, data 0x%04hx.\n", address, data);
    }
}

void gen_write8_ram(cal_cpu cpu, unsigned long address, unsigned char data)
{
    genesis_ram[address & 0xffff] = data;
}

void gen_write16_ram(cal_cpu cpu, unsigned long address, unsigned short data)
{
    if (address & 1) {
	deb_printf("gen_write16_ram: CPU BUG: non-aligned 16-bit access.\n");
    }
    (*((unsigned short *)(genesis_ram + (address & 0xffff)))) = ntohs(data);
}

memread8_t gen_read8[16] = {
    gen_read8_rom,
    gen_read8_rom,
    gen_read8_rom,
    gen_read8_rom,
    gen_read8_reserved,
    gen_read8_reserved,
    gen_read8_reserved,
    gen_read8_reserved,
    gen_read8_reserved,
    gen_read8_reserved,
    gen_read8_misc,
    gen_read8_reserved,
    gen_read8_vdp,
    gen_read8_vdp,
    gen_read8_ram,
    gen_read8_ram,
};
memread16_t gen_read16[16] = {
    gen_read16_rom,
    gen_read16_rom,
    gen_read16_rom,
    gen_read16_rom,
    gen_read16_reserved,
    gen_read16_reserved,
    gen_read16_reserved,
    gen_read16_reserved,
    gen_read16_reserved,
    gen_read16_reserved,
    gen_read16_misc,
    gen_read16_reserved,
    gen_read16_vdp,
    gen_read16_vdp,
    gen_read16_ram,
    gen_read16_ram,
};

memwrite8_t gen_write8[16] = {
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_reserved,
    gen_write8_misc,
    gen_write8_reserved,
    gen_write8_vdp,
    gen_write8_vdp,
    gen_write8_ram,
    gen_write8_ram,
};
memwrite16_t gen_write16[16] = {
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_reserved,
    gen_write16_misc,
    gen_write16_reserved,
    gen_write16_vdp,
    gen_write16_vdp,
    gen_write16_ram,
    gen_write16_ram,
};

void genesis_convert_rom(void)
{
    /* FIXME: kill the person who devised the SMD format */
    unsigned char *newrom;
    unsigned char *cur_ptr;
    int i;
    int j;

    newrom = malloc(romsize - 0x200);
    cur_ptr = newrom;
    
    j = 0x200;
    while (j < romsize) {
	for (i = 0; i < 0x2000; i++) {
	    *cur_ptr++ = romimage[j + 0x2000];
	    *cur_ptr++ = romimage[j];
	    j++;
	}
	j += 0x2000;
    }
    free(romimage);
    romimage = newrom;
}

void genesis_timeslice(cal_cpu cpu)
{
    int i;
    int j;

    for (i = 262; i && (system_flags == F_NONE); i--) {
	cpu->runfor(cpu, 228); /* FIXME: find out how many cycles/hbl */
	for (j = 1; (j < 20) && (!(system_flags & F_UNIMPL)); j++) {
	    cpu->run(cpu);
	}

	if (genesis_vdp->periodic(genesis_vdp)) {
/* 	    cpu->irq(cpu); */
	}

	if (system_flags & F_UNIMPL) {
	    deb_printf("Unimplimented Instruction.\n");
	}
    }
}

void genesis_run(int startup_deb)
{
    genesis_ram = calloc(1, GENESIS_RAMSIZE);
    
    if (genesis_ram && romimage) {
	if (romsize & 0x200) {
	    deb_printf("genesis_run: hi/lo paged rom image (SMD format).\n");
	    deb_printf("genesis_run: converting.\n");
	    genesis_convert_rom();
	} else {
	    deb_printf("genesis_run: linear rom image (BIN format).\n");
	}

	genesis_68k = cal_create(CPUT_JUNK68K);
	genesis_68k->setmmu8(genesis_68k, 20, 15, gen_read8, gen_write8);
	genesis_68k->setmmu16(genesis_68k, 20, 15, gen_read16, gen_write16);
	genesis_68k->reset(genesis_68k);

	genesis_vdp = gen_vdp_create();
	
	set_timeslice((void (*)(void *)) &genesis_timeslice, genesis_68k);
    } else {
	deb_printf("out of memory.\n");
    }
}

/*
 * $Log: genesis.c,v $
 * Revision 1.1  1999/04/17 22:24:47  nyef
 * Initial revision
 *
 */
