/*
 * mappers.c
 *
 * mapper emulation routines
 */

/* $Id: mappers.c,v 1.54 1999/07/05 01:30:02 nyef Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include "mappers.h"
#include "ui.h"
#include "tool.h"
#include "nes_ppu.h"
#include "nes.h"
#include "cal.h"

/* defines for mapper interfaces */
extern mapper_iface_t map0_iface;
extern mapper_iface_t map1_iface;
extern mapper_iface_t map2_iface;
extern mapper_iface_t map3_iface;
extern mapper_iface_t map4_iface;
extern mapper_iface_t map5_iface;
extern mapper_iface_t map6_iface;
extern mapper_iface_t map7_iface;
extern mapper_iface_t map8_iface;
extern mapper_iface_t map9_iface;
extern mapper_iface_t mapA_iface;
extern mapper_iface_t mapB_iface;
extern mapper_iface_t mapC_iface;
extern mapper_iface_t mapD_iface;
extern mapper_iface_t mapE_iface;
extern mapper_iface_t mapF_iface;
extern mapper_iface_t map14_iface;
extern mapper_iface_t map17_iface;
extern mapper_iface_t map18_iface;
extern mapper_iface_t map20_iface;
extern mapper_iface_t map42_iface;
extern mapper_iface_t map44_iface;
extern mapper_iface_t map4E_iface;

struct mapper_support {
    int mapper;
    const char *name;
    mapper_iface_t *iface;
    int support; /* mapper support. 0: none. 1: partial. 2: full. */
};

struct mapper_support mappers[] = {
    {0,   "None",               &map0_iface,  2},
    {1,   "MMC1",               &map1_iface,  2},
    {2,   "UNROM",              &map2_iface,  2},
    {3,   "CNROM",              &map3_iface,  2},
    {4,   "MMC3",               &map4_iface,  1},
    {7,   "AOROM",              &map7_iface,  2},
    {9,   "MMC2",               &map9_iface,  2},
    {10,  "MMC4",               &mapA_iface,  2},
    {11,  "Color Dreams",       &mapB_iface,  2},
    {15,  "100-in-1",           &mapF_iface,  2},
    {20,  "Famicom DiskSystem", &map14_iface, 1},
    {23,  "Konami VRC2 type B", &map17_iface, 1},
    {24,  "Konami VRC6",        &map18_iface, 1},
    {32,  "Irem G-101",         &map20_iface, 1},
    {66,  "GNROM",              &map42_iface, 1},
    {68,  "Sunsoft Mapper 4",   &map44_iface, 2},
    {78,  "Irem 74HC161/32",    &map4E_iface, 1},
    {-1,  NULL,                 NULL,         0},
};

int cur_mapper = -1;

mapinit_t cur_init;
mapwrite_t cur_write;
maphsync_t cur_hsync;
mapvsync_t cur_vsync;

int set_mapper(int mapper)
{
    int i;
    
    for (i = 0; mappers[i].mapper != -1; i++) {
	if ((mappers[i].mapper == mapper) && (mappers[i].support)) {
	    deb_printf("mappers: initializing%s completed mapper %d (%s).\n",
		   (mappers[i].support == 1)? " partially": "",
		   mapper, mappers[i].name);
	    cur_mapper = mapper;
	    cur_init = mappers[i].iface->initcode;
	    cur_write = mappers[i].iface->writecode;
	    cur_hsync = mappers[i].iface->hsynccode;
	    cur_vsync = mappers[i].iface->vsynccode;
	    return 0;
	}
    }
    return -1;
}

int get_mapper(void)
{
    return cur_mapper;
}

int check_mapper(int mapper)
{
    int i;

    for (i = 0; mappers[i].mapper != -1; i++) {
	if (mappers[i].mapper == mapper) {
	    return mappers[i].support;
	}
    }
    return 0;
}


/* dummy routines for when a mapper doesn't need it. */

void dummy_hsync(cal_cpu cpu)
{
}

void dummy_vsync(cal_cpu cpu)
{
}

/* MAPPER 0 Implementation */

void map0_write(unsigned short address, unsigned char value)
{
}

void map0_init(void)
{
    ppu_latchfunc = NULL;
    if (ppu_flags & PPUF_VROM) {
	PPU_bank_00 = &vromimage[0x0000];
	PPU_bank_04 = &vromimage[0x0400];
	PPU_bank_08 = &vromimage[0x0800];
	PPU_bank_0c = &vromimage[0x0c00];
	PPU_bank_10 = &vromimage[0x1000];
	PPU_bank_14 = &vromimage[0x1400];
	PPU_bank_18 = &vromimage[0x1800];
	PPU_bank_1c = &vromimage[0x1c00];
    } else {
	deb_printf("map0_init(): cart has no vrom. not good.\n");
    }
    if (cart_vrom_size > 1) {
	deb_printf("map0_init(): cart has more than 8k vrom. not good.\n");
    }
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t map0_iface = {map0_init, map0_write, dummy_hsync, dummy_vsync};


/* MAPPER 1 Implementation */

int map1_seq;
int map1_acc;
int map1_data[4];

void map1_write(unsigned short address, unsigned char value)
{
    int i;
    int bank_select;
    
    if (value & 0x80) {
	map1_data[0] |= 0x0c;
	map1_acc = map1_data[(address >> 13) & 3];
	map1_seq = 5;
    } else {
	map1_acc |= ((value & 1) << (map1_seq));
	map1_seq++;
    }

    if ((map1_seq) == 5) {
	map1_data[(address >> 13) & 3] = map1_acc;
	map1_seq = 0;
	map1_acc = 0;
	    
	if (!(map1_data[0] & 2)) {
	    PPU_mirror_one_low();
	} else if (map1_data[0] & 1) {
	    PPU_mirror_horizontal();
	} else {
	    PPU_mirror_vertical();
	}
	
	if (cart_rom_size == 0x20) { /* 512k cart */
	    bank_select = (map1_data[1] & 0x10) << 1;
	} else if (cart_rom_size == 0x40) { /* 1024k cart */
	    if (map1_data[0] & 0x10) {
		bank_select = ((map1_data[1] & 0x10) << 1) | ((map1_data[2] & 0x10) << 2);
	    } else {
		bank_select = (map1_data[1] & 0x10)? 0x60: 0x00;
	    }
	} else { /* other carts */
	    bank_select = 0;
	}
	
	if (!(map1_data[0] & 8)) {
	    for (i = 0xf; ((map1_data[3]) & i) >= (cart_rom_size >> 1); i >>= 1) {
		/* mask down to the rom size. */
	    }
	    bank8_reg = ((map1_data[3] & i) << 2) + bank_select;
	    bankA_reg = ((map1_data[3] & i) << 2) + bank_select + 1;
	    bankC_reg = ((map1_data[3] & i) << 2) + bank_select + 2;
	    bankE_reg = ((map1_data[3] & i) << 2) + bank_select + 3;
	} else if (map1_data[0] & 4) {
	    for (i = 0xf; ((map1_data[3]) & i) >= cart_rom_size; i >>= 1) {
		/* mask down to the rom size. */
	    }
	    bank8_reg = ((map1_data[3] & i) << 1) + bank_select;
	    bankA_reg = ((map1_data[3] & i) << 1) + bank_select + 1;
	    bankC_reg = ((cart_rom_size - 1) << 1);
	    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
	} else {
	    for (i = 0xf; ((map1_data[3]) & i) >= cart_rom_size; i >>= 1) {
		/* mask down to the rom size. */
	    }
	    bank8_reg = 0;
	    bankA_reg = 1;
	    bankC_reg = ((map1_data[3] & i) << 1) + bank_select;
	    bankE_reg = ((map1_data[3] & i) << 1) + bank_select + 1;
	}
	sync_banks();
	
	if (ppu_flags & PPUF_VROM) {
	    if (map1_data[0] & 0x10) {
		for (i = 0x1f; ((map1_data[1]) & i) >= (cart_vrom_size << 1); i >>= 1) {
		    /* mask down to the vrom size. */
		}
		PPU_bank_00 = &vromimage[(map1_data[1]&i)*0x1000+0x0000];
		PPU_bank_04 = &vromimage[(map1_data[1]&i)*0x1000+0x0400];
		PPU_bank_08 = &vromimage[(map1_data[1]&i)*0x1000+0x0800];
		PPU_bank_0c = &vromimage[(map1_data[1]&i)*0x1000+0x0c00];
		for (i = 0x1f; ((map1_data[2]) & i) >= (cart_vrom_size << 1); i >>= 1) {
		    /* mask down to the vrom size. */
		}
		PPU_bank_10 = &vromimage[(map1_data[2]&i)*0x1000+0x0000];
		PPU_bank_14 = &vromimage[(map1_data[2]&i)*0x1000+0x0400];
		PPU_bank_18 = &vromimage[(map1_data[2]&i)*0x1000+0x0800];
		PPU_bank_1c = &vromimage[(map1_data[2]&i)*0x1000+0x0c00];
	    } else {
		for (i = 0x1f; ((map1_data[1]) & i) >= (cart_vrom_size << 1); i >>= 1) {
		    /* mask down to the vrom size. */
		}
		i &= 0xfe;
		PPU_bank_00 = &vromimage[(map1_data[1]&i)*0x1000+0x0000];
		PPU_bank_04 = &vromimage[(map1_data[1]&i)*0x1000+0x0400];
		PPU_bank_08 = &vromimage[(map1_data[1]&i)*0x1000+0x0800];
		PPU_bank_0c = &vromimage[(map1_data[1]&i)*0x1000+0x0c00];
		PPU_bank_10 = &vromimage[(map1_data[1]&i)*0x1000+0x1000];
		PPU_bank_14 = &vromimage[(map1_data[1]&i)*0x1000+0x1400];
		PPU_bank_18 = &vromimage[(map1_data[1]&i)*0x1000+0x1800];
		PPU_bank_1c = &vromimage[(map1_data[1]&i)*0x1000+0x1c00];
	    }
	}
    }
}

void map1_init(void)
{
    ppu_latchfunc = NULL;
    map1_data[0] = 0x1f; /* FIXME: may be incorrect */
    map1_data[3] = 0x00;
    if (ppu_flags & PPUF_VROM) {
	/* FIXME: may want to set consistent internal state */
	PPU_bank_00 = &vromimage[0x0000];
	PPU_bank_04 = &vromimage[0x0400];
	PPU_bank_08 = &vromimage[0x0800];
	PPU_bank_0c = &vromimage[0x0c00];
	PPU_bank_10 = &vromimage[0x1000];
	PPU_bank_14 = &vromimage[0x1400];
	PPU_bank_18 = &vromimage[0x1800];
	PPU_bank_1c = &vromimage[0x1c00];
    }
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t map1_iface = {map1_init, map1_write, dummy_hsync, dummy_vsync};


/* MAPPER 2 Implementation */

void map2_write(unsigned short address, unsigned char value)
{
/*     deb_printf("map2_write(%04hx, %02x)\n", address, value); */
    if (value >= cart_rom_size) {
	int i;
	
	deb_printf("map2_write: page %02x beyond end of cartridge, fudging values.\n", value);

	i = 0x7f;
	while ((value &= i) >= cart_rom_size) {
	    i >>= 1;
	}
	deb_printf("map2_write: substituting page %02x.\n", value);
    }
    bank8_reg = (value << 1);
    bankA_reg = (value << 1) + 1;
    sync_banks();
}

void map2_init(void)
{
    ppu_latchfunc = NULL;
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t map2_iface = {map2_init, map2_write, dummy_hsync, dummy_vsync};


/* MAPPER 3 Implementation */

void map3_write(unsigned short address, unsigned char value)
{
/*     deb_printf("map3_write(0x%04hx, 0x%02x)\n", address, value); */
    if (value >= cart_vrom_size) {
	int i;
	
	deb_printf("map3_write: page beyond end of cartridge, fudging values.\n");
	
	i = 0x7f;
	while ((value &= i) >= cart_vrom_size) {
	    i >>= 1;
	}
    }

    PPU_bank_00 = &vromimage[value*0x2000+0x0000];
    PPU_bank_04 = &vromimage[value*0x2000+0x0400];
    PPU_bank_08 = &vromimage[value*0x2000+0x0800];
    PPU_bank_0c = &vromimage[value*0x2000+0x0c00];
    PPU_bank_10 = &vromimage[value*0x2000+0x1000];
    PPU_bank_14 = &vromimage[value*0x2000+0x1400];
    PPU_bank_18 = &vromimage[value*0x2000+0x1800];
    PPU_bank_1c = &vromimage[value*0x2000+0x1c00];
}

void map3_init(void)
{
    ppu_latchfunc = NULL;
    map3_write(0, 0);
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t map3_iface = {map3_init, map3_write, dummy_hsync, dummy_vsync};


/* MAPPER 4 Implementation */

unsigned char map4_cmd; /* current mapper command */
unsigned char map4_irq_count;
unsigned char map4_irq_latch;
unsigned char map4_irq_state;

unsigned char map4_vrom_switch[6];
unsigned char map4_rom_switch[2];

void map4_sync_rom(void)
{
    if (map4_cmd & 0x40) {
	bank8_reg = ((cart_rom_size - 1) << 1);
	bankA_reg = map4_rom_switch[1];
	bankC_reg = map4_rom_switch[0];
	bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    } else {
	bank8_reg = map4_rom_switch[0];
	bankA_reg = map4_rom_switch[1];
	bankC_reg = ((cart_rom_size - 1) << 1);
	bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    }
    sync_banks();
}

void map4_sync_vrom(void)
{
    if (ppu_flags & PPUF_VROM) {
	if (map4_cmd & 0x80) {
	    PPU_bank_10 = &vromimage[(map4_vrom_switch[0]*0x0400+0x0000)];
	    PPU_bank_14 = &vromimage[(map4_vrom_switch[0]*0x0400+0x0400)];
	    PPU_bank_18 = &vromimage[(map4_vrom_switch[1]*0x0400+0x0000)];
	    PPU_bank_1c = &vromimage[(map4_vrom_switch[1]*0x0400+0x0400)];
	    PPU_bank_00 = &vromimage[(map4_vrom_switch[2]*0x0400+0x0000)];
	    PPU_bank_04 = &vromimage[(map4_vrom_switch[3]*0x0400+0x0000)];
	    PPU_bank_08 = &vromimage[(map4_vrom_switch[4]*0x0400+0x0000)];
	    PPU_bank_0c = &vromimage[(map4_vrom_switch[5]*0x0400+0x0000)];
	} else {
	    PPU_bank_00 = &vromimage[(map4_vrom_switch[0]*0x0400+0x0000)];
	    PPU_bank_04 = &vromimage[(map4_vrom_switch[0]*0x0400+0x0400)];
	    PPU_bank_08 = &vromimage[(map4_vrom_switch[1]*0x0400+0x0000)];
	    PPU_bank_0c = &vromimage[(map4_vrom_switch[1]*0x0400+0x0400)];
	    PPU_bank_10 = &vromimage[(map4_vrom_switch[2]*0x0400+0x0000)];
	    PPU_bank_14 = &vromimage[(map4_vrom_switch[3]*0x0400+0x0000)];
	    PPU_bank_18 = &vromimage[(map4_vrom_switch[4]*0x0400+0x0000)];
	    PPU_bank_1c = &vromimage[(map4_vrom_switch[5]*0x0400+0x0000)];
	}
    }
}

void map4_write(unsigned short address, unsigned char value)
{
/*    deb_printf("map4_write(0x%04hx, 0x%02x)\n", address, value); */
    if (address == 0x8000) {
        map4_cmd = value;
	map4_sync_vrom();
	map4_sync_rom();
    } else if (address == 0x8001) {
        switch (map4_cmd & 7) {
	case 0:
	    value &= (cart_vrom_size << 3) - 1;
	    map4_vrom_switch[0] = value;
	    break;
	case 1:
	    value &= (cart_vrom_size << 3) - 1;
	    map4_vrom_switch[1] = value;
	    break;
	case 2:
	    value &= (cart_vrom_size << 3) - 1;
	    map4_vrom_switch[2] = value;
	    break;
	case 3:
	    value &= (cart_vrom_size << 3) - 1;
	    map4_vrom_switch[3] = value;
	    break;
	case 4:
	    value &= (cart_vrom_size << 3) - 1;
	    map4_vrom_switch[4] = value;
	    break;
	case 5:
	    value &= (cart_vrom_size << 3) - 1;
	    map4_vrom_switch[5] = value;
	    break;
	case 6:
	    if (value >= (cart_rom_size << 1)) {
		int i;
		
		i = 0x7f;
		while ((value &= i) >= (cart_rom_size << 1)) {
		    i >>= 1;
		}
	    }
            map4_rom_switch[0] = value;
	    break;
	case 7:
	    if (value >= (cart_rom_size << 1)) {
		int i;
		
		i = 0x7f;
		while ((value &= i) >= (cart_rom_size << 1)) {
		    i >>= 1;
		}
	    }
            map4_rom_switch[1] = value;
	    break;
	default:
	    deb_printf("map4_write: unknown command. (0x%02x)\n", map4_cmd);
        }
	map4_sync_rom();
	map4_sync_vrom();
    } else if (address == 0xa000) {
/*         deb_printf("map4_write: mirror control? (0x%02x)\n", value); */
	if (value & 0x40) {
	    PPU_mirror_one_high();
	} else {
	    if (value & 0x01) {
		PPU_mirror_horizontal();
	    } else {
		PPU_mirror_vertical();
	    }
	}
    } else if (address == 0xc000) {
	map4_irq_count = value;
    } else if (address == 0xc001) {
	map4_irq_latch = value;
    } else if (address == 0xe000) {
	map4_irq_state = 0;
	map4_irq_count = map4_irq_latch;
    } else if (address == 0xe001) {
	map4_irq_state = 1;
    } else {
        deb_printf("map4_write: unknown address. (0x%04hx, 0x%02x)\n", address, value);
    }
}

void map4_hsync(cal_cpu cpu)
{
    extern unsigned char ppu_cntrl2;
    
    if ((ppu_cntrl2 & 0x18) && map4_irq_state && (map4_irq_count-- == 0)) {
	map4_irq_count = map4_irq_latch;
	cpu->irq(cpu);
    }
}

void map4_vsync(cal_cpu cpu)
{
    map4_irq_count = map4_irq_latch;
}

void map4_init(void)
{
    ppu_latchfunc = NULL;
    map4_cmd = 0;
    map4_rom_switch[0] = 0;
    map4_rom_switch[1] = 1;
    map4_sync_rom();
    map4_vrom_switch[0] = 0;
    map4_vrom_switch[1] = 2;
    map4_vrom_switch[2] = 4;
    map4_vrom_switch[3] = 5;
    map4_vrom_switch[4] = 6;
    map4_vrom_switch[5] = 7;
    map4_sync_vrom();
}

mapper_iface_t map4_iface = {map4_init, map4_write, map4_hsync, map4_vsync};


/* MAPPER 7 Implementation */

void map7_write(unsigned short address, unsigned char value)
{
    if (value & 0x10) {
	PPU_mirror_one_high();
    } else {
	PPU_mirror_one_low();
    }
    
    if ((value & 0xf) >= (cart_rom_size >> 1)) {
	int i;

	deb_printf("map7_write: page beyond end of cartridge, fudging values.\n");

	i = 0x7f;
	while ((value &= i) >= (cart_rom_size >> 1)) {
	    i >>= 1;
	}
    }
    bank8_reg = (value & 0xf) << 2;
    bankA_reg = bank8_reg + 1;
    bankC_reg = bankA_reg + 1;
    bankE_reg = bankC_reg + 1;
    sync_banks();
}

void map7_init(void)
{
    ppu_latchfunc = NULL;
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = 2;
    bankE_reg = 3;
    sync_banks();
}

mapper_iface_t map7_iface = {map7_init, map7_write, dummy_hsync, dummy_vsync};


/* MAPPER 9 Implementation */

unsigned char map9_latch1_lo;
unsigned char map9_latch1_hi;
unsigned char map9_latch1_state;
unsigned char map9_latch2_lo;
unsigned char map9_latch2_hi;
unsigned char map9_latch2_state;

void map9_latchfunc(unsigned short address)
{
    if ((address & 0x3ff0) == 0x0fd0) {
	map9_latch1_state = 0;
	PPU_bank_00 = &vromimage[(map9_latch1_lo << 12) + 0x0000];
	PPU_bank_04 = &vromimage[(map9_latch1_lo << 12) + 0x0400];
	PPU_bank_08 = &vromimage[(map9_latch1_lo << 12) + 0x0800];
	PPU_bank_0c = &vromimage[(map9_latch1_lo << 12) + 0x0c00];
    } else if ((address & 0x3ff0) == 0x0fe0) {
	map9_latch1_state = 1;
	PPU_bank_00 = &vromimage[(map9_latch1_hi << 12) + 0x0000];
	PPU_bank_04 = &vromimage[(map9_latch1_hi << 12) + 0x0400];
	PPU_bank_08 = &vromimage[(map9_latch1_hi << 12) + 0x0800];
	PPU_bank_0c = &vromimage[(map9_latch1_hi << 12) + 0x0c00];
    }
    if ((address & 0x3ff0) == 0x1fd0) {
	map9_latch2_state = 0;
	PPU_bank_10 = &vromimage[(map9_latch2_lo << 12) + 0x0000];
	PPU_bank_14 = &vromimage[(map9_latch2_lo << 12) + 0x0400];
	PPU_bank_18 = &vromimage[(map9_latch2_lo << 12) + 0x0800];
	PPU_bank_1c = &vromimage[(map9_latch2_lo << 12) + 0x0c00];
    } else if ((address & 0x3ff0) == 0x1fe0) {
	map9_latch2_state = 1;
	PPU_bank_10 = &vromimage[(map9_latch2_hi << 12) + 0x0000];
	PPU_bank_14 = &vromimage[(map9_latch2_hi << 12) + 0x0400];
	PPU_bank_18 = &vromimage[(map9_latch2_hi << 12) + 0x0800];
	PPU_bank_1c = &vromimage[(map9_latch2_hi << 12) + 0x0c00];
    }
}

void map9_write(unsigned short address, unsigned char value)
{
    if (address < 0xa000) {
	/* Ignore it. It's not in spec. */
    } else if (address < 0xb000) {
	bank8_reg = value;
	sync_banks();
    } else if (address < 0xc000) {
	map9_latch1_lo = value;
	if (map9_latch1_state == 0) {
	    PPU_bank_00 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_04 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_08 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_0c = &vromimage[(value << 12) + 0x0c00];
	}
    } else if (address < 0xd000) {
	map9_latch1_hi = value;
	if (map9_latch1_state == 1) {
	    PPU_bank_00 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_04 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_08 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_0c = &vromimage[(value << 12) + 0x0c00];
	}
    } else if (address < 0xe000) {
	map9_latch2_lo = value;
	if (map9_latch2_state == 0) {
	    PPU_bank_10 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_14 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_18 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_1c = &vromimage[(value << 12) + 0x0c00];
	}
    } else if (address < 0xf000) {
	map9_latch2_hi = value;
	if (map9_latch2_state == 1) {
	    PPU_bank_10 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_14 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_18 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_1c = &vromimage[(value << 12) + 0x0c00];
	}
    } else {
	if (value & 1) {
	    PPU_mirror_horizontal();
	} else {
	    PPU_mirror_vertical();
	}
    }
}

void map9_init(void)
{
    map9_latch1_state = 1;
    map9_latch2_state = 1;
    ppu_latchfunc = map9_latchfunc;
    bank8_reg = 0;
    bankA_reg = ((cart_rom_size - 1) << 1) - 1;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t map9_iface = {map9_init, map9_write, dummy_hsync, dummy_vsync};


/* MAPPER A (10) Implementation */

unsigned char mapa_latch1_lo;
unsigned char mapa_latch1_hi;
unsigned char mapa_latch1_state;
unsigned char mapa_latch2_lo;
unsigned char mapa_latch2_hi;
unsigned char mapa_latch2_state;

void mapA_latchfunc(unsigned short address)
{
    if ((address & 0x3ff0) == 0x0fd0) {
	mapa_latch1_state = 0;
	PPU_bank_00 = &vromimage[(mapa_latch1_lo << 12) + 0x0000];
	PPU_bank_04 = &vromimage[(mapa_latch1_lo << 12) + 0x0400];
	PPU_bank_08 = &vromimage[(mapa_latch1_lo << 12) + 0x0800];
	PPU_bank_0c = &vromimage[(mapa_latch1_lo << 12) + 0x0c00];
    } else if ((address & 0x3ff0) == 0x0fe0) {
	mapa_latch1_state = 1;
	PPU_bank_00 = &vromimage[(mapa_latch1_hi << 12) + 0x0000];
	PPU_bank_04 = &vromimage[(mapa_latch1_hi << 12) + 0x0400];
	PPU_bank_08 = &vromimage[(mapa_latch1_hi << 12) + 0x0800];
	PPU_bank_0c = &vromimage[(mapa_latch1_hi << 12) + 0x0c00];
    } else if ((address & 0x3ff0) == 0x1fd0) {
	mapa_latch2_state = 0;
	PPU_bank_10 = &vromimage[(mapa_latch2_lo << 12) + 0x0000];
	PPU_bank_14 = &vromimage[(mapa_latch2_lo << 12) + 0x0400];
	PPU_bank_18 = &vromimage[(mapa_latch2_lo << 12) + 0x0800];
	PPU_bank_1c = &vromimage[(mapa_latch2_lo << 12) + 0x0c00];
    } else if ((address & 0x3ff0) == 0x1fe0) {
	mapa_latch2_state = 1;
	PPU_bank_10 = &vromimage[(mapa_latch2_hi << 12) + 0x0000];
	PPU_bank_14 = &vromimage[(mapa_latch2_hi << 12) + 0x0400];
	PPU_bank_18 = &vromimage[(mapa_latch2_hi << 12) + 0x0800];
	PPU_bank_1c = &vromimage[(mapa_latch2_hi << 12) + 0x0c00];
    }
}

void mapA_write(unsigned short address, unsigned char value)
{
    if (address < 0xa000) {
	/* Ignore it. It's not in spec. */
    } else if (address < 0xb000) {
	bank8_reg = value << 1;
	bankA_reg = (value << 1) + 1;
	sync_banks();
    } else if (address < 0xc000) {
	mapa_latch1_lo = value;
	if (mapa_latch1_state == 0) {
	    PPU_bank_00 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_04 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_08 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_0c = &vromimage[(value << 12) + 0x0c00];
	}
    } else if (address < 0xd000) {
	mapa_latch1_hi = value;
	if (mapa_latch1_state == 1) {
	    PPU_bank_00 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_04 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_08 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_0c = &vromimage[(value << 12) + 0x0c00];
	}
    } else if (address < 0xe000) {
	mapa_latch2_lo = value;
	if (mapa_latch2_state == 0) {
	    PPU_bank_10 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_14 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_18 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_1c = &vromimage[(value << 12) + 0x0c00];
	}
    } else if (address < 0xf000) {
	mapa_latch2_hi = value;
	if (mapa_latch2_state == 1) {
	    PPU_bank_10 = &vromimage[(value << 12) + 0x0000];
	    PPU_bank_14 = &vromimage[(value << 12) + 0x0400];
	    PPU_bank_18 = &vromimage[(value << 12) + 0x0800];
	    PPU_bank_1c = &vromimage[(value << 12) + 0x0c00];
	}
    } else {
	if (value & 1) {
	    PPU_mirror_horizontal();
	} else {
	    PPU_mirror_vertical();
	}
    }
}

void mapA_init(void)
{
    mapa_latch1_state = 1;
    mapa_latch2_state = 1;
    ppu_latchfunc = mapA_latchfunc;
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t mapA_iface = {mapA_init, mapA_write, dummy_hsync, dummy_vsync};


/* MAPPER B (11) Implementation */

void mapB_write(unsigned short address, unsigned char value)
{
    unsigned char vromptr;
    unsigned char romptr;

    vromptr = (value >> 4) & 0x0f;
    if ((vromptr) > (cart_vrom_size >> 1)) {
	int i;

	i = 0x7;
	while ((vromptr &= i) >= cart_vrom_size) {
	    i >>= 1;
	}
    }

    romptr = value & 0x0f;
    if ((romptr) > (cart_rom_size >> 1)) {
	int i;

	i = 0x7;
	while ((romptr &= i) >= cart_rom_size) {
	    i >>= 1;
	}
    }
    
    PPU_bank_00 = &vromimage[(vromptr << 13)+0x0000];
    PPU_bank_04 = &vromimage[(vromptr << 13)+0x0400];
    PPU_bank_08 = &vromimage[(vromptr << 13)+0x0800];
    PPU_bank_0c = &vromimage[(vromptr << 13)+0x0c00];
    PPU_bank_10 = &vromimage[(vromptr << 13)+0x1000];
    PPU_bank_14 = &vromimage[(vromptr << 13)+0x1400];
    PPU_bank_18 = &vromimage[(vromptr << 13)+0x1800];
    PPU_bank_1c = &vromimage[(vromptr << 13)+0x1c00];
    bank8_reg = (romptr << 2);
    bankA_reg = (romptr << 2) + 1;
    bankC_reg = (romptr << 2) + 2;
    bankE_reg = (romptr << 2) + 3;
    sync_banks();
}

void mapB_init(void)
{
    ppu_latchfunc = NULL;
    mapB_write(0, 0);
    sync_banks();
}

mapper_iface_t mapB_iface = {mapB_init, mapB_write, dummy_hsync, dummy_vsync};


/* MAPPER F (15) Implementation */

void mapF_write(unsigned short address, unsigned char value)
{
    int bank = ((value & (cart_rom_size - 1)) << 1);
    
    if (address == 0x8000) {
	bank8_reg = bank;
	bankA_reg = bank + 1;
	bankC_reg = bank + 2;
	bankE_reg = bank + 3;
	sync_banks();
	
	if (value & 64) {
	    PPU_mirror_horizontal();
	} else {
	    PPU_mirror_vertical();
	}
    } else if (address == 0x8001) {
	bankC_reg = bank;
	bankE_reg = bank + 1;
	sync_banks();
    } else if (address == 0x8002) {
	bank8_reg = bank + ((value & 0x80)? 1 : 0);
	bankA_reg = bank8_reg;
	bankC_reg = bank8_reg;
	bankE_reg = bank8_reg;
	sync_banks();
    } else if (address == 0x8003) {
	bankC_reg = bank;
	bankE_reg = bank + 1;
	sync_banks();
	
	if (value & 64) {
	    PPU_mirror_horizontal();
	} else {
	    PPU_mirror_vertical();
	}
    } else {
	/* FIXME: what about other addresses? */
    }
}

void mapF_init(void)
{
    ppu_latchfunc = NULL;
    mapF_write(0, 0);
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = 2;
    bankE_reg = 3;
    sync_banks();
}

mapper_iface_t mapF_iface = {mapF_init, mapF_write, dummy_hsync, dummy_vsync};


/* MAPPER 14 (20) Implementation */

unsigned char *map14_rom;
unsigned char map14_irqflag1;
unsigned char map14_irqflag2;
unsigned char *map14_diskptr;
unsigned char *map14_ram;

/* #include "m6502.h" */

unsigned char map14_b4read(cal_cpu cpu, unsigned long addr)
{
    if (addr == 0x4030) { /* Disk Status Register 0 */
/* 	deb_printf("FDS: DSR0 at IP=0x%04hx.\n", cpu->data.d_marat6502->PC.W); */
	deb_printf("FDS: DSR0 read.\n");
	map14_diskptr--; /* FIXME: this is a cheap hack */
	return 0x00;
    } else if (addr == 0x4031) { /* Port Input */
/* 	deb_printf("FDS: port input at IP=0x%04hx.\n", cpu->data.d_marat6502->PC.W); */
/* 	deb_printf("FDS: 4031 (0x%02x).\n", *map14_diskptr); */
	return *map14_diskptr++; /* FIXME: incrementing the pointer should probably be done at interrupt generation */
    } else if (addr == 0x4032) { /* Disk Status Register 1 */
/* 	deb_printf("FDS: DSR1 read.\n"); */
	return 0x00; /* FIXME: may be wrong */
    } else if (addr == 0x4033) { /* ExPort Input */
/* 	deb_printf("FDS: ExPort read at IP=0x%04hx.\n", cpu->data.d_marat6502->PC.W); */
	deb_printf("FDS: ExPort read.\n");
	return 0x80; /* FIXME: may be wrong */
    } else {
	deb_printf("FDS: unknown read from 0x%04x.\n", addr);
	return 0;
    }
}

void map14_b4write(cal_cpu cpu, unsigned long addr, unsigned char data)
{
    if (addr < 0x4040) {
	if (addr == 0x4024) {
/* 	    deb_printf("FDS: 4024 0x%02x at PC=0x%04hx.\n", data, cpu->data.d_marat6502->PC.W); */
/* 	    deb_printf("FDS: 4024 0x%02x.\n", data); */
	    map14_irqflag2 = 1;
	} else if (addr == 0x4025) {
/* 	    deb_printf("FDS: 4025 write at PC=0x%04hx.\n", cpu->data.d_marat6502->PC.W); */
	    map14_irqflag1 = 0;
	    if ((data & 0xf7) == 0x26) { /* FIXME: this is silly */
		deb_printf("FDS: 4025 phase A.\n");
    map14_diskptr = romimage + 0x10;
	    } else if ((data & 0xf7) == 0x27) {
		deb_printf("FDS: 4025 phase B1.\n");
    map14_diskptr = romimage + 0x10;
	    } else if ((data & 0xf7) == 0x25) {
		deb_printf("FDS: 4025 phase B2.\n");
	    } else if ((data & 0xf7) == 0x65) {
		deb_printf("FDS: 4025 phase C.\n");
	    } else if ((data & 0xf7) == 0xe5) {
		deb_printf("FDS: 4025 phase D.\n");
		map14_irqflag1 = 1;
		map14_irqflag2 = 1;
	    } else if ((data & 0xf7) == 0xf5) {
		deb_printf("FDS: 4025 phase E?.\n");
		map14_irqflag1 = 1;
		map14_irqflag2 = 1;
	    } else {
		deb_printf("FDS: 4025 unknown phase (0x%02x).\n", data);
	    }
	    if (data & 0x08) {
		PPU_mirror_horizontal();
	    } else {
		PPU_mirror_vertical();
	    }
	} else {
	    deb_printf("FDS: unknown write (0x%04x, 0x%02x).\n", addr, data);
	}
    } else {
	/* It's a sound reg, ignore it for now */
    }
}

void map14_hsync(cal_cpu cpu)
{
/*     if (map14_irqflag1 && map14_irqflag2) { */
    if (map14_irqflag1) {
	map14_irqflag2 = 0;
	cpu->irq(cpu);
    }
}

void map14_write(unsigned short address, unsigned char value)
{
    if (address < 0xe000) {
	map14_ram[address - 0x8000] = value;
    } else {
	deb_printf("map14_write(0x%04hx, 0x%02x).\n", address, value);
    }
}

void map14_init(void)
{
    FILE *romfile;

    map14_rom = malloc(0x2000);
    if (!map14_rom) {
	deb_printf("out of memory.\n");
    }
    romfile = fopen("disksys.rom", "rb");
    if (romfile) {
	fread(map14_rom, 0x2000, 1, romfile);
	fclose(romfile);
    } else {
	deb_printf("error opening FDS rom file.\n");
    }

    map14_diskptr = romimage + 0x10;

    map14_ram = calloc(1, 0x6000);
	
    ppu_latchfunc = NULL;
    bank_8 = map14_ram + 0x0000;
    bank_A = map14_ram + 0x2000;
    bank_C = map14_ram + 0x4000;
    bank_E = map14_rom;

    nes_bank4_read_hook = map14_b4read;
    nes_bank4_write_hook = map14_b4write;
}

mapper_iface_t map14_iface = {map14_init, map14_write, map14_hsync, dummy_vsync};


/* MAPPER 17 (23) Implementation */

unsigned char map17_bank_regs[16];

void map17_sync_vrom(int bank)
{
    unsigned char value;

    value = map17_bank_regs[(bank << 1)];
    value |= map17_bank_regs[(bank << 1) + 1] << 4;

    *PPU_banks[bank] = &vromimage[value << 10];
}

void map17_write(unsigned short address, unsigned char value)
{
    if (address == 0x8000) {
	bank8_reg = value & ((cart_rom_size << 1) - 1);
	sync_banks();
    } else if (address == 0x9000) {
	switch (value & 3) {
	case 0:
	    PPU_mirror_vertical();
	    break;
	case 1:
	    PPU_mirror_horizontal();
	    break;
	case 2:
	    PPU_mirror_one_low();
	    break;
	case 3:
	    PPU_mirror_one_high();
	    break;
	}
    } else if (address == 0xa000) {
   	bankA_reg = value & ((cart_rom_size << 1) - 1);
	sync_banks();
    } else if ((address & 0xfffc) == 0xb000) {
	map17_bank_regs[0+(address & 3)]  = value & 15;
	map17_sync_vrom(0+((address & 2) >> 1));
    } else if ((address & 0xfffc) == 0xc000) {
	map17_bank_regs[4+(address & 3)]  = value & 15;
	map17_sync_vrom(2+((address & 2) >> 1));
    } else if ((address & 0xfffc) == 0xd000) {
	map17_bank_regs[8+(address & 3)]  = value & 15;
	map17_sync_vrom(4+((address & 2) >> 1));
    } else if ((address & 0xfffc) == 0xe000) {
	map17_bank_regs[12+(address & 3)]  = value & 15;
	map17_sync_vrom(6+((address & 2) >> 1));
    } else {
	deb_printf("map23: unknown addr 0x%04hx = 0x%02x.\n", address, value);
    }
}

void map17_init(void)
{
    ppu_latchfunc = NULL;
#if 0
    PPU_bank_00 = &vromimage[0x0000];
    PPU_bank_04 = &vromimage[0x0400];
    PPU_bank_08 = &vromimage[0x0800];
    PPU_bank_0c = &vromimage[0x0c00];
    PPU_bank_10 = &vromimage[0x1000];
    PPU_bank_14 = &vromimage[0x1400];
    PPU_bank_18 = &vromimage[0x1800];
    PPU_bank_1c = &vromimage[0x1c00];
#endif
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t map17_iface = {map17_init, map17_write, dummy_hsync, dummy_vsync};


/* MAPPER 18 (24) Implementation */

void map18_write(unsigned short address, unsigned char value)
{
    if (address == 0x8000) {
	if (value >= cart_rom_size) {
	    int i;
	    
	    i = 0x7f;
	    while ((value &= i) >= cart_rom_size) {
		i >>= 1;
	    }
	    deb_printf("map2_write: substituting page %02x.\n", value);
	}
	bank8_reg = (value << 1);
	bankA_reg = (value << 1) + 1;
	sync_banks();
    } else if (address < 0x9000) {
	deb_printf("map18_write: unknown write in 0x8001-0x8fff range.\n");
    } else if ((address & 0xfffc) == 0xd000) {
	*PPU_banks[address & 3] = &vromimage[value << 10];
    } else if ((address & 0xfffc) == 0xe000) {
	*PPU_banks[4 + (address & 3)] = &vromimage[value << 10];
    }
}

void map18_init(void)
{
    ppu_latchfunc = NULL;
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t map18_iface = {map18_init, map18_write, dummy_hsync, dummy_vsync};


/* MAPPER 20 (32) Implementation */

unsigned char map20_saved_9fff;

void map20_write(unsigned short address, unsigned char value)
{
    if (address == 0x8fff) {
	if (map20_saved_9fff & 0x02) {
	    bankC_reg = value;
	} else {
	    bank8_reg = value;
	}
	sync_banks();
    } else if (address == 0x9fff) {
	map20_saved_9fff = value;
	if (value & 1) {
	    PPU_mirror_horizontal();
	} else {
	    PPU_mirror_vertical();
	}
    } else if (address == 0xafff) {
	bankA_reg = value;
	sync_banks();
    } else if ((address & 0xfff8) == 0xbff0) {
	*PPU_banks[address & 7] = &vromimage[value * 0x400];
    } else {
	deb_printf("map32: unknown address 0x%04hx = 0x%02x.\n", address, value);
    }
}

void map20_init(void)
{
    ppu_latchfunc = NULL;
    bank8_reg = 0;
    bankA_reg = 0;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t map20_iface = {map20_init, map20_write, dummy_hsync, dummy_vsync};


/* MAPPER 42 (66) Implementation */

void map42_write(unsigned short address, unsigned char value)
{
    unsigned char vromptr;
    unsigned char romptr;

    vromptr = value & 0x03;
    romptr = (value >> 4) & 0x03;
    
    PPU_bank_00 = &vromimage[(vromptr << 13)+0x0000];
    PPU_bank_04 = &vromimage[(vromptr << 13)+0x0400];
    PPU_bank_08 = &vromimage[(vromptr << 13)+0x0800];
    PPU_bank_0c = &vromimage[(vromptr << 13)+0x0c00];
    PPU_bank_10 = &vromimage[(vromptr << 13)+0x1000];
    PPU_bank_14 = &vromimage[(vromptr << 13)+0x1400];
    PPU_bank_18 = &vromimage[(vromptr << 13)+0x1800];
    PPU_bank_1c = &vromimage[(vromptr << 13)+0x1c00];
    bank8_reg = (romptr << 2);
    bankA_reg = (romptr << 2) + 1;
    bankC_reg = (romptr << 2) + 2;
    bankE_reg = (romptr << 2) + 3;
    sync_banks();
}

void map42_init(void)
{
    ppu_latchfunc = NULL;
    map42_write(0, 0);
    sync_banks();
}

mapper_iface_t map42_iface = {map42_init, map42_write, dummy_hsync, dummy_vsync};


/* MAPPER 44 (68) Implementation */

unsigned char map44_a10l;
unsigned char map44_a10h;
unsigned char map44_mirr;

void map44_sync_mirrors(void)
{
    if (map44_mirr & 0x10) {
	switch (map44_mirr & 3) {
	case 0:
	    PPU_nametables[0] = &vromimage[map44_a10l << 10] + 0x20000;
	    PPU_nametables[1] = &vromimage[map44_a10h << 10] + 0x20000;
	    PPU_nametables[2] = &vromimage[map44_a10l << 10] + 0x20000;
	    PPU_nametables[3] = &vromimage[map44_a10h << 10] + 0x20000;
	    break;
	case 1:
	    PPU_nametables[0] = &vromimage[map44_a10l << 10] + 0x20000;
	    PPU_nametables[1] = &vromimage[map44_a10l << 10] + 0x20000;
	    PPU_nametables[2] = &vromimage[map44_a10h << 10] + 0x20000;
	    PPU_nametables[3] = &vromimage[map44_a10h << 10] + 0x20000;
	    break;
	case 2:
	    PPU_nametables[0] = &vromimage[map44_a10l << 10] + 0x20000;
	    PPU_nametables[1] = &vromimage[map44_a10l << 10] + 0x20000;
	    PPU_nametables[2] = &vromimage[map44_a10l << 10] + 0x20000;
	    PPU_nametables[3] = &vromimage[map44_a10l << 10] + 0x20000;
	    break;
	case 3:
	    PPU_nametables[0] = &vromimage[map44_a10h << 10] + 0x20000;
	    PPU_nametables[1] = &vromimage[map44_a10h << 10] + 0x20000;
	    PPU_nametables[2] = &vromimage[map44_a10h << 10] + 0x20000;
	    PPU_nametables[3] = &vromimage[map44_a10h << 10] + 0x20000;
	    break;
	}
    } else {
	if (map44_mirr & 2) {
	    if (map44_mirr & 1) {
		PPU_mirror_one_high();
	    } else {
		PPU_mirror_one_low();
	    }
	} else {
	    if (map44_mirr & 1) {
		PPU_mirror_vertical();
	    } else {
		PPU_mirror_horizontal();
	    }
	}
    }
}

void map44_write(unsigned short address, unsigned char value)
{
    switch ((address & 0x7000) >> 12) {
    case 0:
	PPU_bank_00 = &vromimage[value << 11];
	PPU_bank_04 = &vromimage[value << 11] + 0x400;
	break;
    case 1:
	PPU_bank_08 = &vromimage[value << 11];
	PPU_bank_0c = &vromimage[value << 11] + 0x400;
	break;
    case 2:
	PPU_bank_10 = &vromimage[value << 11];
	PPU_bank_14 = &vromimage[value << 11] + 0x400;
	break;
    case 3:
	PPU_bank_18 = &vromimage[value << 11];
	PPU_bank_1c = &vromimage[value << 11] + 0x400;
	break;
    case 4:
	map44_a10l = value;
	map44_sync_mirrors();
	break;
    case 5:
	map44_a10h = value;
	map44_sync_mirrors();
	break;
    case 6:
	map44_mirr = value;
	map44_sync_mirrors();
	break;
    case 7:
	bank8_reg = value << 1;
	bankA_reg = (value << 1) + 1;
	sync_banks();
	break;
    }
}

void map44_init(void)
{
    ppu_latchfunc = NULL;
    bank8_reg = 0;
    bankA_reg = 1;
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    sync_banks();
}

mapper_iface_t map44_iface = {map44_init, map44_write, dummy_hsync, dummy_vsync};


/* MAPPER 4E (78) Implementation */

void map4E_write(unsigned short address, unsigned char value)
{
    unsigned char vromptr;
    unsigned char romptr;

    vromptr = (value >> 4) & 0x0f;
    if ((vromptr) > (cart_vrom_size >> 1)) {
	int i;

	i = 0x7;
	while ((vromptr &= i) >= cart_vrom_size) {
	    i >>= 1;
	}
    }

    romptr = value & 0x0f;
    if ((romptr) > cart_rom_size) {
	int i;

	i = 0x7;
	while ((romptr &= i) >= cart_rom_size) {
	    i >>= 1;
	}
    }
    
    PPU_bank_00 = &vromimage[(vromptr << 13)+0x0000];
    PPU_bank_04 = &vromimage[(vromptr << 13)+0x0400];
    PPU_bank_08 = &vromimage[(vromptr << 13)+0x0800];
    PPU_bank_0c = &vromimage[(vromptr << 13)+0x0c00];
    PPU_bank_10 = &vromimage[(vromptr << 13)+0x1000];
    PPU_bank_14 = &vromimage[(vromptr << 13)+0x1400];
    PPU_bank_18 = &vromimage[(vromptr << 13)+0x1800];
    PPU_bank_1c = &vromimage[(vromptr << 13)+0x1c00];
    bank8_reg = (romptr << 1);
    bankA_reg = (romptr << 1) + 1;
    sync_banks();
}

void map4E_init(void)
{
    bankC_reg = ((cart_rom_size - 1) << 1);
    bankE_reg = ((cart_rom_size - 1) << 1) + 1;
    ppu_latchfunc = NULL;
    map4E_write(0, 0);
    sync_banks();
}

mapper_iface_t map4E_iface = {map4E_init, map4E_write, dummy_hsync, dummy_vsync};


/*
 * $Log: mappers.c,v $
 * Revision 1.54  1999/07/05 01:30:02  nyef
 * fixed outrange rom bank compensation on mapper 4
 *
 * Revision 1.53  1999/05/04 01:55:37  nyef
 * added preliminary support for mapper 24
 *
 * Revision 1.52  1999/05/03 00:35:52  nyef
 * added ppu mirroring control to FDS
 *
 * Revision 1.51  1999/05/02 21:34:11  nyef
 * fixed typo in FDS mapper (14 (20)) that was causing problems
 *
 * Revision 1.50  1999/02/20 19:08:58  nyef
 * added large amounts of stuff to the FDS mapper (still doesn't work)
 *
 * Revision 1.49  1999/02/14 06:39:30  nyef
 * added preliminary support for mapper 14 (20), the FDS mapper
 *
 * Revision 1.48  1999/01/03 03:16:05  nyef
 * added Kritz's changes to mapper 20 (32).
 *
 * Revision 1.47  1999/01/02 19:09:25  nyef
 * added support for mapper F (15) by kritz.
 *
 * Revision 1.46  1999/01/02 16:56:28  nyef
 * fixed infinite loop in mapper 7.
 *
 * Revision 1.45  1999/01/02 01:17:17  nyef
 * added preliminary mapper 17 (23) support.
 *
 * Revision 1.44  1999/01/02 00:51:52  nyef
 * added preliminary (nonfunctional) mapper 20 (32) support.
 *
 * Revision 1.43  1999/01/01 22:11:40  nyef
 * added support for mapper 44 (68).
 *
 * Revision 1.42  1998/12/31 22:35:47  nyef
 * added outrange vrom bank compensation to mapper 4.
 * Crystalis no longer breaks at the name screen.
 *
 * Revision 1.41  1998/12/31 19:06:49  nyef
 * added preliminary (untested) support for mapper 42 (66).
 *
 * Revision 1.40  1998/12/31 18:59:20  nyef
 * added preliminary (untested) support for mapper 4E (78).
 *
 * Revision 1.39  1998/12/31 03:38:31  nyef
 * changed map4_hsync() to only check irq counter if screen is on.
 *
 * Revision 1.38  1998/12/29 02:15:35  nyef
 * fixed map4_sync_rom() so that the SMB3 intro sortof works.
 *
 * Revision 1.37  1998/12/26 01:02:58  nyef
 * changed mapper system to support support for mappers > 15.
 *
 * Revision 1.36  1998/12/20 03:43:13  nyef
 * updated mapper_support[] table.
 *
 * Revision 1.35  1998/12/20 03:31:32  nyef
 * added second latch to mapper 9.
 * mapper 9 now works just about perfectly.
 *
 * Revision 1.34  1998/12/15 04:41:27  nyef
 * added support for 512k and 1024k mapper 1 carts.
 *
 * Revision 1.33  1998/12/15 03:50:14  nyef
 * fixed bank mask bug in mapper 1.
 *
 * Revision 1.32  1998/12/15 03:43:20  nyef
 * reorganized the mapper 1 implementation to be much cleaner looking.
 *
 * Revision 1.31  1998/12/15 01:50:24  nyef
 * changed mapper hsync and vsync routines to conform to new spec.
 *
 * Revision 1.30  1998/11/28 03:59:31  nyef
 * changed map4_hsync() to postdecrement rather than predecrement the irq
 * count. Crystalis looks slightly better.
 *
 * Revision 1.29  1998/11/16 23:25:53  nyef
 * fixed mistake in the mapper 10 latchfunc.
 *
 * Revision 1.28  1998/11/05 03:07:40  nyef
 * added default VRom mapping to mapper 1.
 *
 * Revision 1.27  1998/11/03 01:41:24  nyef
 * added preliminary support for mapper 10.
 *
 * Revision 1.26  1998/11/01 22:09:40  nyef
 * commented out mapper 4 mirror control deb_printf.
 *
 * Revision 1.25  1998/10/04 14:14:16  nyef
 * fixed out-range bank compensation on mappers 11, 7, 3, 2 and 1.
 *
 * Revision 1.24  1998/10/04 14:10:52  nyef
 * fixed out-range rom bank compensation on mapper 4.
 * this will probably need to be propagated to other mappers.
 *
 * Revision 1.23  1998/10/03 23:18:36  nyef
 * added out-range rom bank compensation to mapper 4.
 *
 * Revision 1.22  1998/10/03 23:10:58  nyef
 * changed mapper 1 to only have 1 shift register.
 * this may improve some games.
 *
 * Revision 1.21  1998/10/03 13:46:18  nyef
 * added more code to mapper 11. (still doesn't work)
 *
 * Revision 1.20  1998/10/01 02:43:16  nyef
 * added preliminary support for mapper B (11).
 *
 * Revision 1.19  1998/10/01 02:35:17  nyef
 * added support for mapper 9.
 *
 * Revision 1.18  1998/09/01 01:00:01  nyef
 * fixed mapper 0. (added vrom stuff. oops.)
 *
 * Revision 1.17  1998/08/30 17:07:10  nyef
 * changed to use the new ppu_flags system instead of PPU_use_vrom.
 *
 * Revision 1.16  1998/08/24 01:57:53  nyef
 * fixed bug with 8k vrom pages on mapper 1.
 *
 * Revision 1.15  1998/08/24 01:24:19  nyef
 * more work on mapper 4.
 *
 * Revision 1.14  1998/08/24 00:29:06  nyef
 * more changes to mapper 4.
 *
 * Revision 1.13  1998/08/23 22:42:46  nyef
 * added hooks for calling mappers on vsync and hsync.
 * added preliminary irq support to mapper 4.
 *
 * Revision 1.12  1998/08/23 18:59:16  nyef
 * more changes to mapper 1. effectiveness unknown.
 *
 * Revision 1.11  1998/08/22 19:33:57  nyef
 * changes to mapper 4. effectiveness unknown.
 *
 * Revision 1.10  1998/08/22 18:41:37  nyef
 * more changes to mapper 1.
 *
 * Revision 1.9  1998/08/22 17:53:22  nyef
 * fixed more bugs in the mapper 1 emulation.
 *
 * Revision 1.8  1998/08/22 16:37:59  nyef
 * added support for mapper 1 roms that don't use VROM.
 *
 * Revision 1.7  1998/08/22 16:33:12  nyef
 * added preliminary VROM support to mapper 4.
 *
 * Revision 1.6  1998/08/04 21:56:31  nyef
 * changed the bit masks on the vrom selectors for mapper 1.
 *
 * Revision 1.5  1998/08/04 02:03:04  nyef
 * added an initialization message to set_mapper().
 *
 * Revision 1.4  1998/08/02 13:46:03  nyef
 * changed every printf to deb_printf for debug console.
 *
 * Revision 1.3  1998/07/31 02:28:00  nyef
 * upgraded the mapper 1 emulation somewhat. still incomplete, but better.
 *
 * Revision 1.2  1998/07/18 21:55:52  nyef
 * added preliminary mapper 7 support
 *
 * Revision 1.1  1998/07/11 22:18:27  nyef
 * Initial revision
 *
 */
