/*
 * mappers.h
 *
 * Mapper emulation defines
 */

/* $Id: mappers.h,v 1.3 1998/12/15 01:50:29 nyef Exp $ */

#ifndef __MAPPERS_H__
#define __MAPPERS_H__

#ifndef __EMU6502_H__
#include "emu6502.h"
#endif

#include "cal.h"

/* mapper routine pointer defines */
typedef void (* mapinit_t)(void);
typedef void (* mapwrite_t)(unsigned short address, unsigned char value);
typedef void (* maphsync_t)(cal_cpu cpu);
typedef void (* mapvsync_t)(cal_cpu cpu);

typedef struct _mapper_iface {
    mapinit_t initcode;
    mapwrite_t writecode;
    maphsync_t hsynccode;
    mapvsync_t vsynccode;
} mapper_iface_t;


/* current mapper interface */
extern mapinit_t cur_init;
extern mapwrite_t cur_write;
extern maphsync_t cur_hsync;
extern mapvsync_t cur_vsync;

/* use mapper mapper. returns 0 if successful, -1 on failure. */
int set_mapper(int mapper);

/* return current mapper, or -1 if none */
int get_mapper(void);

/* check mapper support. return 0 for none, 1 for partial, 2 for full */
int check_mapper(int mapper);


#endif /* __MAPPERS_H__ */

/*
 * $Log: mappers.h,v $
 * Revision 1.3  1998/12/15 01:50:29  nyef
 * changed mapper hsync and vsync routines to conform to new spec.
 *
 * Revision 1.2  1998/08/23 22:42:12  nyef
 * added hooks for calling mappers on vsync and hsync.
 *
 * Revision 1.1  1998/07/11 22:18:30  nyef
 * Initial revision
 *
 */
