/*
 * nes_ppu.h
 *
 * emulation interface for NES PPU
 */

/* $Id: nes_ppu.h,v 1.11 1999/06/05 02:52:31 nyef Exp $ */

#ifndef __NES_PPU_H__
#define __NES_PPU_H__

#ifndef __EMU6502_H__
#include "emu6502.h"
#endif

#ifndef __CAL_H__
#include "cal.h"
#endif

typedef struct nes_ppu *nes_ppu;
nes_ppu nes_ppu_true;

/* PPU control */
void PPU_enter_cpu(cal_cpu cpu);
void PPU_leave_cpu(cal_cpu cpu);
void PPU_mirror_horizontal(void);
void PPU_mirror_vertical(void);
void PPU_mirror_one_low(void);
void PPU_mirror_one_high(void);
int nesppu_periodic(nes_ppu ppu);

/* PPU flags */
extern int ppu_flags;
#define PPUF_VROM           (1<<0)
#define PPUF_NTSC           (1<<1)
#define PPUF_DIRTY_PATTERNS (1<<2)
#define PPUF_DIRTY_NAMES    (1<<3)

/* PPU memory initialization */
void PPU_init(void);

/* PPU functions */
void PPU_render_video_frame(void);

/* PPU I/O */
void ppu_io_write(nes_ppu ppu, unsigned short addr, unsigned char value);
unsigned char ppu_io_read(nes_ppu ppu, unsigned short addr);
void PPU_do_dma(int page);

/* PPU debug output */
void PPU_dump_memory(short address);
void PPU_dump_pattern(short address);

/* PPU memory */
extern unsigned char *PPU_bank_00;
extern unsigned char *PPU_bank_04;
extern unsigned char *PPU_bank_08;
extern unsigned char *PPU_bank_0c;
extern unsigned char *PPU_bank_10;
extern unsigned char *PPU_bank_14;
extern unsigned char *PPU_bank_18;
extern unsigned char *PPU_bank_1c;
extern unsigned char *PPU_bank_20;
extern unsigned char *PPU_bank_24;
extern unsigned char *PPU_bank_28;
extern unsigned char *PPU_bank_2c;
extern unsigned char *PPU_bank_30;
extern unsigned char *PPU_bank_34;
extern unsigned char *PPU_bank_38;
extern unsigned char *PPU_bank_3c;

extern unsigned char **PPU_banks[0x10];
extern unsigned char *PPU_nametables[4];

/* support for mappers that use latches */
typedef void (* ppulatch_t)(unsigned short address);
extern ppulatch_t ppu_latchfunc;


#endif /* __NES_PPU_H__ */

/*
 * $Log: nes_ppu.h,v $
 * Revision 1.11  1999/06/05 02:52:31  nyef
 * added prototype for nesppu_periodic()
 *
 * Revision 1.10  1999/05/29 22:31:47  nyef
 * started reorganizing the NES PPU
 *
 * Revision 1.9  1999/02/14 18:29:40  nyef
 * added function prototypes required for I/O access
 *
 * Revision 1.8  1999/01/02 00:41:53  nyef
 * added PPU_banks for mappers that find it easier to index through that.
 *
 * Revision 1.7  1999/01/01 21:47:40  nyef
 * added PPU_nametables for mappers that need to modify them by hand.
 *
 * Revision 1.6  1998/12/11 04:33:57  nyef
 * fixed for full compatability with cal.
 *
 * Revision 1.5  1998/10/01 02:34:00  nyef
 * added preliminary support for latches as per mappers 9 and 10
 *
 * Revision 1.4  1998/08/30 17:03:17  nyef
 * removed PPU_use_vrom in favor of a more general ppu_flags setup.
 *
 * Revision 1.3  1998/08/29 22:11:25  nyef
 * removed reference to PPU_video_buffer. (no longer used)
 *
 * Revision 1.2  1998/07/18 21:43:57  nyef
 * added support for one-screen mirroring
 *
 * Revision 1.1  1998/07/11 22:18:15  nyef
 * Initial revision
 *
 */
