/*
 * nes_psg.c
 *
 * NES sound emulation system
 */

/* $Id: nes_psg.c,v 1.33 1999/02/14 18:27:54 nyef Exp $ */

#define SOUND
#define SOUND_LINUX
/* #define SOUND_DOS */
/* #define DONT_USE_CAL */ /* for RockNES */

#include <unistd.h>

#ifdef SOUND_LINUX
#  include <stdio.h>
#  include <sys/types.h>
#  include <sys/stat.h>
#  include <fcntl.h>
#  include <sys/soundcard.h>
#  include <sys/ioctl.h>
#endif

#ifdef SOUND_DOS
#  include <allegro.h>
#endif

#ifndef DONT_USE_CAL
#include "cal.h"
#endif

#include "ui.h"
#include "nes_psg.h"

struct psg_event {
    long time;
    unsigned char type;
    unsigned char data;
};

/*
 * I have not noticed more than 11 events in one vsync. (usually less.)
 * this merely covers the pathological case.
 * (A massively unrolled loop that just stores
 * into a sound register.)
 */

/*
 * FIXME: the pathological case could be rather precisely calculated for
 * working code as the number of cycles per vsync divided by the number of
 * cycles required to execute a tight loop (possible unrolled) of loading
 * a byte from memory and dumping it to a sound register. probably the DMC
 * volume port, It's the only one that would warrant this type of code.
 */

struct psg_event psg_eventqueue[15000];
int cur_event;

long lasttime;
long entertime;

#define PSGET_MASK 0xfc
#define PSGET_C1 0x00
#define PSGET_W_C1A 0x00
#define PSGET_W_C1B 0x01
#define PSGET_W_C1C 0x02
#define PSGET_W_C1D 0x03
#define PSGET_C2 0x04
#define PSGET_W_C2A 0x04
#define PSGET_W_C2B 0x05
#define PSGET_W_C2C 0x06
#define PSGET_W_C2D 0x07
#define PSGET_C3 0x08
#define PSGET_W_C3A 0x08
#define PSGET_W_C3B 0x09
#define PSGET_W_C3C 0x0a
#define PSGET_W_C3D 0x0b
#define PSGET_C4 0x0c
#define PSGET_W_C4A 0x0c
#define PSGET_W_C4B 0x0d
#define PSGET_W_C4C 0x0e
#define PSGET_W_C4D 0x0f
#define PSGET_W_CTRL 0x10
#define PSGET_SYNC 0x20

/*
 * a psg write function is of the form c#$ where # is the channel number and
 *   $ is the channel register starting from a
 */

#ifndef DONT_USE_CAL
#define PSG_WRITEFUNC(name, evtype) \
void nes_psg_write_##name(cal_cpu cpu, unsigned short addr, unsigned char value) \
{ \
    psg_eventqueue[cur_event].time = lasttime + (entertime-cpu->timeleft(cpu)); \
    psg_eventqueue[cur_event].type = PSGET_W_##evtype; \
    psg_eventqueue[cur_event].data = value; \
    cur_event++; \
}
#else
#define PSG_WRITEFUNC(name, evtype) \
void nes_psg_write_##name(long cycles_left, unsigned char value) \
{ \
    psg_eventqueue[cur_event].time = lasttime + (entertime-cycles_left); \
    psg_eventqueue[cur_event].type = PSGET_W_##evtype; \
    psg_eventqueue[cur_event].data = value; \
    cur_event++; \
}
#endif

PSG_WRITEFUNC(c1a, C1A);
PSG_WRITEFUNC(c1b, C1B);
PSG_WRITEFUNC(c1c, C1C);
PSG_WRITEFUNC(c1d, C1D);

PSG_WRITEFUNC(c2a, C2A);
PSG_WRITEFUNC(c2b, C2B);
PSG_WRITEFUNC(c2c, C2C);
PSG_WRITEFUNC(c2d, C2D);

PSG_WRITEFUNC(c3a, C3A);
PSG_WRITEFUNC(c3b, C3B);
PSG_WRITEFUNC(c3c, C3C);
PSG_WRITEFUNC(c3d, C3D);

PSG_WRITEFUNC(c4a, C4A);
PSG_WRITEFUNC(c4b, C4B);
PSG_WRITEFUNC(c4c, C4C);
PSG_WRITEFUNC(c4d, C4D);

PSG_WRITEFUNC(control, CTRL);

psg_writefunc sound_regs[16] = {
    nes_psg_write_c1a,
    nes_psg_write_c1b,
    nes_psg_write_c1c,
    nes_psg_write_c1d,
    nes_psg_write_c2a,
    nes_psg_write_c2b,
    nes_psg_write_c2c,
    nes_psg_write_c2d,
    nes_psg_write_c3a,
    nes_psg_write_c3b,
    nes_psg_write_c3c,
    nes_psg_write_c3d,
    nes_psg_write_c4a,
    nes_psg_write_c4b,
    nes_psg_write_c4c,
    nes_psg_write_c4d,
};

/*
 * nes_psg_{enter,leave}_cpu are used to control system timing.
 */

void nes_psg_enter_cpu(long cycles_left)
{
    entertime = cycles_left;
}

void nes_psg_leave_cpu(long cycles_left)
{
    lasttime += entertime - cycles_left;
}

/*
 * nes_psg_vsync does the actual sound rendering.
 */

#ifdef SOUND

unsigned char wave_buffers[4][735]; /* 44100 / 60 = 735 samples per sync */

#ifdef SOUND_LINUX
unsigned char final_wave[2048];
int waveptr;
int wavflag;
#endif

unsigned char nes_psg_c1a; unsigned char nes_psg_c1b;
unsigned char nes_psg_c1c; unsigned char nes_psg_c1d;
unsigned char nes_psg_c2a; unsigned char nes_psg_c2b;
unsigned char nes_psg_c2c; unsigned char nes_psg_c2d;
unsigned char nes_psg_c3a; unsigned char nes_psg_c3b;
unsigned char nes_psg_c3c; unsigned char nes_psg_c3d;
unsigned char nes_psg_c4a; unsigned char nes_psg_c4b;
unsigned char nes_psg_c4c; unsigned char nes_psg_c4d;
unsigned char nes_psg_ctrl;
unsigned char nes_psg_c1_atl; /* active time left */
unsigned char nes_psg_c2_atl;
unsigned char nes_psg_c3_atl;
unsigned char nes_psg_c4_atl;
unsigned char *nes_psg_c1_wave;
unsigned long nes_psg_c1_skip;
unsigned long nes_psg_c1_index;
unsigned char nes_psg_c1_vol;
unsigned char *nes_psg_c2_wave;
unsigned long nes_psg_c2_skip;
unsigned long nes_psg_c2_index;
unsigned char nes_psg_c2_vol;
unsigned long nes_psg_c3_skip;
unsigned long nes_psg_c3_index;
unsigned char nes_psg_c3_vol;
unsigned long nes_psg_c4_sr; /* Shift register */
unsigned long nes_psg_c4_fdc; /* Frequency divide counter */
unsigned long nes_psg_c4_skip;
unsigned long nes_psg_c4_index;
unsigned char nes_psg_c4_vol;
#endif

int nes_psg_quality;

#ifdef SOUND
unsigned long nes_psg_pulse_magic;
unsigned long nes_psg_triangle_magic;
unsigned long nes_psg_noise_magic;
unsigned int nes_psg_samples_per_sync;
unsigned int nes_psg_cycles_per_sample;
unsigned int nes_psg_sample_rate;

struct nes_psg_quality_data {
    unsigned long pulse_magic;
    unsigned long triangle_magic;
    unsigned long noise_magic;
    unsigned int samples_per_sync;
    unsigned int cycles_per_sample;
    unsigned int sample_rate;
} nes_psg_qual[] = {
    {0xa2567000, 0x512b3800, 0x512b3800, 367, 82, 22050},
    {0x512b3800, 0x289d9c00, 0x289d9c00, 735, 41, 44100},
};

#ifdef SOUND_LINUX
int sound_fd;
#endif

#ifdef SOUND_DOS
AUDIOSTREAM *sound_stream;
#endif

/*
 * FIXME: there are some unsupported claims and manifest constants in these
 * routines. produce a description of how the NES sound system works (in
 * understandable terms) and place it in a comment in this file. something
 * like what you wish _you_ had when you were looking for information.
 */

/* FIXME: frequency sweep? */

/* NOTE: these routines use 8.24 bit fixed point math in places. */

/* FIXME: some frequency values are not handled well by this system. */

unsigned char pulse_25[0x20] = {
    0x11, 0x11, 0x11, 0x11,
    0x11, 0x11, 0x11, 0x11,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
};

unsigned char pulse_50[0x20] = {
    0x11, 0x11, 0x11, 0x11,
    0x11, 0x11, 0x11, 0x11,
    0x11, 0x11, 0x11, 0x11,
    0x11, 0x11, 0x11, 0x11,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
};

unsigned char pulse_75[0x20] = {
    0x11, 0x11, 0x11, 0x11,
    0x11, 0x11, 0x11, 0x11,
    0x11, 0x11, 0x11, 0x11,
    0x11, 0x11, 0x11, 0x11,
    0x11, 0x11, 0x11, 0x11,
    0x11, 0x11, 0x11, 0x11,
    0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00,
};

unsigned char pulse_87[0x20] = {
    0x11, 0x11, 0x11, 0x11,
    0x11, 0x11, 0x11, 0x11,
    0x11, 0x11, 0x11, 0x11,
    0x11, 0x11, 0x11, 0x11,
    0x11, 0x11, 0x11, 0x11,
    0x11, 0x11, 0x11, 0x11,
    0x11, 0x11, 0x11, 0x11,
    0x00, 0x00, 0x00, 0x00,
};

unsigned char triangle_50[0x20] = {
    0x00, 0x10, 0x20, 0x30,
    0x40, 0x50, 0x60, 0x70,
    0x80, 0x90, 0xa0, 0xb0,
    0xc0, 0xd0, 0xe0, 0xf0,
    0xff, 0xef, 0xdf, 0xcf,
    0xbf, 0xaf, 0x9f, 0x8f,
    0x7f, 0x6f, 0x5f, 0x4f,
    0x3f, 0x2f, 0x1f, 0x0f,
};

unsigned char *pulse_waves[4] = {
    pulse_87, pulse_75, pulse_50, pulse_25,
};

unsigned char nes_psg_atl[0x20] = {
    5, 127, 10, 1, 20,  2, 40,  3, 80,  4, 30,  5, 7,  6, 13,  7,
    6,   8, 12, 9, 24, 10, 48, 11, 96, 12, 36, 13, 8, 14, 16, 15,
};

void nes_psg_wave_1(void)
{
    int i;
    int cycles;
    int event;
    unsigned int freq;
    
    cycles = 0;
    event = 0;
    
    /* note: 41 CPU cycles occur between increments of i */
    for (i = 0; i < nes_psg_samples_per_sync; i++) {
	cycles += nes_psg_cycles_per_sample;
	while ((event < cur_event) && (psg_eventqueue[event].time < cycles)) {
	    if ((psg_eventqueue[event].type & PSGET_MASK) == PSGET_C1) {
		switch (psg_eventqueue[event].type & 3) {
		case 0:
		    nes_psg_c1a = psg_eventqueue[event].data;
		    nes_psg_c1_wave = pulse_waves[(nes_psg_c1a >> 6)];
		    nes_psg_c1_vol = nes_psg_c1a & 0x0f;
		    break;
		case 1:
		    nes_psg_c1b = psg_eventqueue[event].data;
		    break;
		case 2:
		    nes_psg_c1c = psg_eventqueue[event].data;
		    freq = ((((unsigned short)nes_psg_c1d & 0x07) << 8) + nes_psg_c1c);
		    nes_psg_c1_atl = nes_psg_atl[(nes_psg_c1d & 0xf8) >> 3];
		    if (freq) {
			nes_psg_c1_skip = nes_psg_pulse_magic / freq;
		    } else {
			nes_psg_c1_skip = 0;
		    }
		    break;
		case 3:
		    nes_psg_c1d = psg_eventqueue[event].data;
		    freq = ((((unsigned short)nes_psg_c1d & 0x07) << 8) + nes_psg_c1c);
		    nes_psg_c1_atl = nes_psg_atl[(nes_psg_c1d & 0xf8) >> 3];
		    if (freq) {
			nes_psg_c1_skip = nes_psg_pulse_magic / freq;
		    } else {
			nes_psg_c1_skip = 0;
		    }
		}
	    }
	    
	    event++;
	}
	if ((nes_psg_c1_atl) || (nes_psg_c1a & 0x20)) {
	    nes_psg_c1_index += nes_psg_c1_skip;
	    nes_psg_c1_index &= 0x1fffffff;
	    wave_buffers[0][i] = nes_psg_c1_wave[nes_psg_c1_index >> 24] * nes_psg_c1_vol;
	} else {
	    wave_buffers[0][i] = 0;
	}
    }
    if (nes_psg_c1_atl) {
	nes_psg_c1_atl--;
    }
}

void nes_psg_wave_2(void)
{
    int i;
    int cycles;
    int event;
    unsigned int freq;
    
    cycles = 0;
    event = 0;
    
    /* note: 41 CPU cycles occur between increments of i */
    for (i = 0; i < nes_psg_samples_per_sync; i++) {
	cycles += nes_psg_cycles_per_sample;
	while ((event < cur_event) && (psg_eventqueue[event].time < cycles)) {
	    if ((psg_eventqueue[event].type & PSGET_MASK) == PSGET_C2) {
		switch (psg_eventqueue[event].type & 3) {
		case 0:
		    nes_psg_c2a = psg_eventqueue[event].data;
		    nes_psg_c2_wave = pulse_waves[(nes_psg_c2a >> 6)];
		    nes_psg_c2_vol = nes_psg_c2a & 0x0f;
		    break;
		case 1:
		    nes_psg_c2b = psg_eventqueue[event].data;
		    break;
		case 2:
		    nes_psg_c2c = psg_eventqueue[event].data;
		    freq = ((((unsigned short)nes_psg_c2d & 0x07) << 8) + nes_psg_c2c);
		    nes_psg_c2_atl = nes_psg_atl[(nes_psg_c2d & 0xf8) >> 3];
		    if (freq) {
			nes_psg_c2_skip = nes_psg_pulse_magic / freq;
		    } else {
			nes_psg_c2_skip = 0;
		    }
		    break;
		case 3:
		    nes_psg_c2d = psg_eventqueue[event].data;
		    freq = ((((unsigned short)nes_psg_c2d & 0x07) << 8) + nes_psg_c2c);
		    nes_psg_c2_atl = nes_psg_atl[(nes_psg_c2d & 0xf8) >> 3];
		    if (freq) {
			nes_psg_c2_skip = nes_psg_pulse_magic / freq;
		    } else {
			nes_psg_c2_skip = 0;
		    }
		}
	    }
	    
	    event++;
	}
	if ((nes_psg_c2_atl) || (nes_psg_c2a & 0x20)) {
	    nes_psg_c2_index += nes_psg_c2_skip;
	    nes_psg_c2_index &= 0x1fffffff;
	    wave_buffers[1][i] = nes_psg_c2_wave[nes_psg_c2_index >> 24] * nes_psg_c2_vol;
	} else {
	    wave_buffers[1][i] = 0;
	}
    }
    if (nes_psg_c2_atl) {
	nes_psg_c2_atl--;
    }
}

void nes_psg_wave_3(void)
{
    int i;
    int cycles;
    int event;
    unsigned int freq;
    
    cycles = 0;
    event = 0;
    
    /* note: 41 CPU cycles occur between increments of i */
    for (i = 0; i < nes_psg_samples_per_sync; i++) {
	cycles += nes_psg_cycles_per_sample;
	while ((event < cur_event) && (psg_eventqueue[event].time < cycles)) {
	    if ((psg_eventqueue[event].type & PSGET_MASK) == PSGET_C3) {
		switch (psg_eventqueue[event].type & 3) {
		case 0:
		    nes_psg_c3a = psg_eventqueue[event].data;
		    nes_psg_c3_vol = nes_psg_c3a & 0x0f;
		    break;
		case 1:
		    nes_psg_c3b = psg_eventqueue[event].data;
		    break;
		case 2:
		    nes_psg_c3c = psg_eventqueue[event].data;
		    freq = ((((unsigned short)nes_psg_c3d & 0x07) << 8) + nes_psg_c3c);
		    nes_psg_c3_atl = nes_psg_atl[(nes_psg_c3d & 0xf8) >> 3];
		    if (freq) {
			nes_psg_c3_skip = nes_psg_triangle_magic / freq;
		    } else {
			nes_psg_c3_skip = 0;
		    }
		    break;
		case 3:
		    nes_psg_c3d = psg_eventqueue[event].data;
		    freq = ((((unsigned short)nes_psg_c3d & 0x07) << 8) + nes_psg_c3c);
		    nes_psg_c3_atl = nes_psg_atl[(nes_psg_c3d & 0xf8) >> 3];
		    if (freq) {
			nes_psg_c3_skip = nes_psg_triangle_magic / freq;
		    } else {
			nes_psg_c3_skip = 0;
		    }
		}
	    }
	    
	    event++;
	}
	if (((nes_psg_c2_atl) || (nes_psg_c2a & 0x20)) && nes_psg_c3_vol) {
	    nes_psg_c3_index += nes_psg_c3_skip;
	    nes_psg_c3_index &= 0x1fffffff;
	    wave_buffers[2][i] = triangle_50[nes_psg_c3_index >> 24];
	} else {
	    wave_buffers[2][i] = 0;
	}
    }
    if (nes_psg_c3_atl) {
	nes_psg_c3_atl--;
    }
}

void nes_psg_wave_4(void)
{
    int i;
    int cycles;
    int event;
    
    cycles = 0;
    event = 0;
    
    for (i = 0; i < nes_psg_samples_per_sync; i++) {
	cycles += nes_psg_cycles_per_sample;
	while ((event < cur_event) && (psg_eventqueue[event].time < cycles)) {
	    if ((psg_eventqueue[event].type & PSGET_MASK) == PSGET_C4) {
		switch (psg_eventqueue[event].type & 3) {
		case 0:
		    nes_psg_c4a = psg_eventqueue[event].data;
		    nes_psg_c4_vol = nes_psg_c4a & 0x0f;
		    nes_psg_c4_vol |= nes_psg_c4_vol << 4;
/* 		    deb_printf("psg: c4a: vol %d.\n", nes_psg_c4a & 15); */
		    break;
		case 1:
		    nes_psg_c4b = psg_eventqueue[event].data;
/* 		    deb_printf("psg: c4b: 0x%02x.\n", nes_psg_c4b); */
		    break;
		case 2:
		    nes_psg_c4c = psg_eventqueue[event].data;
		    if (nes_psg_c4c & 0x80) {
			nes_psg_c4_sr = 0x001f;
		    } else {
			nes_psg_c4_sr = 0x01ff;
		    }
/* 		    if (nes_psg_c4c & 0x70) { */
/* 			deb_printf("psg: c4c: c4c & 0x70 == 0x%02x.\n", nes_psg_c4c & 0x70); */
/* 		    } */
		    if (nes_psg_c4c & 15) {
			nes_psg_c4_skip = nes_psg_noise_magic / (nes_psg_c4c & 15);
		    } else {
			nes_psg_c4_skip = 0;
		    }
		    nes_psg_c4_atl = nes_psg_atl[(nes_psg_c4d & 0xf8) >> 3];
/* 		    deb_printf("psg: c4: freq: %d.\n", nes_psg_c4c & 15); */
		    break;
		case 3:
		    nes_psg_c4d = psg_eventqueue[event].data;
		    if (nes_psg_c4c & 15) {
			nes_psg_c4_skip = nes_psg_noise_magic / (nes_psg_c4c & 15);
		    } else {
			nes_psg_c4_skip = 0;
		    }
		    nes_psg_c4_atl = nes_psg_atl[(nes_psg_c4d & 0xf8) >> 3];
		}
	    }
	    
	    event++;
	}
	nes_psg_c4_index += nes_psg_c4_skip;
	if (nes_psg_c4_index > 0x00ffffff) {
	    if (nes_psg_c4c & 0x80) { /* FIXME: may be wrong */
		nes_psg_c4_sr |= ((!(nes_psg_c4_sr & 1)) ^ (!(nes_psg_c4_sr & 4))) << 5;
	    } else {
		nes_psg_c4_sr |= ((!(nes_psg_c4_sr & 1)) ^ (!(nes_psg_c4_sr & 16))) << 9;
	    }
	    nes_psg_c4_sr >>= 1;
	}
	nes_psg_c4_index &= 0x00ffffff;
#if 0
	if (nes_psg_c4_atl && (nes_psg_c4_sr & 1)) {
	    wave_buffers[3][i] = nes_psg_c4_vol;
	} else {
	    wave_buffers[3][i] = 0;
	}
#else
	wave_buffers[3][i] = 0;
#endif
    }
    if (nes_psg_c4_atl) {
	nes_psg_c4_atl--;
    }
}

void nes_psg_mix(unsigned char *final_wave)
{
    int i;
    int cycles;
    int event;
    int elapsed;
    
    cycles = 0;
    elapsed = 0;
    event = 0;

    /* note: 41 CPU cycles occur between increments of i */
    for (i = 0; i < nes_psg_samples_per_sync; i++) {
	cycles += nes_psg_cycles_per_sample;
	while ((event < cur_event) && (psg_eventqueue[event].time < cycles)) {
	    if (psg_eventqueue[event].type == PSGET_W_CTRL) {
		nes_psg_ctrl = psg_eventqueue[event].data;
	    }
	    event++;
	}
	/* FIXME: there has to be a better way to do this. */
	if (!(nes_psg_ctrl & 1)) {
	    wave_buffers[0][i] = 0;
	}
	if (!(nes_psg_ctrl & 2)) {
	    wave_buffers[1][i] = 0;
	}
	if (!(nes_psg_ctrl & 4)) {
	    wave_buffers[2][i] = 0;
	}
	if (!(nes_psg_ctrl & 8)) {
	    wave_buffers[3][i] = 0;
	}
#ifdef SOUND_DOS
	final_wave[i] = (wave_buffers[0][i] + wave_buffers[1][i] +
			 wave_buffers[2][i] + wave_buffers[3][i]) >> 2;
#endif
#ifdef SOUND_LINUX
	final_wave[waveptr] = (wave_buffers[0][i] + wave_buffers[1][i] +
			       wave_buffers[2][i] + wave_buffers[3][i]) >> 2;
	waveptr++;
	if (waveptr == 2048) {
	    waveptr = 0;
	    wavflag = 2;
	} else if (waveptr == 1024) {
	    wavflag = 1;
	}
#endif
	elapsed++;
    }
}

#endif

void nes_psg_vsync(void)
{
#ifdef SOUND_DOS
    unsigned char *final_wave = NULL;
#endif
    
#ifdef SOUND_LINUX
    nes_psg_wave_1();
    nes_psg_wave_2();
    nes_psg_wave_3();
    nes_psg_wave_4();
    
    nes_psg_mix(final_wave);

    if (wavflag) {
	if (sound_fd) {
	    if (write(sound_fd, &final_wave[(wavflag - 1) << 10], 1024) < 1024) {
		deb_printf("wrote less than 1024 bytes\n");
	    }
	}
	wavflag = 0;
    }
#endif
    
#ifdef SOUND_DOS
    while (!final_wave) { /* Wheee... Primitive Framesync. :-) */
	final_wave = get_audio_stream_buffer(sound_stream);
    }
    
    nes_psg_wave_1();
    nes_psg_wave_2();
    nes_psg_wave_3();
    nes_psg_wave_4();

    nes_psg_mix(final_wave);
    
    free_audio_stream_buffer(sound_stream);
#endif
    
    lasttime = 0;
    cur_event = 0;
}

#ifdef SOUND_LINUX
void nes_psg_open_sound_linux(void)
{
    int tmp;
    int result;
    int sound_rate;
    int sound_frag;

    waveptr = 0;
    wavflag = 0;
    
    printf("opening /dev/dsp...");
    sound_fd = open("/dev/dsp", O_WRONLY | O_NONBLOCK);
    if (sound_fd < 0) {
	perror("failed");
	sound_fd = 0;
	return;
    } else {
	printf("done.\n");
    }
    
    printf("setting unsigned 8-bit format...");
    tmp = AFMT_U8;
    result = ioctl(sound_fd, SNDCTL_DSP_SETFMT, &tmp);
    if (result < 0) {
	perror("failed");
	close(sound_fd);
	sound_fd = 0;
	return;
    } else {
	printf("done.\n");
    }
    
    printf("setting mono mode...");
    tmp = 0;
    result = ioctl(sound_fd, SNDCTL_DSP_STEREO, &tmp);
    if (result < 0) {
	perror("failed");
	close(sound_fd);
	sound_fd = 0;
	return;
    } else {
	printf("done.\n");
    }
    
    sound_rate = nes_psg_sample_rate;
    printf("setting sound rate to %dHz...", sound_rate);
    result = ioctl(sound_fd, SNDCTL_DSP_SPEED, &sound_rate);
    if (result < 0) {
	perror("failed");
	close(sound_fd);
	sound_fd = 0;
	return;
    } else {
	printf("got %dHz...done.\n", sound_rate);
    }

    /* high word of sound_frag is number of frags, low word is frag size */
    sound_frag = 0x00080008;
    printf("setting soundfrags...");
    result = ioctl(sound_fd, SNDCTL_DSP_SETFRAGMENT, &sound_frag);
    if (result < 0) {
	perror("failed");
	close(sound_fd);
	sound_fd = 0;
	return;
    } else {
	printf("done.\n");
    }
}
#endif

#ifdef SOUND_DOS
void nes_psg_open_sound_dos(void)
{
    if (install_sound(DIGI_AUTODETECT, MIDI_NONE, NULL) != 0) {
	deb_printf("Error initializing sound system\n%s\n", allegro_error);
    } else {
	sound_stream = play_audio_stream(nes_psg_samples_per_sync, 8, nes_psg_sample_rate, 255, 128);
	if (!sound_stream) {
	    deb_printf("unable to create audio stream.\n");
	}
    }
}
#endif

void nes_psg_init(void)
{
#ifdef SOUND
    if (nes_psg_quality > 0) {
	if (nes_psg_quality > 2) {
	    nes_psg_quality = 1;
	} else {
	    nes_psg_quality--;
	}
	nes_psg_pulse_magic = nes_psg_qual[nes_psg_quality].pulse_magic;
	nes_psg_triangle_magic = nes_psg_qual[nes_psg_quality].triangle_magic;
	nes_psg_noise_magic = nes_psg_qual[nes_psg_quality].noise_magic;
	nes_psg_samples_per_sync = nes_psg_qual[nes_psg_quality].samples_per_sync;
	nes_psg_cycles_per_sample = nes_psg_qual[nes_psg_quality].cycles_per_sample;
	nes_psg_sample_rate = nes_psg_qual[nes_psg_quality].sample_rate;
	
#ifdef SOUND_LINUX
	nes_psg_open_sound_linux();
#endif
    
#ifdef SOUND_DOS
	nes_psg_open_sound_dos();
#endif

	nes_psg_c1_wave = pulse_50;
	nes_psg_c2_wave = pulse_50;
    } else {
#ifdef SOUND_LINUX
	sound_fd = 0;
#endif
    
#ifdef SOUND_DOS
	sound_stream = NULL;
#endif
    }
#endif
    lasttime = 0;
    cur_event = 0;
}

void nes_psg_done(void)
{
#ifdef SOUND_LINUX
    if (sound_fd) {
	close(sound_fd);
    }
#endif
    
#ifdef SOUND_DOS
    stop_audio_stream(sound_stream);
#endif
}

/*
 * $Log: nes_psg.c,v $
 * Revision 1.33  1999/02/14 18:27:54  nyef
 * added a function pointer array for writing the sound regs
 *
 * Revision 1.32  1999/01/17 04:21:11  nyef
 * fixed some problems with disabling sound completely
 *
 * Revision 1.31  1998/12/28 04:26:18  nyef
 * added simple fix for "long note" problem on wave 3.
 * full volume support on wave 3 is nessecary for proper fix.
 *
 * Revision 1.30  1998/12/28 04:10:50  nyef
 * added disabled preliminary wave 4 generation.
 *
 * Revision 1.29  1998/12/27 04:57:34  nyef
 * added basic framework for wave 4 generation.
 *
 * Revision 1.28  1998/12/27 01:57:18  nyef
 * added support for volume control and looped sounds.
 *
 * Revision 1.27  1998/12/17 06:22:19  nyef
 * cleaned up wave generation functions slightly.
 *
 * Revision 1.26  1998/12/12 20:33:07  nyef
 * commented out "DONT_USE_CAL" as final step in bringing up to spec.
 *
 * Revision 1.25  1998/12/12 02:46:42  nyef
 * fixed bug in nes_psg_open_sound_dos().
 *
 * Revision 1.24  1998/12/06 04:15:56  nyef
 * added preliminary sound quality (playback rate) adjustment.
 *
 * Revision 1.23  1998/12/05 23:32:14  nyef
 * halved the frequency on the triangle waves.
 *
 * Revision 1.22  1998/12/05 18:43:03  nyef
 * active time length counters improved. Dungeon Magic sounds much better.
 *
 * Revision 1.21  1998/12/05 18:23:51  nyef
 * changed from 16.16 fixed point to 8.24 fixed point. Due to some loss of
 * precision in the conversion, the "sound constant" needs to be recalculated.
 *
 * Revision 1.20  1998/12/05 05:25:48  nyef
 * added pulse wave duty cycles. don't know how well they work.
 *
 * Revision 1.19  1998/12/04 02:55:10  nyef
 * started to bring io write functions up to spec. maintaining backward
 * compatability for ease of transition and for RockNES.
 *
 * Revision 1.18  1998/12/04 02:08:36  nyef
 * changed wave buffer 'triangle' to 'triangle_50' because allegro already
 * defines 'triangle'.
 *
 * Revision 1.17  1998/12/03 04:58:59  nyef
 * major overhaul. changed sound rendering method. system sounds a lot
 * better. channel 3 approaches triangularity for sufficiently large values
 * of 32. some hooks for duty cycles in place. active time length counters
 * improved. floating point stuff removed. improved event handling.
 *
 * Revision 1.16  1998/12/01 04:01:12  nyef
 * obtained speedup by storing the value (44100 / 111860.78) * freq in a
 * local variable instead of calculating it every time it was needed.
 *
 * Revision 1.15  1998/12/01 03:34:15  nyef
 * more doc fixes (mainly adding FIXMEs). added preliminary active time left
 * support (I don't know if I did it right). more bugfixes.
 *
 * Revision 1.14  1998/12/01 02:57:48  nyef
 * doc fix. removed useless debug code from psg write functions. added
 * preliminary wave 3 support (needs to be changed to triangle wave).
 * fixed a few bugs in the wave functions.
 *
 * Revision 1.13  1998/11/28 05:06:40  nyef
 * bashed up linux sound stuff again. reintroduced soundfrags. changed to
 * use 1024 byte writes with dual local buffers.
 *
 * Revision 1.12  1998/11/26 01:39:54  nyef
 * gratuitous changes. mainly linux latency stuff.
 *
 * Revision 1.11  1998/11/05 02:58:31  nyef
 * added preliminary emulation of psg control register.
 *
 * Revision 1.10  1998/11/05 02:29:34  nyef
 * integrated new dos code changes. sound output now works on dos.
 *
 * Revision 1.9  1998/11/04 03:25:45  nyef
 * moved the code to open the sound device from nes_psg_init() to
 * nes_psg_open_sound_linux().
 *
 * Revision 1.8  1998/11/04 03:20:16  nyef
 * added (commented out) defines for SOUND_LINUX and SOUND_DOS.
 * made inclusion of system header files required for compile with
 * sound support conditional upon definition of SOUND_LINUX.
 *
 * Revision 1.7  1998/11/04 03:10:08  nyef
 * changed all psg event type defines to hexadecimal.
 * added psg event type PSGET_SYNC.
 *
 * Revision 1.6  1998/11/03 00:43:26  nyef
 * Added preliminary wave 2 generation.
 *
 * Revision 1.5  1998/10/21 00:15:33  nyef
 * commented out the define for sound.
 * changed some of the debug output in nes_psg_wave_1().
 *
 * Revision 1.4  1998/09/19 00:48:25  nyef
 *   wave 1 generation now working (the video display and rendering must be
 * disabled for nes_ppu.c for it not to skip on my machine). started
 * instrumenting wave 1 control regs for furthur insight into wave generation.
 *
 * Revision 1.3  1998/09/02 01:25:40  nyef
 * added test output of wave 1. wave 1 generation still not working.
 *
 * Revision 1.2  1998/08/30 20:37:57  nyef
 * added event queue, filled in the write functions, and added vsync hook.
 *
 * Revision 1.1  1998/08/22 00:35:43  nyef
 * Initial revision
 *
 */
