/*
 * opc6502.h
 *
 * 6502 opcode emulation macros
 */

/* $Id: opc6502.h,v 1.8 1999/06/14 16:50:38 nyef Exp $ */

#ifndef __OPC6502_H__
#define __OPC6502_H__

#include "machdep.h"

#define reg_A       0x0(%ebp)
#define reg_Y       0x2(%ebp)
#define reg_X       0x4(%ebp)
#define reg_P       0x6(%ebp)
#define reg_S       0x7(%ebp)
#define reg_PC      0xb(%ebp)
#define readmap     0xd(%ebp)
#define writemap    0x11(%ebp)
#define decodemap   0x15(%ebp)
#define dispatch    0x19(%ebp)
#define cycles_left 0x1d(%ebp)
#define page_0      0x21(%ebp)
#define mmu_shift   0x25(%ebp)
#define mmu_mask    0x29(%ebp)
#define user        0x2d(%ebp)

    CYC_ADDMODE = 0
    CYC_OP = 1
#define Continue subl $(CYC_ADDMODE+CYC_OP), cycles_left; jmp *dispatch; CYC_ADDMODE = 0; CYC_OP = 1

#define flag_N_bit $0xf
#define flag_Z_bit $0xe
#define flag_I_bit $0xd
#define flag_D_bit $0xc
#define flag_V_bit $0xb
#define flag_C_bit $0x8

#define flag_N $0x80
#define flag_Z $0x40
#define flag_I $0x20 /* foo? */
#define flag_D $0x10 /* foo? */
#define flag_V $0x08 /* foo? */
#define flag_C $0x01

#define flag_NZ $0xc0
#define flag_NZC $0xc1
#define flag_NZVC $0xc9

#define flag_N_not $0x7f
#define flag_Z_not $0xbf
#define flag_I_not $0xdf /* foo? */
#define flag_D_not $0xef /* foo? */
#define flag_V_not $0xf7 /* foo? */
#define flag_C_not $0xfe

#define flag_NZ_not $0x3f
#define flag_NZC_not $0x3e
#define flag_NZVC_not $0x36
#define flag_NV_not $0x77

/* emulation register map */
/* %edi = effective address */
/* %esi = memaccess scratch */ /* used only in the GetInd(X/Y) macros */
/* %edx = PC */
/* %cl = A */
/* %ch = P */
/* %bl = X */
/* %bh = Y */
/* %ax = opcode scratch */

#define FlagZ lahf; andb flag_Z_not, %ch; andb flag_Z, %ah; orb %ah, %ch
#define FlagNZ lahf; andb flag_NZ_not, %ch; andb flag_NZ, %ah; orb %ah, %ch
#define FlagNZC lahf; andb flag_NZC_not, %ch; andb flag_NZC, %ah; orb %ah, %ch
#define FlagNZVC lahf; jno 0f; orb flag_V, %ah; 0: andb flag_NZVC_not, %ch; andb flag_NZVC, %ah; orb %ah, %ch

/* #define MemReadNorm movl readmap, %esi; call (%esi,%edi,4) */
#define MemReadNorm pushl %edx; pushl %ecx; pushl %eax; pushl %edi; pushl user; movl mmu_shift, %ecx; movl mmu_mask, %edx; shrl %cl, %edi; andl %edx, %edi; movl readmap, %esi; call *(%esi,%edi,4); movb %al, 8(%esp); popl %edi; popl %edi; popl %eax; popl %ecx; popl %edx
#define MemReadZPage movl page_0, %esi; movb (%esi,%edi,1), %al
#define MemReadW(x) incw %di; MemRead##x; decw %di; shlw $8, %ax; MemRead##x

/* #define MemWriteNorm movl writemap, %esi; call (%esi,%edi,4) */
#define MemWriteNorm pushl %edx; pushl %ecx; pushl %eax; pushl %edi; pushl user; movl mmu_shift, %ecx; movl mmu_mask, %ebx; shrl %cl, %edi; andl %ebx, %edi; movl writemap, %esi; call *(%esi,%edi,4); popl %edi; popl %edi; popl %eax; popl %ecx; popl %edx
#define MemWriteZPage movl page_0, %esi; movb %al, (%esi,%edi,1)

#define GetFromPC_B movl %edx, %edi; incw %dx
#define GetFromPC_W movl %edx, %edi; incw %dx; incw %dx

#define GetImm GetFromPC_B; CYC_ADDMODE = 0
#define GetAbs GetFromPC_W; MemReadW(Norm); movw %ax, %di; CYC_ADDMODE = 2
#define GetAbsX GetFromPC_W; MemReadW(Norm); addb reg_X, %al; setc cycle_temp; adcb $0, %ah; movw %ax, %di; movl cycle_temp, %eax; subl %eax, cycles_left; CYC_ADDMODE = 2
#define GetAbsY GetFromPC_W; MemReadW(Norm); addb reg_Y, %al; setc cycle_temp; adcb $0, %ah; movw %ax, %di; movl cycle_temp, %eax; subl %eax, cycles_left; CYC_ADDMODE = 2
#define GetIndX xorl %eax, %eax; GetFromPC_B; MemReadNorm; addb reg_X, %al; movl page_0, %esi; movw (%eax, %esi), %ax; movl %eax, %edi; CYC_ADDMODE = 4
#define GetIndY xorl %eax, %eax; GetFromPC_B; MemReadNorm; movl page_0, %esi; movw (%eax, %esi), %ax; addb reg_Y, %al; setc cycle_temp; adcb $0, %ah; movl %eax, %edi; movl cycle_temp, %eax; subl %eax, cycles_left; CYC_ADDMODE = 3
#define GetZPage GetFromPC_B; MemReadNorm; xorb %ah, %ah; movw %ax, %di; CYC_ADDMODE = 1
#define GetZPageX GetFromPC_B; MemReadNorm; xorb %ah, %ah; addb reg_X, %al; movw %ax, %di; CYC_ADDMODE = 2
#define GetZPageY GetFromPC_B; MemReadNorm; xorb %ah, %ah; addb reg_Y, %al; movw %ax, %di; CYC_ADDMODE = 2

#define BxC(cond) GetFromPC_B; MemReadNorm; bt cond, %ecx; cbw; jc 0f; addw %ax, %dx; decl cycles_left; 0:
#define BxS(cond) GetFromPC_B; MemReadNorm; bt cond, %ecx; cbw; jnc 0f; addw %ax, %dx; decl cycles_left; 0:

#define DoLDA(x) MemRead##x; movb %al, reg_A; orb %al, %al; FlagNZ
#define DoLDX(x) MemRead##x; movb %al, reg_X; orb %al, %al; FlagNZ
#define DoLDY(x) MemRead##x; movb %al, reg_Y; orb %al, %al; FlagNZ

#define DoSTA(x) movb reg_A, %al; MemWrite##x
#define DoSTX(x) movb reg_X, %al; MemWrite##x
#define DoSTY(x) movb reg_Y, %al; MemWrite##x

#define DoDEX decb reg_X; FlagNZ
#define DoDEY decb reg_Y; FlagNZ

#define DoINX incb reg_X; FlagNZ
#define DoINY incb reg_Y; FlagNZ

#define DoINC(x) MemRead##x; incb %al; FlagNZ; MemWrite##x; CYC_OP = 3
#define DoDEC(x) MemRead##x; decb %al; FlagNZ; MemWrite##x; CYC_OP = 3

#define DoTXS movb reg_X, %al; movb %al, reg_S
#define DoTSX movb reg_S, %al; movb %al, reg_X; orb %al, %al; FlagNZ

#define DoTAX movb reg_A, %al; movb %al, reg_X; orb %al, %al; FlagNZ
#define DoTAY movb reg_A, %al; movb %al, reg_Y; orb %al, %al; FlagNZ

#define DoTXA movb reg_X, %al; movb %al, reg_A; orb %al, %al; FlagNZ
#define DoTYA movb reg_Y, %al; movb %al, reg_A; orb %al, %al; FlagNZ

#define DoADCb bt flag_C_bit, %ecx; adcb %al, reg_A; FlagNZVC
#define DoADCd bt flag_C_bit, %ecx; adcb reg_A, %al; daa; movb %al, reg_A; FlagNZVC
#define DoSBCb bt flag_C_bit, %ecx; cmc; sbbb %al, reg_A; cmc; FlagNZVC
#define DoSBCd bt flag_C_bit, %ecx; cmc; sbbb %al, reg_A; xchgb %al, reg_A; das; movb %al, reg_A; cmc; FlagNZVC

#define DoBIT(x) MemRead##x; movb %al, %ah; andb reg_A, %ah; FlagZ; andl $0xc0, %eax; movb SYMNAME(flags_to_8088)(%eax), %al; andb flag_NV_not, %ch; orb %al, %ch
#define DoEOR(x) MemRead##x; xorb %al, reg_A; FlagNZ
#define DoADC(x) MemRead##x; bt flag_D_bit, %ecx; jc 1f; DoADCb; jmp 2f; 1: DoADCd; 2:
#define DoSBC(x) MemRead##x; bt flag_D_bit, %ecx; jc 1f; DoSBCb; jmp 2f; 1: DoSBCd; 2:
#define DoAND(x) MemRead##x; andb %al, reg_A; FlagNZ
#define DoORA(x) MemRead##x; orb %al, reg_A; FlagNZ

#define DoCMP(x) MemRead##x; cmpb %al, reg_A; cmc; FlagNZC
#define DoCPX(x) MemRead##x; cmpb %al, reg_X; cmc; FlagNZC
#define DoCPY(x) MemRead##x; cmpb %al, reg_Y; cmc; FlagNZC

#define DoASLa clc; rclb $1, reg_A; FlagNZC; movb reg_A, %al; orb %al, %al; FlagNZ
#define DoASL(x) MemRead##x; clc; rclb $1, %al; FlagNZC; orb %al, %al; FlagNZ; MemWrite##x; CYC_OP = 3
#define DoLSRa clc; rcrb $1, reg_A; FlagNZC; movb reg_A, %al; orb %al, %al; FlagNZ
#define DoLSR(x) MemRead##x; clc; rcrb $1, %al; FlagNZC; orb %al, %al; FlagNZ; MemWrite##x; CYC_OP = 3
#define DoROLa bt flag_C_bit, %ecx; rclb $1, reg_A; FlagNZC; movb reg_A, %al; orb %al, %al; FlagNZ
#define DoROL(x) MemRead##x; bt flag_C_bit, %ecx; rclb $1, %al; FlagNZC; orb %al, %al; FlagNZ; MemWrite##x; CYC_OP = 3
#define DoRORa bt flag_C_bit, %ecx; rcrb $1, reg_A; FlagNZC; movb reg_A, %al; orb %al, %al; FlagNZ
#define DoROR(x) MemRead##x; bt flag_C_bit, %ecx; rcrb $1, %al; FlagNZC; orb %al, %al; FlagNZ; MemWrite##x; CYC_OP = 3

#define push(x) movl page_0, %edi; addl reg_S, %edi; movb x, (%edi); decb reg_S
#define pop(x) incb reg_S; movl page_0, %edi; addl reg_S, %edi; movb (%edi), x

#endif /* __OPC6502_H__ */

/*
 * $Log: opc6502.h,v $
 * Revision 1.8  1999/06/14 16:50:38  nyef
 * fixed to compile clean with GNU as 2.9.1
 *
 * Revision 1.7  1999/05/23 16:22:39  nyef
 * fixed decimal mode to work properly
 *
 * Revision 1.6  1998/12/12 21:34:05  nyef
 * really finished converting to the new memory access spec
 *
 * Revision 1.5  1998/11/29 18:26:43  nyef
 * finished converting to the new memory access spec.
 *
 * Revision 1.4  1998/11/29 17:19:01  nyef
 * started to convert to the new memory access spec.
 *
 * Revision 1.3  1998/11/01 23:29:33  nyef
 * fixed all ZPage addressing modes to use direct ZPage access.
 *
 * Revision 1.2  1998/07/26 23:25:24  nyef
 * changed ZPage, ZPageX and ZPageY macros to move the zpage outside of the
 *    normal address range so that they don't cause I/O accesses on the PCE
 *
 * Revision 1.1  1998/07/11 22:18:54  nyef
 * Initial revision
 *
 */
