/*
 * sms.c
 *
 * Sega Master System emulation.
 */

/* $Id: sms.c,v 1.27 1999/07/01 01:26:37 nyef Exp $ */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "ui.h"
#include "system.h"
#include "cal.h"
#include "tool.h"
/* #include "emuz80.h" */
/* #include "mz80.h" */
#include "sms9918.h"
#include "sms_psg.h"

#define SMS_RAMSIZE 0xa000 /* 0x2000 ram + 0x8000 cart ram */

unsigned char *sms_memory;
cal_cpu sms_cpu;
sms9918 sms_vdp;

unsigned char *sms_rombase;
unsigned char *sms_bank0;
unsigned char *sms_bank1;
unsigned char *sms_bank2;
unsigned char *sms_banks[8];
int sms_romsize;
int sms_uses_battery;
char *battery_filename;

const struct joypad_template sms_joypad_template = {
    6, {0x01, 0x02, 0x04, 0x08, 0x10, 0x20}
};

const struct joypad_template gg_joypad_template = {
    7, {0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x80} /* has pause button */
};

struct joypad sms_joypad_1 = {
    &sms_joypad_template, 0
};

unsigned char sms_pageregs[4];

int is_gg;
int last_pausestate;

void sms_pagesync(void)
{
    sms_bank0 = &sms_rombase[(sms_pageregs[1] << 14) & (sms_romsize - 1)];
    sms_bank1 = &sms_rombase[(sms_pageregs[2] << 14) & (sms_romsize - 1)];
    if (sms_pageregs[0] & 0x08) {
	sms_bank2 = &sms_memory[0x2000 + ((sms_pageregs[0] & 0x04) << 12)];
	sms_uses_battery = 1;
    } else {
	sms_bank2 = &sms_rombase[(sms_pageregs[3] << 14) & (sms_romsize - 1)];
    }

    sms_banks[0] = sms_bank0;
    sms_banks[1] = sms_bank0 + 0x2000;
    sms_banks[2] = sms_bank1;
    sms_banks[3] = sms_bank1 + 0x2000;
    sms_banks[4] = sms_bank2;
    sms_banks[5] = sms_bank2 + 0x2000;
    sms_banks[6] = sms_memory;
    sms_banks[7] = sms_memory;
    
}

unsigned char sms_read_rom0(cal_cpu cpu, unsigned long addr)
{
    if (addr < 0x400) {
	return sms_rombase[addr];
    } else {
	return sms_bank0[addr];
    }
}

unsigned char sms_read_rom1(cal_cpu cpu, unsigned long addr)
{
    return sms_bank1[addr & 0x3fff];
}

unsigned char sms_read_rom2(cal_cpu cpu, unsigned long addr)
{
    return sms_bank2[addr & 0x3fff];
}

unsigned char sms_read_ram(cal_cpu cpu, unsigned long addr)
{
    return sms_memory[addr & 0x1fff];
}

unsigned char sms_read(cal_cpu cpu, unsigned long addr)
{
    if (addr < 0x400) {
	return sms_rombase[addr];
    } else if (addr < 0x4000) {
	return sms_bank0[addr];
    } else if (addr < 0x8000) {
	return sms_bank1[addr & 0x3fff];
    } else if (addr < 0xc000) {
	return sms_bank2[addr & 0x3fff];
    } else {
	return sms_memory[addr & 0x1fff];
    }
}

void sms_write_ignore(cal_cpu cpu, unsigned long addr, unsigned char data)
{
}

void sms_write_rom2(cal_cpu cpu, unsigned long addr, unsigned char data)
{
    if (sms_pageregs[0] & 0x08) {
	sms_bank2[addr & 0x3fff] = data;
    }
}

void sms_write_ram(cal_cpu cpu, unsigned long addr, unsigned char data)
{
    if (addr >= 0xfffc) {
	sms_pageregs[addr & 3] = data;
	sms_pagesync();
    }
    sms_memory[addr & 0x1fff] = data;
}

void sms_write(cal_cpu cpu, unsigned long addr, unsigned char data)
{
    if (addr >= 0xc000) {
	if (addr >= 0xfffc) {
	    sms_pageregs[addr & 3] = data;
	    sms_pagesync();
	}
	sms_memory[addr & 0x1fff] = data;
    } else if (addr > 0x8000) {
	if (sms_pageregs[0] & 0x08) {
	    sms_bank2[addr & 0x3fff] = data;
	}
    }
}

memread8_t sms_readtable[4] = {
    sms_read_rom0, sms_read_rom1, sms_read_rom2, sms_read_ram,
};

memwrite8_t sms_writetable[4] = {
    sms_write_ignore, sms_write_ignore, sms_write_rom2, sms_write_ram,
};

memread8_t sms_readfunc = &sms_read;
memwrite8_t sms_writefunc = &sms_write;

unsigned char sms_io_read(cal_cpu cpu, unsigned long address)
{
    if ((address == 0x00) && is_gg) {
	return sms_joypad_1.data & 0x80; /* simulated pause button */
    } else if ((address & 0xfe) == 0xbe) {
	if (address & 1) {
/* 	    deb_printf("sms: VDPstat read at PC 0x%04hx.\n", sms_cpu->data.d_maratz80->PC.W); */
	    return sms_vdp->readport1(sms_vdp);
	} else {
	    return sms_vdp->readport0(sms_vdp);
	}
    } else if (((address & 0xfe) == 0xdc) || ((address & 0xfe) == 0xc0)) {
	/* Dummy controller emulation */
	if (address & 1) {
	    return 0xff;
	} else {
	    ui_update_joypad(&sms_joypad_1);
	    return (sms_joypad_1.data) ^ 255;
	}
    } else if ((address & 0xfe) == 0x7e) {
	return sms_vdp->readscanline(sms_vdp);
    } else {
	deb_printf("sms: I/O read 0x%02x.\n", address & 0xff);
    }

    return 0;
}

void sms_io_write(cal_cpu cpu, unsigned long address, unsigned char data)
{
    if ((address & 0xfe) == 0xbe) {
	if (address & 1) {
	    sms_vdp->writeport1(sms_vdp, data);
	} else {
	    sms_vdp->writeport0(sms_vdp, data);
	}
    } else if ((address & 0xfe) == 0x7e) {
	sms_psg_write(data);
    } else if ((address & 0xfe) == 0xf0) {
	/* I don't like seeing the sound I/O */
    } else {
	deb_printf("sms: I/O write 0x%02x = 0x%02x.\n", address & 0xff, data);
    }
}

void sms_timeslice(cal_cpu cpu)
{
    int i;

    for (i = 262; i && (system_flags == F_NONE); i--) {
	cpu->runfor(cpu, 228); /* FIXME: this is .5 over what it should be. */
	cpu->run(cpu);
/* 	deb_printf("sms: PC 0x%04hx.\n", sms_cpu->data.d_maratz80->PC.W); */
	if (sms_vdp->periodic(sms_vdp)) {
	    cpu->irq(cpu);
	}
	if (system_flags & F_UNIMPL) {
	    deb_printf("Unimplimented Instruction.\n");
	    system_flags |= F_QUIT;
/* 	    deb_printf("PC: 0x%04hx, opcode: 0x%02x.\n", cpu->data.d_emuz80->reg_PC, sms_readtable[(cpu->data.d_emuz80->reg_PC >> 14) & 3](cpu, cpu->data.d_emuz80->reg_PC)); */
	}
    }
}

void sms_shutdown(void)
{
    FILE *batfile;
    
    if (sms_uses_battery) {
	deb_printf("attempting to save battery file.\n");

	/* FIXME: not robust enough. */
	batfile = fopen(battery_filename, "wb");
	
	if (batfile != NULL) {
	    fwrite(sms_memory + 0x2000, 0x4000, 1, batfile);
	    fclose(batfile);
	} else {
	    deb_printf("could not open battery file.\n");
	}
    }

    sms_psg_done();
}

void sms_run(int startup_deb)
{
    sms_memory = calloc(1, SMS_RAMSIZE);
    
    if (sms_memory && romimage) {
	if (romsize & 0x200) {
	    sms_rombase = romimage + 0x200;
	    sms_romsize = romsize - 0x200;
	} else {
	    sms_rombase = romimage;
	    sms_romsize = romsize;
	}

	if (sms_rombase[0x7ffd] < 0x50) {
	    deb_printf("file is probably a Game Gear image.\n");
	    is_gg = 1;
	} else {
	    deb_printf("file is probably a Master System image.\n");
	    is_gg = 0;
	}

	{
	    FILE *batfile;
	    
	    deb_printf("attempting to load battery file.\n");
	    
	    /* FIXME: this is not the most robust solution I've seen. */
	    battery_filename = strdup(romfilename);
	    *((unsigned long *) &battery_filename[strlen(battery_filename) - 4]) = 0x7661732e; /* ".sav" */
	    
	    batfile = fopen(battery_filename, "rb");
	    
	    if (batfile != NULL) {
		fread(sms_memory + 0x2000, 0x4000, 1, batfile);
		fclose(batfile);
	    } else {
		deb_printf("could not open battery file.\n");
	    }
	}
	
	last_pausestate = 0;
	
	sms_pageregs[0] = 0;
	sms_pageregs[1] = 0;
	sms_pageregs[2] = 1;
	sms_pageregs[3] = 2;
	sms_pagesync();

	sms_uses_battery = 0;
	
	sms_cpu = cal_create(CPUT_MARATZ80);
/* 	sms_cpu->setmmu8(sms_cpu, 14, 3, sms_readtable, sms_writetable); */
	sms_cpu->setmmu8(sms_cpu, 0, 0, &sms_readfunc, &sms_writefunc);
	sms_cpu->setiou(sms_cpu, sms_io_read, sms_io_write);
	sms_cpu->reset(sms_cpu);

	sms_vdp = sms9918_create(is_gg);

	nes_psg_quality = 2;
	sms_psg_init();

	dn_shutdown = sms_shutdown;

	if (is_gg) {
	    sms_joypad_1.template = &gg_joypad_template;
	}
	
	if (ui_register_joypad(&sms_joypad_1)) {
	    deb_printf("sms_run: joypad 1... registered.\n");
	} else {
	    deb_printf("sms_run: joypad 1... not registered.\n");
	}
	
	set_timeslice((void (*)(void *)) &sms_timeslice, sms_cpu);
    } else {
	deb_printf("out of memory.\n");
    }
}

/*
 * $Log: sms.c,v $
 * Revision 1.27  1999/07/01 01:26:37  nyef
 * added code to shutdown the sound system
 *
 * Revision 1.26  1999/06/06 17:50:47  nyef
 * changed to use the new mz80 memory interface
 *
 * Revision 1.25  1999/06/05 02:41:32  nyef
 * converted to use new joypad interface
 *
 * Revision 1.24  1999/05/02 18:27:19  nyef
 * added preliminary battery file support
 *
 * Revision 1.23  1999/05/01 23:57:49  nyef
 * added stuff to allow mz80 to directly access SMS memory
 *
 * Revision 1.22  1999/05/01 21:55:35  nyef
 * changed to quit on unimplimented instruction
 *
 * Revision 1.21  1999/05/01 21:55:02  nyef
 * fixed memory read/write routines
 *
 * Revision 1.20  1999/04/28 13:44:21  nyef
 * changed to add sound emulation
 *
 * Revision 1.19  1999/04/17 20:09:25  nyef
 * fixed for new version of the CAL.
 *
 * Revision 1.18  1999/03/12 02:13:18  nyef
 * fixed to tell vdp what system to emulate
 *
 * Revision 1.17  1999/02/23 02:57:28  nyef
 * removed aliasing of bankswitch regs to 0xdffc-0xdfff
 *
 * Revision 1.16  1999/02/15 03:26:08  nyef
 * added patch from Michael Vance to squash warnings
 *
 * Revision 1.15  1999/01/26 03:05:49  nyef
 * fixed to use 32bit "address bus"
 *
 * Revision 1.14  1999/01/14 02:18:43  nyef
 * fixed polarity on gg pause button.
 *
 * Revision 1.13  1999/01/13 02:48:31  nyef
 * added emulation of gg start button.
 *
 * Revision 1.12  1999/01/10 19:55:38  nyef
 * added aliasing of bankswitch regs to 0xdffc-0xdfff.
 * added detection of sms/gg images.
 *
 * Revision 1.11  1999/01/10 02:20:19  nyef
 * disabled debug logging of writes to I/O ports 0xf0 and 0xf1.
 *
 * Revision 1.10  1999/01/09 23:10:06  nyef
 * added support for reading the current scanline.
 *
 * Revision 1.9  1999/01/09 22:57:01  nyef
 * added aliased I/O ports for controller reads.
 *
 * Revision 1.8  1999/01/09 21:52:13  nyef
 * fixed stupid mistake in sms_pagesync().
 *
 * Revision 1.7  1999/01/08 19:35:58  nyef
 * fixed timing in sms_timeslice().
 *
 * Revision 1.6  1999/01/06 02:41:12  nyef
 * added code to prevent non-extant banks on rom from being swapped in.
 *
 * Revision 1.5  1999/01/05 05:40:19  nyef
 * hacked in primitive controller support.
 *
 * Revision 1.4  1999/01/03 22:07:39  nyef
 * added vdp emulation stuff.
 * added some other stuff too.
 *
 * Revision 1.3  1999/01/03 02:20:19  nyef
 * switched from CPUT_EMUZ80 to CPUT_MARATZ80.
 *
 * Revision 1.2  1999/01/02 06:40:50  nyef
 * added dummy I/O interface.
 *
 * Revision 1.1  1999/01/01 20:43:47  nyef
 * Initial revision
 *
 */
