/*
 * sms9918.c
 *
 * sms custom VDP emulation.
 */

/* $Id: sms9918.c,v 1.31 1999/07/08 02:36:33 nyef Exp $ */

#include <stdlib.h>
#include <string.h>
#include "sms9918.h"
#include "ui.h"
#include "video.h"
#include "blitters.h"
#include "sms_psg.h"

/* manifest constants and flag defines */

#define SMS_VDP_RAMSIZE 0x4000

#define TF_ADDRWRITE 1
#define TF_GAMEGEAR 2


/* Palette definition */

int sms_palbase_red[64] = {
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
    0x00, 0x55, 0xaa, 0xff, 0x00, 0x55, 0xaa, 0xff,
};

int sms_palbase_green[64] = {
    0x00, 0x00, 0x00, 0x00, 0x55, 0x55, 0x55, 0x55,
    0xaa, 0xaa, 0xaa, 0xaa, 0xff, 0xff, 0xff, 0xff,
    0x00, 0x00, 0x00, 0x00, 0x55, 0x55, 0x55, 0x55,
    0xaa, 0xaa, 0xaa, 0xaa, 0xff, 0xff, 0xff, 0xff,
    0x00, 0x00, 0x00, 0x00, 0x55, 0x55, 0x55, 0x55,
    0xaa, 0xaa, 0xaa, 0xaa, 0xff, 0xff, 0xff, 0xff,
    0x00, 0x00, 0x00, 0x00, 0x55, 0x55, 0x55, 0x55,
    0xaa, 0xaa, 0xaa, 0xaa, 0xff, 0xff, 0xff, 0xff,
};

int sms_palbase_blue[64] = {
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
    0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55,
    0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55, 0x55,
    0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa,
    0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
    0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff,
};


/* tilecache stuff */

char sms9918_tilecache_dirty[512];
unsigned char sms9918_tilecache[0x8000];
unsigned char sms9918_tilecache_rev[0x8000];

/* port 0 (data) emulation */

unsigned char sms9918_readport0(sms9918 vdp)
{
    unsigned char retval;

    if (vdp->address & 0x4000) {
	if (!(vdp->flags & TF_GAMEGEAR)) {
	    retval = vdp->palette[(vdp->address++) & 0x1f];
	    vdp->address &= 0x401f;
	} else {
	    retval = vdp->palette[(vdp->address++) & 0x3f];
	    vdp->address &= 0x403f;
	}
    } else {
	retval = vdp->readahead;
	vdp->readahead = vdp->memory[vdp->address++];
	vdp->address &= 0x3fff;
    }
    vdp->flags &= ~TF_ADDRWRITE;
    return retval;
}

void sms9918_writeport0(sms9918 vdp, unsigned char data)
{
    if (vdp->address & 0x4000) {
	if (!(vdp->flags & TF_GAMEGEAR)) {
	    vdp->palette_xlat[(vdp->address) & 0x1f] = vid_pre_xlat[data&0x3f];
	    vdp->palette[(vdp->address++) & 0x1f] = data;
	    vdp->address &= 0x401f;
	} else {
	    unsigned char tmp;
	    vdp->palette[(vdp->address) & 0x3f] = data;

	    tmp = (vdp->palette[((vdp->address) & 0x3f) | 1] << 2) & 0x30;
	    tmp |= (vdp->palette[(vdp->address) & 0x3e] >> 4) & 0x0c;
	    tmp |= (vdp->palette[(vdp->address) & 0x3e] >> 2) & 0x03;

	    vdp->palette_xlat[((vdp->address) >> 1) & 0x1f] = vid_pre_xlat[tmp];
	    vdp->address++;
	    vdp->address &= 0x403f;
	}
    } else {
	sms9918_tilecache_dirty[vdp->address >> 5] = 1;
	vdp->readahead = data;
	vdp->memory[vdp->address++] = data;
	vdp->address &= 0x3fff;
    }
    vdp->flags &= ~TF_ADDRWRITE;
}


/* port 1 (status/address/palette/registers) emulation */

unsigned char sms9918_readport1(sms9918 vdp)
{
    unsigned char retval;

    retval = vdp->status;
    vdp->status &= 0x3f;
    vdp->flags &= ~TF_ADDRWRITE;
    return retval;
}

void sms9918_writeport1(sms9918 vdp, unsigned char data)
{
    if (vdp->flags & TF_ADDRWRITE) {
	if (data & 0x80) {
	    if (data & 0x40) {
		vdp->address = (vdp->addrsave | (data << 8)) & 0x401f;
	    } else {
		vdp->regs[data & 15] = vdp->addrsave;
	    }
	} else {
	    vdp->address = (vdp->addrsave | (data << 8)) & 0x3fff;
	    if (!(data & 0x40)) {
		vdp->readahead = vdp->memory[vdp->address++];
		vdp->address &= 0x3fff;
	    }
	}
	vdp->flags &= ~TF_ADDRWRITE;
    } else {
	vdp->addrsave = data;
	vdp->flags |= TF_ADDRWRITE;
    }
}


/* scanline read emulation */

unsigned char sms9918_readscanline(sms9918 vdp)
{
    return (vdp->cur_scanline < 192)? vdp->cur_scanline: 255;
}

const unsigned char sms9918_cache_palette[16] = {0, 1, 2, 3, 4, 5, 6, 7, 8,
						 9, 10, 11, 12, 13, 14, 15};
void sms9918_cache_tile(sms9918 vdp, unsigned short tile)
{
    unsigned char *cur_vbp;
    unsigned char *cur_vbp2;
    unsigned char *tiledata;
    int i;
    
    sms9918_tilecache_dirty[tile] = 0;
    cur_vbp = &sms9918_tilecache[tile << 6];
    cur_vbp2 = &sms9918_tilecache_rev[tile << 6];
    tiledata = &vdp->memory[tile << 5];

    for (i = 8; i; i--) {
	blit_4_8(tiledata[3], tiledata[2], tiledata[1], tiledata[0], 8, 0, sms9918_cache_palette, cur_vbp);
	blit_4_8_rev(tiledata[3], tiledata[2], tiledata[1], tiledata[0], 8, 0, sms9918_cache_palette, cur_vbp2);
	cur_vbp += 8;
	cur_vbp2 += 8;
	tiledata += 4;
    }
}

/* video rendering */

unsigned char *sms_back_tiles[32];
unsigned char *sms_back_palettes[32];
unsigned char sms_back_priorities[32];
unsigned char sms_back_finescroll;

void sms9918_init_backcache(sms9918 vdp)
{
    unsigned short *cur_pos;
    unsigned short cur_tile;
    unsigned char *tiledata;
    int cur_scanline;
    int cur_pattern_line;
    int scroll;
    int i;

    cur_scanline = vdp->cur_scanline + vdp->regs[9];
    while (cur_scanline > 223) {
	cur_scanline -= 224;
    }
    cur_pattern_line = cur_scanline & 7;
    cur_pos = (unsigned short *)&vdp->memory[(vdp->regs[2] & 14) << 10];
    cur_pos += (cur_scanline >> 3) << 5;
    if ((vdp->regs[0] & 0x40) && (cur_scanline < 16)) {
	scroll = 0;
	sms_back_finescroll = 0;
    } else {
	scroll = 32 - ((vdp->regs[8] + 7) >> 3);
	sms_back_finescroll = (8 - (vdp->regs[8] & 7)) & 7;
    }
    
    for (i = 0; i < 32; i++) {
	cur_tile = cur_pos[(i + scroll) & 0x1f];
	if (sms9918_tilecache_dirty[cur_tile & 0x1ff]) {
	    sms9918_cache_tile(vdp, cur_tile & 0x1ff);
	}
	if (cur_tile & 0x200) {
	    tiledata = sms9918_tilecache_rev;
	} else {
	    tiledata = sms9918_tilecache;
	}
	if (cur_tile & 0x400) {
	    tiledata += ((cur_tile & 0x1ff) << 6) + ((7 - cur_pattern_line) << 3);
	} else {
	    tiledata += ((cur_tile & 0x1ff) << 6) + (cur_pattern_line << 3);
	}

	sms_back_tiles[i] = tiledata;
	
	if (cur_tile & 0x800) {
	    sms_back_palettes[i] = &vdp->palette_xlat[16];
	} else {
	    sms_back_palettes[i] = &vdp->palette_xlat[0];
	}

	if (cur_tile & 0x1000) {
	    sms_back_priorities[i] = 1;
	} else {
	    sms_back_priorities[i] = 0;
	}
    }
}

void sms9918_render_background(sms9918 vdp, unsigned char *cur_vbp)
{
    int i;
    unsigned char *palette;
    unsigned char *tiledata;

    memset(cur_vbp, sms_back_palettes[0][0], 256);

    if (vdp->regs[0] & 0x20) {
	cur_vbp[0] = vid_pre_xlat[0];
	cur_vbp[1] = vid_pre_xlat[0];
	cur_vbp[2] = vid_pre_xlat[0];
	cur_vbp[3] = vid_pre_xlat[0];
	cur_vbp[4] = vid_pre_xlat[0];
	cur_vbp[5] = vid_pre_xlat[0];
	cur_vbp[6] = vid_pre_xlat[0];
	cur_vbp[7] = vid_pre_xlat[0];
	cur_vbp += 8;
	i = 1;
    } else {
	i = 0;
    }
    
    if (sms_back_priorities[i]) {
	cur_vbp += 8 - sms_back_finescroll;
    } else {
	palette = sms_back_palettes[i];
	tiledata = sms_back_tiles[i] + sms_back_finescroll;
	switch (sms_back_finescroll) {
	case 0:
	    *cur_vbp++ = palette[*tiledata++];
	case 1:
	    *cur_vbp++ = palette[*tiledata++];
	case 2:
	    *cur_vbp++ = palette[*tiledata++];
	case 3:
	    *cur_vbp++ = palette[*tiledata++];
	case 4:
	    *cur_vbp++ = palette[*tiledata++];
	case 5:
	    *cur_vbp++ = palette[*tiledata++];
	case 6:
	    *cur_vbp++ = palette[*tiledata++];
	case 7:
	    *cur_vbp++ = palette[*tiledata++];
	}
    }
    
    for (i++; i < 32; i++) {
	if (sms_back_priorities[i]) {
	    cur_vbp += 8;
	} else {
	    palette = sms_back_palettes[i];
	    tiledata = sms_back_tiles[i];
	    *cur_vbp++ = palette[*tiledata++];
	    *cur_vbp++ = palette[*tiledata++];
	    *cur_vbp++ = palette[*tiledata++];
	    *cur_vbp++ = palette[*tiledata++];
	    *cur_vbp++ = palette[*tiledata++];
	    *cur_vbp++ = palette[*tiledata++];
	    *cur_vbp++ = palette[*tiledata++];
	    *cur_vbp++ = palette[*tiledata++];
	}
    }
    
    if (sms_back_finescroll) {
	palette = sms_back_palettes[0];
	if (!sms_back_priorities[0]) {
	    tiledata = sms_back_tiles[0];
	    switch (sms_back_finescroll) {
	    case 7:
		*cur_vbp++ = palette[*tiledata++];
	    case 6:
		*cur_vbp++ = palette[*tiledata++];
	    case 5:
		*cur_vbp++ = palette[*tiledata++];
	    case 4:
		*cur_vbp++ = palette[*tiledata++];
	    case 3:
		*cur_vbp++ = palette[*tiledata++];
	    case 2:
		*cur_vbp++ = palette[*tiledata++];
	    case 1:
		*cur_vbp++ = palette[*tiledata++];
	    }
	}
    }
}

void sms9918_render_foreground(sms9918 vdp, unsigned char *cur_vbp)
{
    int i;
    unsigned char *palette;
    unsigned char *tiledata;
    
    if (vdp->regs[0] & 0x20) {
	cur_vbp += 8;
	i = 1;
    } else {
	i = 0;
    }
    
    if (!sms_back_priorities[i]) {
	cur_vbp += 8 - sms_back_finescroll;
    } else {
	palette = sms_back_palettes[i];
	tiledata = sms_back_tiles[i] + sms_back_finescroll;
	switch (sms_back_finescroll) {
	case 0:
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	case 1:
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	case 2:
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	case 3:
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	case 4:
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	case 5:
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	case 6:
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	case 7:
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	}
    }
    
    for (i++; i < 32; i++) {
	if (!sms_back_priorities[i]) {
	    cur_vbp += 8;
	} else {
	    palette = sms_back_palettes[i];
	    tiledata = sms_back_tiles[i];
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	}
    }
    
    if (sms_back_finescroll && (!sms_back_priorities[0])) {
	palette = sms_back_palettes[0];
	tiledata = sms_back_tiles[0];
	switch (sms_back_finescroll) {
	case 7:
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	case 6:
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	case 5:
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	case 4:
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	case 3:
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	case 2:
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	case 1:
	    if (*tiledata)
	    *cur_vbp = palette[*tiledata];
	    cur_vbp++;
	    tiledata++;
	}
    }
}

void sms9918_render_sprites(sms9918 vdp, unsigned char *cur_vbp)
{
    unsigned char *sprite_table;
    unsigned char *tiledata[8];
    unsigned short cur_tile;
    unsigned char sprite_x[8];
    int cur_pattern_line;
    int sprite_8x16;
    int i;
    int k;

    k = 0;
    sprite_table = &vdp->memory[(vdp->regs[5] & 0x7e) << 7];
    sprite_8x16 = (vdp->regs[1] & 2)? 16: 8;
    
    for (i = 0; (i < 64) && (k < 8); i++) {
	cur_pattern_line = vdp->cur_scanline - sprite_table[i];
	cur_pattern_line &= 0xff;
	if (cur_pattern_line < sprite_8x16) {
	    cur_tile = sprite_table[128 + (i << 1) + 1] | ((vdp->regs[6] & 4) << 6);
	    if (vdp->regs[1] & 2) {
		cur_tile &= 0x1fe;
	    }
	    tiledata[k] = &vdp->memory[(cur_tile << 5) + (cur_pattern_line << 2)];
	    sprite_x[k] = sprite_table[128 + (i << 1) + 0];
	    k++;
	} else if (sprite_table[i] == 208) {
	    break;
	}
    }
    for (i = k - 1; i >= 0; i--) {
	blit_4_8_czt(tiledata[i][3], tiledata[i][2],
		     tiledata[i][1], tiledata[i][0],
		     8, 0, &vdp->palette_xlat[16],
		     &cur_vbp[sprite_x[i]]);
    }
}

void sms9918_render_line(sms9918 vdp)
{
    unsigned char *video_buffer;

    video_buffer = video_get_vbp(vdp->cur_scanline);
    
    if (vdp->regs[1] & 0x40) {
	sms9918_init_backcache(vdp);
	sms9918_render_background(vdp, video_buffer);
	sms9918_render_sprites(vdp, video_buffer);
	sms9918_render_foreground(vdp, video_buffer);
    } else {
	/* FIXME: floodfill with black */
    }
}


/* periodic task */

int sms9918_periodic(sms9918 vdp)
{
    /* FIXME: does not account for overscan. */
    if (vdp->cur_scanline < 192) {
	sms9918_render_line(vdp);
    } else if (vdp->cur_scanline == 192) {
/* 	video_events(); */
	video_display_buffer();
	sms_psg_vsync();
	vdp->status |= 0x80;
    }
    if (vdp->cur_scanline == 261) {
	vdp->cur_scanline = 0;
	vdp->linecounter = vdp->regs[10];
    } else {
	vdp->cur_scanline++;
	if ((vdp->cur_scanline < 192) && !vdp->linecounter--) {
	    vdp->status |= 0x40;
	    vdp->linecounter = vdp->regs[10];
	}
    }
    return (((vdp->status & 0x80) && (vdp->regs[1] & 0x20)) ||
	    ((vdp->status & 0x40) && (vdp->regs[0] & 0x10)));
}


/* initialization */

sms9918 sms9918_create(int is_gamegear)
{
    sms9918 retval;

    retval = calloc(1, sizeof(struct sms9918));
    if (retval) {
	retval->memory = calloc(1, SMS_VDP_RAMSIZE);
	if (retval->memory) {
	    if (is_gamegear) {
		retval->flags = TF_GAMEGEAR;
	    }
	    retval->readport0 = sms9918_readport0;
	    retval->readport1 = sms9918_readport1;
	    retval->writeport0 = sms9918_writeport0;
	    retval->writeport1 = sms9918_writeport1;
	    retval->readscanline = sms9918_readscanline;
	    retval->periodic = sms9918_periodic;
	    video_setsize(256, 192);
	    video_setpal(64, sms_palbase_red, sms_palbase_green, sms_palbase_blue);
	} else {
	    free(retval);
	    retval = NULL;
	}
    }
    
    if (!retval) {
	deb_printf("sms9918_create(): out of memory.\n");
    }
    
    return retval;
}

/*
 * $Log: sms9918.c,v $
 * Revision 1.31  1999/07/08 02:36:33  nyef
 * fixed to not display background on left 8 pixels of the screen when
 * display of left column disabled
 *
 * fixed broken scrolling of leftmost 0-7 pixels of the background
 *
 * Revision 1.30  1999/07/05 23:28:25  nyef
 * fixed sprite renderer to display sprites at the top of the screen
 *
 * Revision 1.29  1999/06/13 20:44:12  nyef
 * cleaned up a silliness in the code generated for sms9918_cache_tile()
 *
 * Revision 1.28  1999/06/13 20:15:02  nyef
 * changed to use memset(3) to handle background color on transparent tiles
 *
 * Revision 1.27  1999/05/08 15:41:44  nyef
 * fixed problem with blue color in gg palette code
 *
 * Revision 1.26  1999/05/05 00:24:45  nyef
 * fixed tile transparencies in renderer
 *
 * Revision 1.25  1999/05/01 05:36:57  nyef
 * split up the renderer code into "background" and "forground" versions
 *
 * Revision 1.24  1999/05/01 05:16:09  nyef
 * split the background renderer up into two separate functions
 *
 * Revision 1.23  1999/05/01 04:41:31  nyef
 * removed non-tilecache renderer code
 *
 * Revision 1.22  1999/05/01 03:32:05  nyef
 * cleaned up the background renderer a little more
 * added a reversed tilecache
 *
 * Revision 1.21  1999/04/28 16:48:51  nyef
 * cleaned up the background renderer a little
 * added a tilecache system
 *
 * Revision 1.20  1999/04/28 13:44:47  nyef
 * changed to add sound emulation
 *
 * Revision 1.19  1999/03/12 02:12:45  nyef
 * added preliminary GameGear palette support
 *
 * Revision 1.18  1999/02/07 17:06:47  nyef
 * changed to use video_setsize()
 *
 * Revision 1.17  1999/02/06 03:37:42  nyef
 * removed call to video_events()
 *
 * Revision 1.16  1999/01/24 17:26:48  nyef
 * fixed background renderer to display the final few pixels on the
 * right side while h-scrolling.
 *
 * Revision 1.15  1999/01/24 17:17:39  nyef
 * switched to using blitters from blitters.c for sprite rendering.
 * got rid of the old blitter system.
 *
 * Revision 1.14  1999/01/11 02:25:05  nyef
 * switched to using blitters from blitters.c for background rendering.
 *
 * Revision 1.13  1999/01/10 02:13:32  nyef
 * reversed sprite display order.
 *
 * Revision 1.12  1999/01/10 00:02:42  nyef
 * changed sms9918_readscanline() to return 255 for values > 191.
 *
 * Revision 1.11  1999/01/09 23:09:55  nyef
 * added support for reading the current scanline.
 *
 * Revision 1.10  1999/01/09 22:01:47  nyef
 * fixed problem with tile priorities and horizontal scroll.
 *
 * Revision 1.9  1999/01/09 01:31:58  nyef
 * added support for 8x16 sprites.
 * limited number of sprites per line to 8 (supposedly the real system
 * does this too.).
 *
 * Revision 1.8  1999/01/09 00:48:35  nyef
 * fixed stupid mistake in tile priority code.
 *
 * Revision 1.7  1999/01/09 00:45:47  nyef
 * added preliminary support for vscroll.
 *
 * Revision 1.6  1999/01/09 00:39:12  nyef
 * added preliminary implimentation of tile priority flag.
 *
 * Revision 1.5  1999/01/08 19:36:21  nyef
 * improved timing in sms9918_periodic().
 *
 * Revision 1.4  1999/01/08 02:24:25  nyef
 * added line interrupts.
 *
 * Revision 1.3  1999/01/08 02:07:25  nyef
 * added horizontal scroll.
 *
 * Revision 1.2  1999/01/07 05:20:35  nyef
 * fixed sprite data table lookup.
 * added sprite tileset selection.
 * fixed sprite palette lookup.
 *
 * Revision 1.1  1999/01/07 03:17:07  nyef
 * Initial revision
 *
 */
