/*
 * sms_psg.c
 *
 * SMS sound emulation system
 */

/* $Id: sms_psg.c,v 1.3 1999/07/10 02:44:36 nyef Exp $ */

#define SOUND
#define SOUND_LINUX
/* #define SOUND_DOS */
/* #define DONT_USE_CAL */ /* for RockNES */

#include <unistd.h>

#ifdef SOUND_LINUX
#  include <stdio.h>
#  include <sys/types.h>
#  include <sys/stat.h>
#  include <fcntl.h>
#  include <sys/soundcard.h>
#  include <sys/ioctl.h>
#endif

#ifdef SOUND_DOS
#  include <allegro.h>
#endif

#ifndef DONT_USE_CAL
#include "cal.h"
#endif

#include "ui.h"
#include "sms_psg.h"

struct {
    int lastchan; /* for frequency writes */
    unsigned char vols[4];
    unsigned char freqls[4];
    unsigned char freqhs[4];
    unsigned long index[4];
    unsigned long skip[4];
    unsigned long shiftreg;
} sms_psg;

/*
 * sms_psg_vsync does the actual sound rendering.
 */

#ifdef SOUND

unsigned char wave_buffers[4][735]; /* 44100 / 60 = 735 samples per sync */

#ifdef SOUND_LINUX
unsigned char final_wave[2048];
int waveptr;
int wavflag;
#endif

#endif

/* int nes_psg_quality; */

#ifdef SOUND
unsigned long sms_psg_magic;
unsigned int sms_psg_samples_per_sync;

extern unsigned int nes_psg_sample_rate;
extern unsigned int nes_psg_samples_per_sync;

/*
 * magic number calculation: (NOTE: the figure 125000 below is probably wrong)
 *
 * 44100 / 125000 = 0.3528
 * 32 / 0.3528 = 90.702947
 * 90.702947 * 2^16 = 5944308.3
 * 5944308 as hex = 0x5ab3f4
 */

struct sms_psg_quality_data {
    unsigned long magic;
    unsigned int samples_per_sync;
    unsigned int sample_rate;
} sms_psg_qual[] = {
    {0xb567e800, 367, 22050},
    {0x5ab3f400, 735, 44100},
};

#ifdef SOUND_LINUX
int sound_fd;
#endif

#ifdef SOUND_DOS
AUDIOSTREAM *sound_stream;
#endif

/* NOTE: these routines use 8.24 bit fixed point math in places. */

/* FIXME: some frequency values are not handled well by this system. */

extern unsigned char pulse_50[0x20];

void sms_psg_calc_freqs(int channel)
{
    unsigned int freq;
    
    if (channel != 3) {
	freq = (sms_psg.freqhs[channel] << 4) + sms_psg.freqls[channel];
    } else {
	if ((sms_psg.freqls[3] & 3) != 3) {
	    freq = (0x10 << (sms_psg.freqls[3] & 3));
	} else {
	    freq = ((sms_psg.freqhs[2] << 4) + sms_psg.freqls[2]) << 1;
	}
    }
    
    if (freq) {
	sms_psg.skip[channel] = sms_psg_magic / freq;
    } else {
	sms_psg.skip[channel] = 0;
    }

    if ((channel == 2) && ((sms_psg.freqls[3] & 3) == 3)) {
	freq = ((sms_psg.freqhs[2] << 4) + sms_psg.freqls[2]) << 1;
	if (freq) {
	    sms_psg.skip[3] = sms_psg_magic / freq;
	} else {
	    sms_psg.skip[3] = 0;
	}
    }
}

void sms_psg_write(unsigned char data)
{
    if (data & 0x80) {
	if (data & 0x10) {
	    /* volume */
	    sms_psg.vols[(data >> 5) & 3] = 15 - (data & 15);
	} else {
	    /* freq low */
	    sms_psg.lastchan = (data >> 5) & 3;
	    sms_psg.freqls[sms_psg.lastchan] = data & 15;
	    sms_psg_calc_freqs(sms_psg.lastchan);
	}
    } else {
	/* freq high */
	sms_psg.freqhs[sms_psg.lastchan] = data & 0x3f;
	sms_psg_calc_freqs(sms_psg.lastchan);
    }

}

void sms_psg_wave_1(void)
{
    int i;

    for (i = 0; i < sms_psg_samples_per_sync; i++) {
	sms_psg.index[0] += sms_psg.skip[0];
	sms_psg.index[0] &= 0x1fffffff;
	wave_buffers[0][i] = pulse_50[sms_psg.index[0] >> 24] * sms_psg.vols[0];
    }
}

void sms_psg_wave_2(void)
{
    int i;

    for (i = 0; i < sms_psg_samples_per_sync; i++) {
	sms_psg.index[1] += sms_psg.skip[1];
	sms_psg.index[1] &= 0x1fffffff;
	wave_buffers[1][i] = pulse_50[sms_psg.index[1] >> 24] * sms_psg.vols[1];
    }
}

void sms_psg_wave_3(void)
{
    int i;

    for (i = 0; i < sms_psg_samples_per_sync; i++) {
	sms_psg.index[2] += sms_psg.skip[2];
	sms_psg.index[2] &= 0x1fffffff;
	wave_buffers[2][i] = pulse_50[sms_psg.index[2] >> 24] * sms_psg.vols[2];
    }
}

void sms_psg_wave_4(void)
{
#if 0
    int i;

    if (!sms_psg.shiftreg) {
	sms_psg.shiftreg = 1;
    }

    for (i = 0; i < sms_psg_samples_per_sync; i++) {
	sms_psg.index[3] += sms_psg.skip[3];
	if (sms_psg.index[3] > 0x1fffffff) {
	    if (sms_psg.freqls[3] & 0x04) {
		sms_psg.shiftreg |= ((!(sms_psg.shiftreg & 1)) ^ (!(sms_psg.shiftreg & 4))) << 5;
	    } else {
		sms_psg.shiftreg |= (!(sms_psg.shiftreg & 1)) << 5;
	    }
	    sms_psg.shiftreg >>= 1;
	}
	sms_psg.index[3] &= 0x1fffffff;
	if (sms_psg.shiftreg & 1) {
	    wave_buffers[3][i] = 0x11 * sms_psg.vols[3];
	} else {
	    wave_buffers[3][i] = 0;
	}
    }
#endif
}

void sms_psg_mix(unsigned char *final_wave)
{
    int i;
    
    /* note: 41 CPU cycles occur between increments of i */
    for (i = 0; i < sms_psg_samples_per_sync; i++) {
#ifdef SOUND_DOS
	final_wave[i] = (wave_buffers[0][i] + wave_buffers[1][i] +
			 wave_buffers[2][i] + wave_buffers[3][i]) >> 2;
#endif
#ifdef SOUND_LINUX
	final_wave[waveptr] = (wave_buffers[0][i] + wave_buffers[1][i] +
			       wave_buffers[2][i] + wave_buffers[3][i]) >> 2;
	waveptr++;
	if (waveptr == 2048) {
	    waveptr = 0;
	    wavflag = 2;
	} else if (waveptr == 1024) {
	    wavflag = 1;
	}
#endif
    }
}

#endif

void sms_psg_vsync(void)
{
#ifdef SOUND_DOS
    unsigned char *final_wave = NULL;
#endif
    
#ifdef SOUND_LINUX
    sms_psg_wave_1();
    sms_psg_wave_2();
    sms_psg_wave_3();
    sms_psg_wave_4();

    sms_psg_mix(final_wave);

    if (wavflag) {
	if (sound_fd) {
	    if (write(sound_fd, &final_wave[(wavflag - 1) << 10], 1024) < 1024) {
		deb_printf("wrote less than 1024 bytes\n");
	    }
	}
	wavflag = 0;
    }
#endif
    
#ifdef SOUND_DOS
    while (!final_wave) { /* Wheee... Primitive Framesync. :-) */
	final_wave = get_audio_stream_buffer(sound_stream);
    }
    
    sms_psg_wave_1();
    sms_psg_wave_2();
    sms_psg_wave_3();
    sms_psg_wave_4();
    
    sms_psg_mix(final_wave);
    
    free_audio_stream_buffer(sound_stream);
#endif
}

#ifdef SOUND_LINUX
void nes_psg_open_sound_linux(void);
#endif

#ifdef SOUND_DOS
void nes_psg_open_sound_dos(void);
#endif

void sms_psg_init(void)
{
#ifdef SOUND
    if (nes_psg_quality > 0) {
	if (nes_psg_quality > 2) {
	    nes_psg_quality = 1;
	} else {
	    nes_psg_quality--;
	}
	sms_psg_magic = sms_psg_qual[nes_psg_quality].magic;
	sms_psg_samples_per_sync = sms_psg_qual[nes_psg_quality].samples_per_sync;
	nes_psg_samples_per_sync = sms_psg_samples_per_sync;
	nes_psg_sample_rate = sms_psg_qual[nes_psg_quality].sample_rate;

#ifdef SOUND_LINUX
	nes_psg_open_sound_linux();
#endif
    
#ifdef SOUND_DOS
	nes_psg_open_sound_dos();
#endif
    } else {
#ifdef SOUND_LINUX
	sound_fd = 0;
#endif
    
#ifdef SOUND_DOS
	sound_stream = NULL;
#endif
    }
#endif
}

void sms_psg_done(void)
{
#ifdef SOUND_LINUX
    if (sound_fd) {
	close(sound_fd);
    }
#endif
    
#ifdef SOUND_DOS
    stop_audio_stream(sound_stream);
#endif
}

/*
 * $Log: sms_psg.c,v $
 * Revision 1.3  1999/07/10 02:44:36  nyef
 * rearranged the calculation of frequencies and skip factors
 *
 * Revision 1.2  1999/05/04 23:50:30  nyef
 * fixed problem with initialization in DOS version
 *
 * Revision 1.1  1999/04/28 14:11:08  nyef
 * Initial revision
 *
 */
