/*
 * tms9918.c
 *
 * tms9918 VDP emulation.
 */

/* $Id: tms9918.c,v 1.4 1999/06/10 01:50:08 nyef Exp $ */

#include <stdlib.h>
#include <string.h>
#include "tms9918.h"
#include "ui.h"
#include "video.h"
#include "sms_psg.h" /* FIXME: this is silly */

#define TMS_RAMSIZE 0x4000

#define TF_ADDRWRITE 1

/* palette definition */

int tms9918_palbase_red[16] = {
    0x00, 0x00, 0x22, 0x66, 0x22, 0x44, 0xbb, 0x44,
    0xff, 0xff, 0xdd, 0xdd, 0x22, 0xdd, 0xbb, 0xff,
};

int tms9918_palbase_green[16] = {
    0x00, 0x00, 0xdd, 0xff, 0x22, 0x66, 0x22, 0xdd,
    0x22, 0x66, 0xdd, 0xdd, 0x99, 0x44, 0xbb, 0xff,
};

int tms9918_palbase_blue[16] = {
    0x00, 0x00, 0x22, 0x66, 0xff, 0xff, 0x11, 0xff,
    0x22, 0x66, 0x22, 0x99, 0x22, 0xbb, 0xbb, 0xff,
};

unsigned char tms9918_readport0(tms9918 vdp)
{
    unsigned char retval;
    
    retval = vdp->readahead;
    vdp->readahead = vdp->memory[vdp->address++];
    vdp->address &= 0x3fff;
    vdp->flags &= ~TF_ADDRWRITE;
    return retval;
}

unsigned char tms9918_readport1(tms9918 vdp)
{
    unsigned char retval;

    retval = vdp->status;
    vdp->status &= 0x5f;
    vdp->flags &= ~TF_ADDRWRITE;
    return retval;
}

void tms9918_writeport0(tms9918 vdp, unsigned char data)
{
    vdp->readahead = data;
    vdp->memory[vdp->address++] = data;
    vdp->address &= 0x3fff;
    vdp->flags &= ~TF_ADDRWRITE;
}

void tms9918_writeport1(tms9918 vdp, unsigned char data)
{
    if (vdp->flags & TF_ADDRWRITE) {
	if (data & 0x80) {
	    deb_printf("tms9918: register write %d = %02x.\n", data & 7, vdp->addrsave);
	    vdp->regs[data & 7] = vdp->addrsave;
	} else {
/* 	    deb_printf("tms9918: address write.\n"); */
	    vdp->address = (vdp->addrsave | (data << 8)) & 0x3fff;
	    if (!(data & 0x40)) {
		vdp->readahead = vdp->memory[vdp->address++];
		vdp->address &= 0x3fff;
	    }
	}
	vdp->flags &= ~TF_ADDRWRITE;
    } else {
	vdp->addrsave = data;
	vdp->flags |= TF_ADDRWRITE;
    }
}

void tms9918_render_line(tms9918 vdp)
{
    unsigned char *cur_vbp;
    unsigned char *nametable;
    unsigned char *patterntable;
    unsigned char *colortable;
    unsigned char color0;
    unsigned char color1;
    unsigned char pattern;
    int i;
    
    cur_vbp = video_get_vbp(vdp->scanline);
    nametable = vdp->memory + ((vdp->regs[2] & 0x0f) << 10);

    if (vdp->regs[1] & 0x40) {
	if ((!(vdp->regs[0] & 0x02)) && (!(vdp->regs[1] & 0x18))) {
	    /* MODE 0 */
	    patterntable = vdp->memory + ((vdp->regs[4] & 0x07) << 10);
	    patterntable += vdp->scanline & 7;
	    nametable += (vdp->scanline & ~7) << 2;
	    colortable = vdp->memory + (vdp->regs[3] << 6);
	    
	    for (i = 0; i < 32; i++) {
		pattern = patterntable[nametable[i] << 3];
		color0 = vid_pre_xlat[colortable[nametable[i] >> 3] & 15];
		color1 = vid_pre_xlat[colortable[nametable[i] >> 3] >> 4];
		if (pattern & 0x80) {
		    *cur_vbp++ = color1;
		} else {
		    *cur_vbp++ = color0;
		}
		if (pattern & 0x40) {
		    *cur_vbp++ = color1;
		} else {
		    *cur_vbp++ = color0;
		}
		if (pattern & 0x20) {
		    *cur_vbp++ = color1;
		} else {
		    *cur_vbp++ = color0;
		}
		if (pattern & 0x10) {
		    *cur_vbp++ = color1;
		} else {
		    *cur_vbp++ = color0;
		}
		if (pattern & 0x08) {
		    *cur_vbp++ = color1;
		} else {
		    *cur_vbp++ = color0;
		}
		if (pattern & 0x04) {
		    *cur_vbp++ = color1;
		} else {
		    *cur_vbp++ = color0;
		}
		if (pattern & 0x02) {
		    *cur_vbp++ = color1;
		} else {
		    *cur_vbp++ = color0;
		}
		if (pattern & 0x01) {
		    *cur_vbp++ = color1;
		} else {
		    *cur_vbp++ = color0;
		}
	    }
	} else if ((vdp->regs[0] & 0x02) && (!(vdp->regs[1] & 0x18))) {
	    /* MODE 2 */
	    patterntable = vdp->memory + ((vdp->regs[4] & 0x04)? 0x2000: 0);
	    patterntable += vdp->scanline & 7;
	    if ((vdp->scanline >= 0x80) && (vdp->regs[4] & 0x02)) {
		patterntable += (0x200 << 3);
	    } else if ((vdp->scanline >= 0x40) && (vdp->regs[4] & 0x01)) {
		patterntable += (0x100 << 3);
	    }

	    nametable += (vdp->scanline & ~7) << 2;
	    
	    for (i = 0; i < 32; i++) {
		pattern = patterntable[nametable[i] << 3];
		if (pattern & 0x80) {
		    *cur_vbp++ = vid_pre_xlat[5];
		} else {
		    *cur_vbp++ = vid_pre_xlat[1];
		}
		if (pattern & 0x40) {
		    *cur_vbp++ = vid_pre_xlat[5];
		} else {
		    *cur_vbp++ = vid_pre_xlat[1];
		}
		if (pattern & 0x20) {
		    *cur_vbp++ = vid_pre_xlat[5];
		} else {
		    *cur_vbp++ = vid_pre_xlat[1];
		}
		if (pattern & 0x10) {
		    *cur_vbp++ = vid_pre_xlat[5];
		} else {
		    *cur_vbp++ = vid_pre_xlat[1];
		}
		if (pattern & 0x08) {
		    *cur_vbp++ = vid_pre_xlat[5];
		} else {
		    *cur_vbp++ = vid_pre_xlat[1];
		}
		if (pattern & 0x04) {
		    *cur_vbp++ = vid_pre_xlat[5];
		} else {
		    *cur_vbp++ = vid_pre_xlat[1];
		}
		if (pattern & 0x02) {
		    *cur_vbp++ = vid_pre_xlat[5];
		} else {
		    *cur_vbp++ = vid_pre_xlat[1];
		}
		if (pattern & 0x01) {
		    *cur_vbp++ = vid_pre_xlat[5];
		} else {
		    *cur_vbp++ = vid_pre_xlat[1];
		}
	    }
	} else {
	    deb_printf("tms9918: unsupported display mode %c%c%c.\n",
		       (vdp->regs[1] & 0x10)? '1': '-',
		       (vdp->regs[0] & 0x02)? '2': '-',
		       (vdp->regs[1] & 0x08)? '3': '-');
	}
    } else {
	memset(cur_vbp, vid_pre_xlat[0], 256);
    }
}

int tms9918_periodic(tms9918 vdp)
{
    if (vdp->scanline < 192) {
	tms9918_render_line(vdp);
    } else if (vdp->scanline == 192) {
	video_display_buffer();
	sms_psg_vsync(); /* FIXME: this is silly */
	vdp->status |= 0x80; /* signal vblank */
    }
    if (vdp->scanline == 261) {
	vdp->scanline = 0;
    } else {
	vdp->scanline++;
    }
    return (vdp->status & 0x80) && (vdp->regs[1] & 0x20);
}

tms9918 tms9918_create(void)
{
    tms9918 retval;

    retval = calloc(1, sizeof(struct tms9918));
    if (retval) {
	retval->memory = calloc(1, TMS_RAMSIZE);
	if (retval->memory) {
	    retval->readport0 = tms9918_readport0;
	    retval->readport1 = tms9918_readport1;
	    retval->writeport0 = tms9918_writeport0;
	    retval->writeport1 = tms9918_writeport1;
	    retval->periodic = tms9918_periodic;
	    retval->scanline = 0;
            video_setsize(256, 192);
	    video_setpal(16, tms9918_palbase_red, tms9918_palbase_green, tms9918_palbase_blue);
	} else {
	    free(retval);
	    retval = NULL;
	}
    }
    
    if (!retval) {
	deb_printf("tms9918_create(): out of memory.\n");
    }
    
    return retval;
}

/*
 * $Log: tms9918.c,v $
 * Revision 1.4  1999/06/10 01:50:08  nyef
 * fixed pattern lookup in mode 2
 *
 * Revision 1.3  1999/06/10 00:41:59  nyef
 * fixed colors in mode 0
 *
 * Revision 1.2  1999/06/09 00:34:08  nyef
 * added preliminary implementation of modes 0 and 2
 *
 * Revision 1.1  1999/06/08 01:49:25  nyef
 * Initial revision
 *
 */
