/*
 * tool.c
 *
 * a program to aide in hacking FF3j
 */

/* $Id: tool.c,v 1.14 1999/06/08 01:48:02 nyef Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <ctype.h>
#include <string.h>

#include "machdep.h"
#include "ui.h"
#include "video.h"
#include "tool.h"
#include "system.h"

int system_type; /* currently emulated system */

unsigned long system_flags;

char *romfilename; /* name of rom file */
unsigned char *romimage = NULL;   /* Cart Rom Image */
unsigned char *ramblock = NULL;   /* System and Cart Ram */
int romsize; /* Rom Image size */

static const unsigned char gb_sig[0x30] = {
    0xCE, 0xED, 0x66, 0x66, 0xCC, 0x0D, 0x00, 0x0B,
    0x03, 0x73, 0x00, 0x83, 0x00, 0x0C, 0x00, 0x0D,
    0x00, 0x08, 0x11, 0x1F, 0x88, 0x89, 0x00, 0x0E,
    0xDC, 0xCC, 0x6E, 0xE6, 0xDD, 0xDD, 0xD9, 0x99,
    0xBB, 0xBB, 0x67, 0x63, 0x6E, 0x0E, 0xEC, 0xCC,
    0xDD, 0xDC, 0x99, 0x9F, 0xBB, 0xB9, 0x33, 0x3E
};

void read_romimage(char *filename)
{
    int fd;
    int numread;

    fd = open(filename, O_RDONLY|O_BINARY);
  
    if (fd < 0) {
	romimage = NULL;
    } else {
  
	romsize = lseek(fd, 0, SEEK_END);
	lseek(fd, 0, SEEK_SET);
  
	romimage = malloc(romsize);
	if (romimage != NULL) {
  
	    numread = read(fd, romimage, romsize);
	    
	    if (numread != romsize) {
		free(romimage);
		romimage = NULL;
	    }

	}
	close(fd);
    }

    if (romimage == NULL) {
	deb_printf("error reading file.\n");
	return;
    }
    
    if (!memcmp(romimage, "NES\x1a", 4)) {
	deb_printf("file is NES image.\n");
	system_type = ST_NES;
    } else if (!memcmp(romimage, "FDS\x1a", 4)) {
	deb_printf("file is Famicom Disk System image.\n");
	system_type = ST_NES;
    } else if ((((unsigned short *)romimage)[0] == 0x55aa) ||
	       (((unsigned short *)romimage)[0] == 0xaa55)) {
	deb_printf("file is probably ColecoVision image.\n");
	system_type = ST_COLECO;
    } else if (((romsize >= 0x8000) &&
	       (!memcmp(romimage + 0x7ff0, "TMR SEGA", 8))) ||
	       ((romsize >= 0x8200) &&
	       (!memcmp(romimage + 0x81f0, "TMR SEGA", 8)))) {
	deb_printf("file is SMS/GG image.\n");
	system_type = ST_MASTERSYS;
    } else if (!memcmp(romimage + 0x104, gb_sig, 0x30)) {
	deb_printf("file is Gameboy image.\n");
	system_type = ST_GAMEBOY;
    } else if ((romsize % 0x2000) == 0) {
	deb_printf("file is a multiple of 0x2000 bytes data.\n");
	if (!memcmp(&romimage[0x100], "SEGA ", 5)) {
	    deb_printf("file is Genesis image.\n");
	    system_type = ST_GENESIS;
	} else {
	    deb_printf("assuming file is PC-Engine image.\n");
	    system_type = ST_PCENGINE;
	}
    } else if ((romsize % 0x2000) == 0x200) {
	deb_printf("file is a multiple of 0x2000 bytes data with 512-byte header.\n");
	if ((!memcmp(&romimage[0x2280], "SG ", 3)) &&
	    (!memcmp(&romimage[0x0280], "EA", 2))) {
	    deb_printf("file is Genesis image.\n");
	    system_type = ST_GENESIS;
	} else {
	    deb_printf("assuming file is PC-Engine image.\n");
	    system_type = ST_PCENGINE;
	}
    }
    romfilename = strdup(filename);
}

void activate_system(int system_type, int startup_deb)
{
    switch (system_type) {
    case ST_NONE:
	deb_printf("No rom image loaded.\n");
	break;
	
    case ST_NES:
	nes_run(startup_deb);
	break;
	
    case ST_PCENGINE:
#if 1
	pce_run(startup_deb);
#else
	deb_printf("PC Engine system unsupported.\n");
#endif
	break;

    case ST_MASTERSYS:
#if 1
	sms_run(startup_deb);
#else
	deb_printf("Master System system insupported.\n");
#endif
	break;
	
    case ST_GENESIS:
#if 1
	genesis_run(startup_deb);
#else
	deb_printf("Genesis system unsupported.\n");
#endif
	break;

    case ST_COLECO:
#if 1
	cv_run(startup_deb);
#else
	deb_printf("ColecoVision system unsupported.\n");
#endif
	break;
	
    default:
	deb_printf("unknown system type.\n");
    }
}

/*
 * $Log: tool.c,v $
 * Revision 1.14  1999/06/08 01:48:02  nyef
 * added colecovision support (the rom detect is really lame, though)
 *
 * Revision 1.13  1999/04/17 22:24:41  nyef
 * added genesis support.
 *
 * Revision 1.12  1999/02/20 20:12:07  nyef
 * added minimum size checks to SMS/GG detection (fixes small PCE roms)
 * changed most if not all strncmp()s to memcmp()s
 *
 * Revision 1.11  1999/02/14 06:18:19  nyef
 * added detection of FDS images
 *
 * Revision 1.10  1999/01/30 17:53:02  nyef
 * reduced the 0x4000s in the genesis/pce detect to 0x2000 (this doesn't
 * affect the genesis detect much, and fixes the pce detect to catch some
 * roms produced with the MagicKit assembler).
 *
 * Revision 1.9  1999/01/01 06:26:50  nyef
 * added support for Master System.
 *
 * Revision 1.8  1998/12/21 02:58:59  nyef
 * added char * romfilename to save the name of the rom file.
 *
 * Revision 1.7  1998/10/11 20:34:33  nyef
 * added SMS/GG detection to read_romimage()
 * added preliminary GB detection to read_romimage()
 *
 * Revision 1.6  1998/08/02 13:42:40  nyef
 * changed every printf to deb_printf for debug console.
 *
 * Revision 1.5  1998/08/01 01:07:44  nyef
 * fixed stupid mistake in activate_system().
 *
 * Revision 1.4  1998/08/01 01:03:20  nyef
 * moved main() out to the video interfaces. retained some of the logic in
 * activate_system().
 *
 * Revision 1.3  1998/07/12 23:07:07  nyef
 * added a romsize variable and fixed read_romimage() to set it
 *
 * Revision 1.2  1998/07/12 20:52:06  nyef
 * enabled PC-Engine emulation
 *
 * Revision 1.1  1998/07/11 22:19:09  nyef
 * Initial revision
 *
 */
