/*
 * ui_x.c
 *
 * user interface for the X Window System.
 */

/* $Id: ui_x.c,v 1.15 1999/04/17 20:12:46 nyef Exp $ */

#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/MenuButton.h>
#include <X11/Xaw/SimpleMenu.h>
#include <X11/Xaw/SmeBSB.h>

#include <stdio.h>
#include <stdarg.h>
#include "ui.h"
#include "system.h"
#include "video.h"
#include "tool.h"

/* Overhead for Xt and main window */
XtAppContext context;
Widget toplevel;
Widget manager;

/* Menus */
Widget menubar;
Widget menubutton_file;
Widget menu_file;
#if 0
Widget menu_file_nes;
Widget menu_file_zarzon;
#endif
Widget menu_file_exit;
#if 0
Widget menubutton_debug;
Widget menu_debug;
Widget menu_debug_console;
Widget menu_debug_regs;
Widget menu_debug_disas;
#endif

/* the drawing area */
Widget drawbox;

#if 0
/* the code window */
Widget code_shell;
Widget code_manager;
Widget linelabels[16];
#endif

/* the shutdown callback */
shutdown_t dn_shutdown;

/* miscellany */
void do_file_nes(Widget, XtPointer, XtPointer);
void do_file_zarzon(Widget, XtPointer, XtPointer);
void do_file_exit(Widget, XtPointer, XtPointer);
void do_debug_console(Widget, XtPointer, XtPointer);
void do_debug_regs(Widget, XtPointer, XtPointer);
void do_debug_disas(Widget, XtPointer, XtPointer);
Boolean emulation_timeslice(XtPointer);

void init_menus(void)
{
    /* the menubar */
    menubar = XtVaCreateManagedWidget("menubar", boxWidgetClass, manager, XtNorientation, XtorientHorizontal, XtNborderWidth, 0, NULL);

    /* the file menu */
    menubutton_file = XtCreateManagedWidget("File", menuButtonWidgetClass, menubar, NULL, 0);
    menu_file = XtCreatePopupShell("fileMenu", simpleMenuWidgetClass, menubutton_file, NULL, 0);
    XtVaSetValues(menubutton_file, XtNmenuName, "fileMenu", NULL);
#if 0
    menu_file_nes = XtCreateManagedWidget("Nes", smeBSBObjectClass, menu_file, NULL, 0);
    XtAddCallback(menu_file_nes, XtNcallback, do_file_nes, NULL);
#endif
#if 0
    menu_file_zarzon = XtCreateManagedWidget("Zarzon", smeBSBObjectClass, menu_file, NULL, 0);
    XtAddCallback(menu_file_zarzon, XtNcallback, do_file_zarzon, NULL);
#endif
    menu_file_exit = XtCreateManagedWidget("Exit", smeBSBObjectClass, menu_file, NULL, 0);
    XtAddCallback(menu_file_exit, XtNcallback, do_file_exit, NULL);

#if 0
    /* the debug menu */
    menubutton_debug = XtCreateManagedWidget("Debug", menuButtonWidgetClass, menubar, NULL, 0);
    menu_debug = XtCreatePopupShell("debugMenu", simpleMenuWidgetClass, menubutton_debug, NULL, 0);
    XtVaSetValues(menubutton_debug, XtNmenuName, "debugMenu", NULL);
    menu_debug_console = XtCreateManagedWidget("Show console", smeBSBObjectClass, menu_debug, NULL, 0);
    XtAddCallback(menu_debug_console, XtNcallback, do_debug_console, NULL);
    menu_debug_regs = XtCreateManagedWidget("Dump regs", smeBSBObjectClass, menu_debug, NULL, 0);
    XtAddCallback(menu_debug_regs, XtNcallback, do_debug_regs, NULL);
    menu_debug_disas = XtCreateManagedWidget("Disassemble", smeBSBObjectClass, menu_debug, NULL, 0);
    XtAddCallback(menu_debug_disas, XtNcallback, do_debug_disas, NULL);
#endif
}

#if 0
void init_codeshell(void)
{
    int i;
    char buf[12];
    
    code_shell = XtVaCreatePopupShell("codeShell", topLevelShellWidgetClass, manager, NULL);
    code_manager = XtVaCreateManagedWidget("codeManager", boxWidgetClass, code_shell, XtNhSpace, 0, XtNvSpace, 0, NULL);

    for (i = 0; i < 16; i++) {
	sprintf(buf, "linelabel%d\n", i);
	linelabels[i] = XtVaCreateManagedWidget(buf, labelWidgetClass, code_manager, XtNlabel, "insert code here", NULL);
    }
}
#endif

int main(int argc, char *argv[])
{
    toplevel = XtAppInitialize(&context, "DarcNES", NULL, 0, &argc, argv, NULL, NULL, 0);
    manager = XtVaCreateManagedWidget("manager", boxWidgetClass, toplevel, XtNhSpace, 0, XtNvSpace, 0, NULL);

    init_menus();
    
    drawbox = XtVaCreateManagedWidget("drawbox", boxWidgetClass, manager, XtNwidth, 512, XtNheight, 256, NULL);

#if 0
    init_codeshell();
#endif
    
    XtRealizeWidget(toplevel);
    
    video_init();

    if (argc > 1) {
	read_romimage(argv[1]);
    } else {
	read_romimage("cart.nes");
    }
    activate_system(system_type, 0);

    XtAppMainLoop(context);
    
    return 0;
}


/* debug console handling */

void deb_printf(const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    vprintf(fmt, args);
    va_end(args);
}


/* emulation timeslicing */

void (*timeslice)(void *) = NULL;
void *timeslice_data;

void set_timeslice(void (*proc)(void *), void *data)
{
    if (!timeslice) {
	XtAppAddWorkProc(context, emulation_timeslice, NULL);
    }
    
    timeslice = proc;
    timeslice_data = data;
}

void unset_timeslice(void)
{
    timeslice = NULL;
    timeslice_data = NULL;
}

Boolean emulation_timeslice(XtPointer client_data)
{
    if (timeslice) {
	timeslice(timeslice_data);
    }
    
    if ((system_flags == F_NONE) && timeslice) {
	return False;
    } else {
	return True;
    }
}


/* menu callbacks */

#if 0
void do_file_nes(Widget w, XtPointer client_data, XtPointer call_data)
{
    read_romimage("cart.nes");
    activate_system(system_type, 0);
}
#endif

#if 0
void do_file_zarzon(Widget w, XtPointer client_data, XtPointer call_data)
{
    void zarzon_run(int);
    zarzon_run(0);
}
#endif

void do_file_exit(Widget w, XtPointer client_data, XtPointer call_data)
{
    if (dn_shutdown) {
	dn_shutdown();
    }
    exit(0);
}

#if 0
void do_debug_console(Widget w, XtPointer client_data, XtPointer call_data)
{
    printf("do_debug_console()\n");
    XtPopup(code_shell, XtGrabNone);
}

#include "debug.h"
#include "emu6502.h"

void do_debug_regs(Widget w, XtPointer client_data, XtPointer call_data)
{
    dump_regs(timeslice_data);
}

void do_debug_disas(Widget w, XtPointer client_data, XtPointer call_data)
{
    disassemble(((struct emu6502_context *)timeslice_data)->reg_PC, ((((struct emu6502_context *)timeslice_data)->reg_PC+16) & 0xffff));
}
#endif

/* video interface */

void ui_set_drawbox_size(int width, int height)
{
    XtVaSetValues(drawbox, XtNwidth, width, XtNheight, height, NULL);
}

/*
 * $Log: ui_x.c,v $
 * Revision 1.15  1999/04/17 20:12:46  nyef
 * changed shutdown() to dn_shutdown().
 *
 * Revision 1.14  1999/03/03 02:53:05  nyef
 * changed the initial size of the drawbox to "really big"
 * as a workaround for the resize problem
 *
 * Revision 1.13  1999/02/13 17:25:34  nyef
 * disabled file->nes.
 * added code to accept a filename on the command line.
 * changed to start running game immediately after starting.
 *
 * Revision 1.12  1999/01/08 02:45:25  nyef
 * disabled file->zarzon.
 *
 * Revision 1.11  1998/12/24 04:56:51  nyef
 * moved for loop from emulation_timeslice() to nes.c.
 *
 * Revision 1.10  1998/12/24 01:23:18  nyef
 * removed debug menu (it was at best useless, and at worst dangerous).
 * added two file menu items ("Nes", and "Zarzon") to start the system.
 * fixed file->quit to call exit(3).
 * fixed background task to only run when we need an emulation timeslice.
 *
 * Revision 1.9  1998/12/21 02:58:03  nyef
 * added a shutdown callback.
 *
 * Revision 1.8  1998/12/18 04:22:25  nyef
 * changed application class to "DarcNES".
 *
 * Revision 1.7  1998/10/28 01:50:00  nyef
 * changed emulation_timeslice() to return aproximately once per screen
 * refresh. this sped up the emulation appreciably but does not appear
 * to have degraded the key response.
 *
 * Revision 1.6  1998/10/02 01:22:07  nyef
 * changed timeslice function to remove itself when emulation halts.
 *
 * Revision 1.5  1998/08/24 01:25:41  nyef
 * started working on new interface.
 *
 * Revision 1.4  1998/08/02 04:01:56  nyef
 * added function deb_printf() for debug console support.
 * added ui.h to the list of includes.
 *
 * Revision 1.3  1998/08/02 03:56:28  nyef
 * added comments grouping functions into related sections.
 *
 * Revision 1.2  1998/08/02 02:31:47  nyef
 * added a separate menubar widget.
 * added a "Debug" menu with one item. ("Show console", which calls a stub.)
 *
 * Revision 1.1  1998/08/02 02:04:26  nyef
 * Initial revision
 *
 */
