/*
 * zarzon.c
 *
 * zarzon machine emulation.
 */

/* $Id: zarzon.c,v 1.2 1998/12/31 04:36:22 nyef Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cal.h"
#include "ui.h"
#include "tool.h"
#include "video.h"
#include "nes_blit.h"

#define ZARZON_RAMSIZE 0x100 + 0xc00 + 0x1000 /* ram + display tables + shape patterns */
#define ZARZON_ROMSIZE 0x5800

unsigned char *zarzon_ram;
unsigned char *zarzon_shapes0;
unsigned char *zarzon_shapes1;
unsigned char *zarzon_rom;
unsigned char *zarzon_vrom0;
unsigned char *zarzon_vrom1;

unsigned char zarzon_palette[32] = {
    0, 1, 2, 3, 0, 1, 2, 3,
    0, 1, 2, 3, 0, 1, 2, 3,
    0, 1, 2, 3, 0, 1, 2, 3,
    0, 1, 2, 3, 0, 1, 2, 3,
};

unsigned char zarzon_back; /* background color */

unsigned char zarzon_read_ram(cal_cpu cpu, unsigned short addr)
{
    if (!(addr & 0x400)) {
	return zarzon_ram[addr & 0xff];
    } else {
	return zarzon_ram[addr - 0x300];
    }
}

unsigned char zarzon_read_vram(cal_cpu cpu, unsigned short addr)
{
    /* FIXME: range 0xc00 - 0xfff should only be 4 bits wide */
    return zarzon_ram[addr - 0x300];
}

unsigned char zarzon_read_rom(cal_cpu cpu, unsigned short addr)
{
    return zarzon_rom[addr - 0x4000];
}

unsigned char zarzon_read_rom8(cal_cpu cpu, unsigned short addr)
{
    return zarzon_rom[(addr & 0x7ff) + 0x3800];
}

unsigned char zarzon_read_trap(cal_cpu cpu, unsigned short addr)
{
    deb_printf("zarzon_read_trap: 0x%04hx, returning 0x00.\n", addr);
    return 0;
}

void zarzon_write_ram(cal_cpu cpu, unsigned short addr, unsigned char data)
{
    if (!(addr & 0x400)) {
	zarzon_ram[addr & 0xff] = data;
    } else {
	deb_printf("zarzon: vram write 0x%04hx:0x%02x.\n", addr, data);
	zarzon_ram[addr - 0x300] = data;
    }
}

void zarzon_write_vram(cal_cpu cpu, unsigned short addr, unsigned char data)
{
    if (addr < 0x1000) {
	deb_printf("zarzon: vram write 0x%04hx:0x%02x.\n", addr, data);
    }
    zarzon_ram[addr - 0x300] = data;
}

void zarzon_write_ignore(cal_cpu cpu, unsigned short addr, unsigned char data)
{
}

void zarzon_write_trap(cal_cpu cpu, unsigned short addr, unsigned char data)
{
    deb_printf("zarzon_write_trap: 0x%04hx, 0x%02x, ignoring.\n", addr, data);
}

unsigned char zarzon_read_crtc(cal_cpu cpu, unsigned short addr)
{
    /* FIXME: add crtc read */
/*     return zarzon_read_trap(cpu, addr); */
    return 0;
}

void zarzon_write_crtc(cal_cpu cpu, unsigned short addr, unsigned char data)
{
    /* FIXME: add crtc write */
/*     zarzon_write_trap(cpu, addr, data); */
}

unsigned char zarzon_read_io(cal_cpu cpu, unsigned short addr)
{
    switch (addr & 7) {
    case 0: /* Sound */
    case 1: /* Sound */
    case 2: /* Video? */
    case 3: /* BG Color? */
	deb_printf("zarzon: read from write-only port %d.\n", addr & 3);
	break;
    case 4: /* Inputs? */
    case 5: /* Inputs? */
    case 6: /* DIP Switch? */
    case 7: /* ??? */
	/* FIXME: implimentation? */
    }
/*     return zarzon_read_trap(cpu, addr); */
    return 0;
}

void zarzon_write_io(cal_cpu cpu, unsigned short addr, unsigned char data)
{
    switch (addr & 7) {
    case 0: /* Sound */
    case 1: /* Sound */
    case 2: /* Video? */
	/* FIXME: implimentation? */
	break;
    case 3: /* BG Color? */
	zarzon_back = data & 3;
	break;
    case 4: /* Inputs? */
    case 5: /* Inputs? */
    case 6: /* DIP Switch? */
    case 7: /* ??? */
	deb_printf("zarzon: write to read-only port %d.\n", addr & 3);
	break;
    }
/*     zarzon_write_trap(cpu, addr, data); */
}

memread_t zarzon_readtable[32] = {
    zarzon_read_ram,      /* 0000 - 07ff */
    zarzon_read_vram,     /* 0800 - 0fff */
    zarzon_read_vram,     /* 1000 - 17ff */
    zarzon_read_vram,     /* 1800 - 1fff */
    zarzon_read_trap,     /* 2000 - 27ff */ /* Empty? */
    zarzon_read_trap,     /* 2800 - 2fff */ /* Empty? */
    zarzon_read_crtc,     /* 3000 - 37ff */
    zarzon_read_crtc,     /* 3800 - 3fff */
    zarzon_read_rom,      /* 4000 - 47ff */ /* ZARZ122.07 */
    zarzon_read_rom,      /* 4800 - 4fff */ /* ZARZ123.08 */
    zarzon_read_rom,      /* 5000 - 57ff */ /* ZARZ124.09 */
    zarzon_read_rom,      /* 5800 - 5fff */ /* ZARZ125.10 */
    zarzon_read_rom,      /* 6000 - 67ff */ /* ZARZ126.13 */
    zarzon_read_rom,      /* 6800 - 6fff */ /* ZARZ127.14 */
    zarzon_read_rom,      /* 7000 - 77ff */ /* ZARZ128.15 */
    zarzon_read_rom,      /* 7800 - 7fff */ /* ZARZ129.16 */
    zarzon_read_rom,      /* 8000 - 87ff */ /* ZARZ130.22 */
    zarzon_read_rom,      /* 8800 - 8fff */ /* ZARZ131.23 */
    zarzon_read_rom,      /* 9000 - 97ff */ /* ZARZ132.24 */
    zarzon_read_trap,     /* 9800 - 9fff */ /* Unused ROM socket */
    zarzon_read_trap,     /* a000 - a7ff */ /* Empty? */
    zarzon_read_trap,     /* a800 - afff */ /* Empty? */
    zarzon_read_io,       /* b000 - b7ff */ /* I/O */
    zarzon_read_trap,     /* b800 - bfff */ /* Empty? */
    zarzon_read_rom8,     /* c000 - c7ff */ /* ZARZ129.16 alias */
    zarzon_read_rom8,     /* c800 - cfff */ /* ZARZ129.16 alias */
    zarzon_read_rom8,     /* d000 - d7ff */ /* ZARZ129.16 alias */
    zarzon_read_rom8,     /* d800 - dfff */ /* ZARZ129.16 alias */
    zarzon_read_rom8,     /* e000 - e7ff */ /* ZARZ129.16 alias */
    zarzon_read_rom8,     /* e800 - efff */ /* ZARZ129.16 alias */
    zarzon_read_rom8,     /* f000 - f7ff */ /* ZARZ129.16 alias */
    zarzon_read_rom8,     /* f800 - ffff */ /* ZARZ129.16 alias */
};

memwrite_t zarzon_writetable[32] = {
    zarzon_write_ram,     /* 0000 - 07ff */
    zarzon_write_vram,    /* 0800 - 0fff */
    zarzon_write_vram,    /* 1000 - 17ff */
    zarzon_write_vram,    /* 1800 - 1fff */
    zarzon_write_trap,    /* 2000 - 27ff */ /* Empty? */
    zarzon_write_trap,    /* 2800 - 28ff */ /* Empty? */
    zarzon_write_crtc,    /* 3000 - 37ff */
    zarzon_write_crtc,    /* 3800 - 3fff */
    zarzon_write_ignore,  /* 4000 - 47ff */ /* ROM */
    zarzon_write_ignore,  /* 4800 - 4fff */ /* ROM */
    zarzon_write_ignore,  /* 5000 - 57ff */ /* ROM */
    zarzon_write_ignore,  /* 5800 - 5fff */ /* ROM */
    zarzon_write_ignore,  /* 6000 - 67ff */ /* ROM */
    zarzon_write_ignore,  /* 6800 - 6fff */ /* ROM */
    zarzon_write_ignore,  /* 7000 - 77ff */ /* ROM */
    zarzon_write_ignore,  /* 7800 - 7fff */ /* ROM */
    zarzon_write_ignore,  /* 8000 - 87ff */ /* ROM */
    zarzon_write_ignore,  /* 8800 - 8fff */ /* ROM */
    zarzon_write_ignore,  /* 9000 - 97ff */ /* ROM */
    zarzon_write_ignore,  /* 9800 - 9fff */ /* ROM */
    zarzon_write_trap,    /* a000 - a7ff */ /* Empty? */
    zarzon_write_trap,    /* a800 - afff */ /* Empty? */
    zarzon_write_io,      /* b000 - b7ff */ /* I/O */
    zarzon_write_trap,    /* b800 - bfff */ /* Empty? */
    zarzon_write_ignore,  /* c000 - c7ff */ /* ROM alias */
    zarzon_write_ignore,  /* c800 - cfff */ /* ROM alias */
    zarzon_write_ignore,  /* d000 - d7ff */ /* ROM alias */
    zarzon_write_ignore,  /* d800 - dfff */ /* ROM alias */
    zarzon_write_ignore,  /* e000 - e7ff */ /* ROM alias */
    zarzon_write_ignore,  /* e800 - efff */ /* ROM alias */
    zarzon_write_ignore,  /* f000 - f7ff */ /* ROM alias */
    zarzon_write_ignore,  /* f800 - f8ff */ /* ROM alias */
};

void zarzon_render_scanline(int scanline)
{
    int i;
    unsigned char *cur_vbp;
    unsigned char cur_attrs;
    unsigned long colors;
    unsigned char cur_tile;
    
    cur_vbp = video_get_vbp(scanline);

    for (i = 0; i < 32; i++) {
	cur_attrs = zarzon_ram[0x900 + ((scanline << 2) & 0x3e0) + i];
	memset(cur_vbp, vid_pre_xlat[zarzon_palette[(0x10 + ((cur_attrs & 3) << 2)) | zarzon_back]], 8);

	cur_tile = zarzon_ram[0x100 + ((scanline << 2) & 0x3e0) + i];
	colors = *((unsigned long *)&zarzon_palette[0x10 + ((cur_attrs & 3) << 2)]);
	nes_do_blit[zarzon_vrom0[(cur_tile << 3) + (scanline & 7)]](cur_vbp, colors, zarzon_vrom1[(cur_tile << 3) + (scanline & 7)]);

	cur_tile = zarzon_ram[0x500 + ((scanline << 2) & 0x3e0) + i];
	colors = *((unsigned long *)&zarzon_palette[cur_attrs & 0xc]);
	nes_do_blit[zarzon_shapes0[(cur_tile << 3) + (scanline & 7)]](cur_vbp, colors, zarzon_shapes1[(cur_tile << 3) + (scanline & 7)]);

	cur_vbp += 8;
    }
}

void zarzon_timeslice(cal_cpu cpu)
{
    int i;

    for (i = 0; i < 256; i++) {
	if (system_flags == F_NONE) {
	    cpu->runfor(cpu, 113); /* FIXME: 113 cycles is a bit arbitrary. */
	    cpu->run(cpu);
	    if (i < 240) {
		zarzon_render_scanline(i);
	    }
	    if (system_flags & F_UNIMPL) {
		deb_printf("Unimplimented Instruction\n");
	    }
	}
    }
    if (system_flags == F_NONE) {
	cpu->irq(cpu);
    }
    video_display_buffer();
}

static unsigned char *romnames[11] = {
    "zarzon/ZARZ122.07", "zarzon/ZARZ123.08", "zarzon/ZARZ124.09",
    "zarzon/ZARZ125.10", "zarzon/ZARZ126.13", "zarzon/ZARZ127.14",
    "zarzon/ZARZ128.15", "zarzon/ZARZ129.16", "zarzon/ZARZ130.22",
    "zarzon/ZARZ131.23", "zarzon/ZARZ132.24",
};

static unsigned char *vromnames[2] = {
    "zarzon/ZARZ135.73", "zarzon/ZARZ136.75",
};

void zarzon_run(int startup_deb)
{
    cal_cpu cpu;

    blit_init();
    
    zarzon_ram = calloc(ZARZON_RAMSIZE, 1);
    zarzon_rom = malloc(ZARZON_ROMSIZE);

    zarzon_shapes0 = zarzon_ram + 0xd00;
    zarzon_shapes1 = zarzon_shapes0 + 0x800;

    zarzon_vrom0 = malloc(0x800);
    zarzon_vrom1 = malloc(0x800);

    if (zarzon_ram && zarzon_rom) {
	int i;
	int errors;
	FILE *romfile;
	unsigned char *romptr;

	errors = 0;
	romptr = zarzon_rom;
	
	for (i = 0; i < 11; i++) {
	    if (!(romfile = fopen(romnames[i], "rb"))) {
		deb_printf("error opening %s: %s.\n", romnames[i], strerror(errno));
		errors++;
	    } else if (fread(romptr, 1, 0x800, romfile) < 0x800) {
		deb_printf("error reading %s: %s.\n", romnames[i], strerror(ferror(romfile)));
		errors++;
	    }
	    fclose(romfile);
	    romptr += 0x800;
	}

	if (!(romfile = fopen(vromnames[0], "rb"))) {
	    deb_printf("error opening %s: %s.\n", vromnames[0], strerror(errno));
	    errors++;
	} else if (fread(zarzon_vrom0, 1, 0x800, romfile) < 0x800) {
	    deb_printf("error reading %s: %s.\n", vromnames[0], strerror(ferror(romfile)));
	    errors++;
	}
	fclose(romfile);
	
	if (!(romfile = fopen(vromnames[1], "rb"))) {
	    deb_printf("error opening %s: %s.\n", vromnames[1], strerror(errno));
	    errors++;
	} else if (fread(zarzon_vrom1, 1, 0x800, romfile) < 0x800) {
	    deb_printf("error reading %s: %s.\n", vromnames[1], strerror(ferror(romfile)));
	    errors++;
	}
	fclose(romfile);
	
	if (errors) {
	    free(zarzon_rom);
	    free(zarzon_ram);
	    zarzon_rom = zarzon_ram = NULL;
	}
    }
    
    if (zarzon_ram && zarzon_rom) {
	cpu = cal_create(CPUT_MARAT6502);
	cpu->setzpage(cpu, zarzon_ram);
	cpu->setmmu(cpu, 11, 31, zarzon_readtable, zarzon_writetable);
	cpu->reset(cpu);

	set_timeslice(&zarzon_timeslice, cpu);
    }
}

/*
 * $Log: zarzon.c,v $
 * Revision 1.2  1998/12/31 04:36:22  nyef
 * added logging of memory writes in the 0x400-0xfff range.
 * tweaked video renderer.
 *
 * Revision 1.1  1998/12/24 04:32:44  nyef
 * Initial revision
 *
 */
